/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aeAnimation.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aeanimation.h"
#include "aedebug.h"
#include "asig.h"
#include "aefileio.h"



/*-------------------
Animation Settings
    Settings for an animation, such as speed and repeat
    Some subtypes allow more than one set of settings.
-------------------*/
typedef struct 
{
    int                 speed;
    int                 repeat;
    int                 runLength;
    aeAnimStyle_t       style;      /* single-pass or wrapped */
    aeAnimDirection_t   direction;  /* horizontal or vertical */
    int                 vector;
    aeAnimTransition_t  transitionStyle;
    int                 transitionDuration;
    int                 reserved0;
    int                 reserved1;
    char                hideAnimFrame;
}aeAnimSettings_t; 

/*-------------------
Animation Data
    Complete information about an animation
    Includes architecture (list of function ptrs), data, and layouts
-------------------*/
#define ANIM_MAX_BYTES_PER_FRAME 64

struct aeAnimationData_tag
{
    int                 safety;
    int                 signature;
    aeAnimSubtype_t     subtype;
    aeAnimSubtype_t     editSubtype;
    int                 layoutWidth;
    int                 layoutHeight;
    char                name[ANIM_NAME_LEN];
    char                uniqueSuffix[ANIM_NAME_UNIQUE_SUFFIX_LEN];  /* Set by sequence if name must distiquished from another name */
    char*               pathptr;
    aeSequence_t*       sequence;
    aeAnimSettings_t    settings[ANIM_MAX_VARIANTS];
    int                 currSetting;
    int                 settingsCount;
    int                 animFrameIndex;
    int                 animRepeatIndex;
    int                 editFrameIndex;
    int                 changed;
    AnimIOState_t       ioState;
    int                 ioCount;
    aErr_t              err;            /* To track internal errors, such as during reading */
    int                 flags;
    char                hideAnimFrame;
    char                needsResizeOnFrameCountChange;
    int                 ledsPerEntry;
    int                 bytesPerEntry;  /* For data storage */
    int                 bytesPerFrame;  /* For animated frame */
    int                 bitsPerPixel;   /* For animated frame */
    int                 frameCount;
    int                 allocatedCount;
    void*               entries;
    unsigned long*      animFrame;
    unsigned char       animFrameData[ANIM_MAX_BYTES_PER_FRAME];
};



/*----------------------------------------------------------------------
    Constants
----------------------------------------------------------------------*/
#define ANIMATION_SAFETY 0x1d7733ee
#define ANIM_ALLOCATE_INCREMENT 10

#define AECHECKANIM(_anim_) AEASSERT(_anim_ && _anim_->safety == ANIMATION_SAFETY)

/*----------------------------------------------------------------------
    Prototypes
----------------------------------------------------------------------*/
static void         aeAnimation_p_SetDefaults(aeAnimationData_t* animData, aeAnimationDataSettings_t* settings);
static aErr_t       aeAnimation_p_IncreaseEntryAllocation(aeAnimationData_t* animData, int newAllocatedCount);
static void*        aeAnimation_p_GetEntryPtr(aeAnimationData_t* animData, int frameIndex);
static unsigned int aeAnimation_p_GetEntryValue(aeAnimationData_t* animData, int frameIndex);
static unsigned int aeAnimation_p_SetEntryValue(aeAnimationData_t* animData, int frameIndex, unsigned int value);
static char         aeAnimation_p_ClearFrame(aeAnimationData_t* animData, int frameIndex);
static void         aeAnimation_p_SetError(aeAnimationData_t* animData, aErr_t err);
static void         aeAnimation_p_CalcOffsetShiftMask(aeAnimationData_t* animData, int ledIndex, char calcForEntry, 
                                    unsigned int* pValueOffset, unsigned int* pValueShift, unsigned int* pValueMask);
static void         aeAnimation_p_CalcPtrShiftMask(aeAnimationData_t* animData, int frameIndex, int ledIndex, 
                                    unsigned char** pptr, unsigned int* pValueShift, unsigned int* pValueMask);
static aeLedValue_t aeAnimation_p_GetLedValueField(aeAnimationData_t* animData, int frameIndex, int ledIndex);
static aeLedValue_t aeAnimation_p_SetLedValueField(aeAnimationData_t* animData, int frameIndex, int ledIndex, unsigned int value);

#define CALC_FOR_FRAME FALSE
#define CALC_FOR_ENTRY TRUE

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_InitDataSettings(aeAnimationDataSettings_t* settings)
{
    AEASSERT(settings);

    memset(settings, 0, sizeof(aeAnimationDataSettings_t));
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimationData_t* aeAnimation_Create(aeAnimationDataSettings_t* settings)
{
    aeAnimationData_t* animData;
    aErr_t             err;

    /* Always allocate at least one item */
    if (settings->allocateCount == 0)
        settings->allocateCount = 1;

    animData = malloc(sizeof(aeAnimationData_t));

    if (animData != NULL)
    {
        memset(animData, 0, sizeof(aeAnimationData_t));
        animData->safety = ANIMATION_SAFETY;
        aeAnimation_p_SetDefaults(animData, settings);

        if (settings->allocateCount > 0)
            err = aeAnimation_p_IncreaseEntryAllocation(animData, settings->allocateCount);
        else
            err = AERR_NO_ERROR;

        if (err == AERR_NO_ERROR)
        {
            /* Must always have at least one frame, but we */
            /* can't set this until the entries data is allocated */
            animData->frameCount = 1;

            /* To simplify smaller animations.... */
            animData->animFrame = (unsigned long*)animData->animFrameData;
        }
        else
        {
            free(animData);
            animData = NULL;
        }
    }

    return animData;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeAnimation_p_IncreaseEntryAllocation(aeAnimationData_t* animData, int newAllocatedCount)
{
    aErr_t  err = AERR_NO_ERROR;
    void*   entries;
    int     entriesSize;

    AECHECKANIM(animData);
    AEASSERT(newAllocatedCount > animData->allocatedCount);

    entriesSize = newAllocatedCount * animData->bytesPerEntry;

    entries = malloc(entriesSize);

    if (entries != NULL)
    {
        /* Clear the new block */
        memset(entries, 0, entriesSize);

        if (animData->frameCount > 0)
        {
            AEASSERT(animData->entries);
            /* Copy the old to the new */
            memcpy(entries, animData->entries, animData->frameCount * animData->bytesPerEntry);
        }

        if (animData->entries != NULL)
        {
            /* Free the old */
            free(animData->entries);
            animData->entries = NULL;
        }

        /* Update the allocated number */
        animData->allocatedCount = newAllocatedCount;

        /* Assign the new */
        animData->entries = entries;
    }
    else
    {
        err = AERR_OUT_OF_MEMORY;
    }

    return err;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_p_SetDefaults(aeAnimationData_t* animData, aeAnimationDataSettings_t* settings)
{
    AECHECKANIM(animData);
    AEASSERT(settings);
    AEASSERT(settings->bitsPerPixel > 0);
    AEASSERT(settings->bytesPerEntry > 0);
    AEASSERT(settings->bytesPerFrame > 0);
    /* Must know the number of pixels/LEDs */
    AEASSERT(settings->ledsPerEntry);


    strcpy(animData->name, "untitled");


    animData->signature   = settings->signature;
    animData->subtype     = settings->subtype;
    animData->editSubtype = settings->subtype;
    animData->frameCount  = 0;

    animData->ledsPerEntry  = settings->ledsPerEntry;
    animData->bitsPerPixel  = settings->bitsPerPixel;
    animData->bytesPerEntry = settings->bytesPerEntry;
    animData->bytesPerFrame = settings->bytesPerFrame;
    
    animData->settings[0].speed     = 0x80;
    animData->settings[0].repeat    = 4;
    animData->settings[0].direction	= aeAnimDirection_eVertical;
    animData->settings[0].runLength = 1;
    animData->settings[0].style     = aeAnimStyle_eSinglePass;

	animData->currSetting   = 0;
    animData->settingsCount = 1;

    /* Use for strip animations */
    animData->layoutWidth  = settings->layoutWidth;
    animData->layoutHeight = settings->layoutHeight;

    if (settings->subtype == aeAnimSubtype_eStrip)
    {
        AEASSERT(settings->layoutWidth > 0 && settings->layoutHeight > 0);
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_Delete(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

    if (animData->entries != NULL)
    {
        free(animData->entries);
        animData->entries = NULL;
    }

    if (animData->pathptr != NULL)
    {
        free(animData->pathptr);
        animData->pathptr = NULL;
    }

    memset(animData, 0, sizeof(aeAnimationData_t));
    free(animData);
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimationData_t* aeAnimation_Copy(aeAnimationData_t* animData)
{
    aeAnimationData_t*          copyData;
    aeAnimationDataSettings_t   settings;
    void*                       entries;

    aeAnimation_InitDataSettings(&settings);
    settings.signature      = animData->signature;
    settings.subtype        = animData->subtype;
    settings.layoutWidth    = animData->layoutWidth;
    settings.layoutHeight   = animData->layoutHeight;
    settings.allocateCount  = animData->frameCount;
    settings.bitsPerPixel   = animData->bitsPerPixel;
    settings.bytesPerEntry  = animData->bytesPerEntry;
    settings.bytesPerFrame  = animData->bytesPerFrame;
    settings.ledsPerEntry   = animData->ledsPerEntry;

    copyData = aeAnimation_Create(&settings);

    if (copyData != NULL)
    {
        /* Cache entries data */
        entries = copyData->entries;

        /* Copy everything */
        *copyData = *animData;

        /* Restore the entries data */
        copyData->entries = entries;

        /* Copy the entries */
        if (copyData->frameCount > 0)
            memcpy(copyData->entries, animData->entries, copyData->frameCount * copyData->bytesPerEntry);

        /* Data owned by original does not transfer */
        copyData->pathptr  = NULL;

        copyData->sequence = NULL;
    }

    return copyData;
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetEditSubtype(aeAnimationData_t* animData, aeAnimSubtype_t editSubtype)
{
    AECHECKANIM(animData);

    animData->editSubtype = editSubtype;

    if (editSubtype == aeAnimSubtype_eStrip)
    {
        AEASSERT(animData->layoutWidth > 0 && animData->layoutHeight > 0);
    }

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_Reset(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
	animData->animFrameIndex = 0;
	animData->animRepeatIndex = 0;
}

/*----------------------------------------------------------------------
Summary
    Returns TRUE if animation at first frame of first pass.    

Desc
    
----------------------------------------------------------------------*/
char aeAnimation_IsAtInitialAnimationState(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    return 	animData->animFrameIndex == 0 && animData->animRepeatIndex == 0;
}


/*----------------------------------------------------------------------
Summary
    Sets the 'hide animatin flag'

Desc
    Used by editor window to show the end of a repeat sequence by
    hiding the animation data.
----------------------------------------------------------------------*/
void aeAnimation_HideAnimatedFrame(aeAnimationData_t* animData, char hideFrame)
{
    AECHECKANIM(animData);
    animData->hideAnimFrame = hideFrame;
}

/*----------------------------------------------------------------------
Summary
    Returns flag indicating if animated frame is hidden or not    

Desc
    Used by layout to suppress animation display so that edit window
    can show end of a repeat sequence
----------------------------------------------------------------------*/
char aeAnimation_GetHideAnimatedFrame(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    return animData->hideAnimFrame;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_Advance(aeAnimationData_t* animData)
{
	int	    steps = 1;
    char    needsRotate;
    int     scrollDirectionSize;

	AECHECKANIM(animData);

	if (animData->frameCount > 0)
	{
		animData->animFrameIndex++;

		/* Frame-by-Frame */
		if (animData->subtype == aeAnimSubtype_eFrameByFrame)
		{
			steps = animData->frameCount;
		}

		/* Pixel List */
		else if (animData->subtype == aeAnimSubtype_ePixelList)
		{
			steps = animData->frameCount;
			
			if (animData->animRepeatIndex == 0 || animData->settings[animData->currSetting].style == aeAnimStyle_eSinglePass)
				steps += animData->settings[animData->currSetting].runLength;
		}

		/* Strip */
		else if (animData->subtype == aeAnimSubtype_eStrip)
		{
            AEASSERT(animData->layoutHeight > 0);
			steps = animData->frameCount;

            /* By definition, "vertical direction" is not rotated, and "horizontal direction" is rotated */
            needsRotate = (animData->settings[animData->currSetting].direction == aeAnimDirection_eHorizontal);
            scrollDirectionSize    = needsRotate ? animData->layoutWidth  : animData->layoutHeight;
			/* Add scroll-on time */
			if ((animData->settings[animData->currSetting].style == aeAnimStyle_eSinglePass) ||
				(animData->animRepeatIndex == 0))
				steps += scrollDirectionSize;
		}
		
		else
			AEASSERTED("unknown subtype");

		/* If index reached end, then reset index and advance repeat */
		if (animData->animFrameIndex >= steps)
		{
			animData->animFrameIndex = 0;
			animData->animRepeatIndex++;

			if (animData->animRepeatIndex >= animData->settings[animData->currSetting].repeat)
				animData->animRepeatIndex = 0;
		}
	}

	aeAnimation_BuildFrame(animData);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static void aeAnimation_BuildFrame_FrameByFrame(aeAnimationData_t *animData)
{
    unsigned char*   ptr;

    if (animData->bytesPerEntry <= sizeof(int))
        animData->animFrame[0] = aeAnimation_p_GetEntryValue(animData, animData->animFrameIndex);
    else
    {
        ptr = aeAnimation_p_GetEntryPtr(animData, animData->animFrameIndex);

        memcpy(animData->animFrameData, ptr, animData->bytesPerEntry);
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static void aeAnimation_BuildFrame_PixelList(aeAnimationData_t* animData)
{
    int	i;
    int index;
    unsigned int     pixelIndex;
    unsigned int     valueOffset, valueShift, valueMask;

    if (animData->bytesPerFrame <= sizeof(int))
        animData->animFrame[0] = 0;
    else
        memset(animData->animFrameData, 0, animData->bytesPerFrame);

    if (animData->settings[animData->currSetting].style == aeAnimStyle_eSinglePass)
    {
        for (i = 0; i < animData->settings[animData->currSetting].runLength; i++)
        {
            index = animData->animFrameIndex - animData->settings[animData->currSetting].runLength + i;

            if (index >= 0 && index < animData->frameCount)
            {
                pixelIndex = aeAnimation_p_GetEntryValue(animData, index);

                if (animData->bytesPerFrame <= sizeof(int))
                    animData->animFrame[0] |= (1 << pixelIndex);
                else
                {
                    aeAnimation_p_CalcOffsetShiftMask(animData, pixelIndex, CALC_FOR_FRAME, &valueOffset, &valueShift, &valueMask);
                    animData->animFrameData[valueOffset] |= (valueMask << valueShift);
                }
            }
        }
    }
    else /* wrapped style */
    {

        for (i = 0; i < animData->settings[animData->currSetting].runLength; i++)
        {

            index = animData->animFrameIndex + i;

            if (animData->animRepeatIndex == 0)
                index -= animData->settings[animData->currSetting].runLength;

            if (animData->animRepeatIndex < animData->settings[animData->currSetting].repeat - 1 && index >= animData->frameCount)
                index -= animData->frameCount;

            if (index >= 0 && index < animData->frameCount)
            {
                pixelIndex = aeAnimation_p_GetEntryValue(animData, index);
                if (animData->bytesPerFrame <= sizeof(int))
                    animData->animFrame[0] |= (1 << pixelIndex);
                else
                {
                    aeAnimation_p_CalcOffsetShiftMask(animData, pixelIndex, CALC_FOR_FRAME, &valueOffset, &valueShift, &valueMask);
                    animData->animFrameData[valueOffset] |= (valueMask << valueShift);
                }
            }
        }
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static void aeAnimation_BuildFrame_Strip(aeAnimationData_t* animData)
{
    int	    i;
    int     index;
    int	    row;
    char    rotateApplied = TRUE;
    char    needsRotate;
    unsigned int     pixelIndex;
    unsigned int     value;
    unsigned int     valueOffset, valueShift, valueMask;
    unsigned int     scrollDirectionSize;
    unsigned int     nonScrollDirectionSize;

    AEASSERT(animData->layoutHeight > 0);
    AEASSERT(animData->layoutWidth  > 0);

    /* By definition, "vertical direction" is not rotated, and "horizontal direction" is rotated */
    needsRotate = (animData->settings[animData->currSetting].direction == aeAnimDirection_eHorizontal);
    scrollDirectionSize    = needsRotate ? animData->layoutWidth  : animData->layoutHeight;
    nonScrollDirectionSize = needsRotate ? animData->layoutHeight : animData->layoutWidth;

    if (animData->bytesPerFrame <= sizeof(int))
        animData->animFrame[0] = 0;
    else
        memset(animData->animFrameData, 0, animData->bytesPerFrame);

    for (i = 0; i < scrollDirectionSize; i++)
    {
        if (animData->settings[animData->currSetting].style == aeAnimStyle_eSinglePass)
            index = i + animData->animFrameIndex - scrollDirectionSize;
        else 
        {
            if (animData->animRepeatIndex > 0)
                index = i + animData->animFrameIndex;
            else
                index = i + animData->animFrameIndex - scrollDirectionSize;
        }

        if (animData->settings[animData->currSetting].style == aeAnimStyle_eWrapped && 
            index >= animData->frameCount && animData->animRepeatIndex < animData->settings[animData->currSetting].repeat - 1)
            index -= animData->frameCount;

        if (index >= 0 && index < animData->frameCount)
        {
            row = aeAnimation_p_GetEntryValue(animData, index);
            if (animData->bytesPerFrame <= sizeof(int))
            {
                animData->animFrame[0] = (animData->animFrame[0] | (row << (i * animData->layoutWidth)));
                /* If there was a rotate, we applied when entering the values */
                rotateApplied = FALSE;
            }
            else if (animData->bytesPerFrame == 8)
            {
                animData->animFrameData[scrollDirectionSize - 1 - i] = row;
                /* If there was a rotate, we applied when entering the values */
                rotateApplied = FALSE;
            }
            else
            {
                for (pixelIndex = 0; pixelIndex < nonScrollDirectionSize; pixelIndex++)
                {
                    value = aeAnimation_p_GetLedValueField(animData, index, pixelIndex);
                    if (value != 0)
                    {
                        int secondindex;

                        if (needsRotate)
                            secondindex = scrollDirectionSize - 1 - i + pixelIndex * scrollDirectionSize;
                        else
                            secondindex = pixelIndex + i * nonScrollDirectionSize;

                        if (secondindex > animData->layoutWidth * animData->layoutHeight)
                            secondindex = animData->layoutWidth * animData->layoutHeight;

                        aeAnimation_p_CalcOffsetShiftMask(animData, secondindex, CALC_FOR_FRAME, &valueOffset, &valueShift, &valueMask);
                        animData->animFrameData[valueOffset] |= (valueMask << valueShift);
                    }
                }

                /* If there was a rotate, we applied when entering the values */
                rotateApplied = TRUE;
            }
        }
    }

    if (needsRotate && !rotateApplied)
        aeAnimation_RotateFrame(animData);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_BuildFrame(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

	if (animData->subtype == aeAnimSubtype_eFrameByFrame)
        aeAnimation_BuildFrame_FrameByFrame(animData);

    else if (animData->subtype == aeAnimSubtype_ePixelList)
        aeAnimation_BuildFrame_PixelList(animData);

    else if (animData->subtype == aeAnimSubtype_eStrip)
        aeAnimation_BuildFrame_Strip(animData);

    else
	{
		AEASSERTED("unknown subtype");
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_RotateFrame(aeAnimationData_t* animData)
{
    	unsigned long	orig[2];
    	unsigned long	rotated[2];
        unsigned char	origData[ANIM_MAX_BYTES_PER_FRAME];
        unsigned char	rotatedData[ANIM_MAX_BYTES_PER_FRAME];
    	int				row, col;
    	int				bitpos;
    	int				bit;
        int             i;

    
        AEASSERT(animData->layoutHeight > 0);
        AEASSERT(animData->layoutWidth  > 0);
        AEASSERT(animData->layoutHeight == animData->layoutWidth);
        AEASSERT(animData->bitsPerPixel == 1);

        if (animData->bytesPerFrame <= sizeof(int))
        {
        	orig[0] = animData->animFrame[0];
        	rotated[0] = 0;
        
        	for (row = 0; row < animData->layoutHeight; row++)
        	{
        		for (col = 0; col < animData->layoutWidth; col++)
        		{
        			bitpos = row * animData->layoutWidth + col;
        
        			bit = (orig[bitpos/32] & (1 << (bitpos%32))) != 0;
        
        			/* Calculate new location for bit; note that col and row are reversed from above */
        			bitpos = col * animData->layoutWidth + (animData->layoutWidth - 1 - row);
        			if (bit)
        				rotated[bitpos/32] |= (1 << (bitpos%32));
        		}
        	}
        
        	animData->animFrame[0] = rotated[0];
        }
        else
        {
    
            for (i = 0; i < ANIM_MAX_BYTES_PER_FRAME; i++)
            {
                origData[i] = animData->animFrameData[i];
                rotatedData[i] = 0;
            }
        
        	for (row = 0; row < animData->layoutHeight; row++)
        	{
        		for (col = 0; col < animData->layoutWidth; col++)
        		{
        			bitpos = row * animData->layoutWidth + col;
        
        			bit = (origData[bitpos/8] & (1 << (bitpos%8))) != 0;
        
        			/* Calculate new location for bit; note that col and row are reversed from above */
        			bitpos = (animData->layoutWidth - 1 - col) * animData->layoutWidth + (row);
        			if (bit)
        				rotatedData[bitpos/8] |= (1 << (bitpos%8));
        
        		}
        
        	}
        
            for (i = 0; i < ANIM_MAX_BYTES_PER_FRAME; i++)
                animData->animFrameData[i] = rotatedData[i];

        }

    
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetSpeed(aeAnimationData_t* animData)
{
    return aeAnimation_GetSpeedAt(animData, animData->currSetting);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetSpeedAt(aeAnimationData_t* animData, int index)
{
    int speed = 0;

    AECHECKANIM(animData);

    speed = animData->settings[index].speed;

    return speed;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetRepeat(aeAnimationData_t* animData)
{
    return aeAnimation_GetRepeatAt(animData, animData->currSetting);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetRepeatAt(aeAnimationData_t* animData, int index)
{
	int repeat = 0;

    AECHECKANIM(animData);

    repeat = animData->settings[index].repeat;

	return repeat;
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeLedValue_t aeAnimation_Edit_GetLedValue(aeAnimationData_t* animData, int ledIndex)
{
	return aeAnimation_GetLedValue(animData, animData->editFrameIndex, ledIndex);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeLedValue_t aeAnimation_Anim_GetLedValue(aeAnimationData_t* animData, int ledIndex)
{
	unsigned long	frame;
	unsigned long	mask;
	aeLedValue_t	value;
    unsigned int    valueMask, valueShift, valueOffset;

	AECHECKANIM(animData);

    if (animData->bytesPerFrame <= sizeof(int) && animData->bitsPerPixel == 1)
    {
    	frame = animData->animFrame[0];
    
    	mask = (1 << ledIndex);
    
    	value = (frame & mask) ? 1 : 0;
    }
    else
    {
        aeAnimation_p_CalcOffsetShiftMask(animData, ledIndex, CALC_FOR_FRAME, &valueOffset, &valueShift, &valueMask);

        value = ((animData->animFrameData[valueOffset] >> valueShift) & valueMask);
    }

    /* If the data is hidden, then return 0 */
    if (animData->hideAnimFrame)
        value = 0;

	return value;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void* aeAnimation_p_GetEntryPtr(aeAnimationData_t* animData, int frameIndex)
{
    unsigned char*   entryPtr;

    AECHECKANIM(animData);
    AEASSERT(animData->entries);
    AEASSERT(frameIndex < animData->allocatedCount);
    AEASSERT(frameIndex < animData->frameCount);

    entryPtr = (unsigned char*)animData->entries;
    entryPtr += frameIndex * animData->bytesPerEntry;

    return (void*)entryPtr;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
unsigned int aeAnimation_p_GetEntryValue(aeAnimationData_t* animData, int frameIndex)
{
    unsigned int    result = 0;
    int             count;
    unsigned char*  p;

    AEASSERT(animData->bytesPerEntry <= sizeof(unsigned int));

    count = animData->bytesPerEntry;
    p = aeAnimation_p_GetEntryPtr(animData, frameIndex);

    do
    {
        result = ((result << 8) | *p);

        p++;
        count--;

    } while (count > 0);

    return result;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
unsigned int aeAnimation_p_SetEntryValue(aeAnimationData_t* animData, int frameIndex, unsigned int value)
{
    unsigned int    result = 0;
    int             count;
    unsigned char*  p;

    AECHECKANIM(animData);
    AEASSERT(animData->bytesPerEntry <= sizeof(unsigned int));
    AEASSERT(frameIndex < animData->allocatedCount);

    /* For safety, insure we are in-range in a non-debug build */
    if (frameIndex < animData->allocatedCount)
    {
        count = animData->bytesPerEntry;
    
        /* Point to last byte of entry and work backward */
        p = aeAnimation_p_GetEntryPtr(animData, frameIndex);
        p += animData->bytesPerEntry - 1;
    
        do
        {
            *p = (value & 0xff);
            value >>= 8;
    
            p--;
            count--;
    
        } while (count > 0);
    }

    return result;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeAnimation_p_ClearFrame(aeAnimationData_t* animData, int frameIndex)
{
    unsigned char* ptr;
    int            count;
    char           changed = FALSE;

    AECHECKANIM(animData);

    ptr = aeAnimation_p_GetEntryPtr(animData, frameIndex);
    count = animData->bytesPerEntry;

    do
    {
        if (*ptr != 0)
            changed = TRUE;

        *ptr++ = 0;
    } while (count-- > 0);

    return changed;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/

void aeAnimation_p_CalcOffsetShiftMask(aeAnimationData_t* animData, int ledIndex, char calcForEntry,
                                    unsigned int* pValueOffset, unsigned int* pValueShift, unsigned int* pValueMask)
{
    int             pixelsPerByte;
    int             bytesPerField;
    int             valuePosition;
    int             valueByteOffset;
    unsigned int    valueShift;
    unsigned int    valueMask;
    unsigned int    valueOffset;

    bytesPerField = calcForEntry ? animData->bytesPerEntry : animData->bytesPerFrame;

    /* The pixels per byte */
    pixelsPerByte = 8/animData->bitsPerPixel;
    AEASSERT(pixelsPerByte);

    /* The position with the byte to find the value */
    valuePosition   = ledIndex % pixelsPerByte;

    /* The byte within the entry */
    valueByteOffset = ledIndex / pixelsPerByte;

    /* This is a little weak, since we need to be one of two values, but this lets us use */
    /* this routine to calc the values when getting the animation frame data for for pixel */
    /*  list animations */
    AEASSERT(valueByteOffset < bytesPerField);

    /* The bit position within the byte */ 
    valueShift      = valuePosition * animData->bitsPerPixel;

    /* The mask for the value */
    valueMask       = (1 << animData->bitsPerPixel) - 1;

    /* Offset from the start. Low entries are least-significant byte */
    valueOffset = bytesPerField - 1 - valueByteOffset;
    AEASSERT(valueOffset < bytesPerField);

    *pValueShift  = valueShift;
    *pValueMask   = valueMask;
    *pValueOffset = valueOffset;
}

void aeAnimation_p_CalcPtrShiftMask(aeAnimationData_t* animData, int frameIndex, int ledIndex, 
                                    unsigned char** pptr, unsigned int* pValueShift, unsigned int* pValueMask)
{
    unsigned char*  ptr;
    unsigned int    valueOffset;

    aeAnimation_p_CalcOffsetShiftMask(animData, ledIndex, CALC_FOR_ENTRY, &valueOffset, pValueShift, pValueMask);

    /* The pointer to the entry */
    ptr = aeAnimation_p_GetEntryPtr(animData, frameIndex);

    /* The pointer to the byte */
    *pptr = ptr + valueOffset;

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeLedValue_t aeAnimation_p_GetLedValueField(aeAnimationData_t* animData, int frameIndex, int ledIndex)
{
    aeLedValue_t    value = 0;
    unsigned int    valueShift;
    unsigned int    valueMask;
    unsigned char*  ptr;

   aeAnimation_p_CalcPtrShiftMask(animData, frameIndex, ledIndex, &ptr, &valueShift, &valueMask);

    /* Get the byte, shift down, and mask off everything else */
    value = (*ptr >> valueShift) & valueMask;

    return value;

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeLedValue_t aeAnimation_p_SetLedValueField(aeAnimationData_t* animData, int frameIndex, int ledIndex, unsigned int value)
{
    unsigned int    valueShift;
    unsigned int    valueMask;
    unsigned char*  ptr;

    unsigned char   entryByte;

    aeAnimation_p_CalcPtrShiftMask(animData, frameIndex, ledIndex, &ptr, &valueShift, &valueMask);

    /* Mask and  shift the input value so that it is ready for placement in the byte */
    value = (value & valueMask) << valueShift;

    /* Get the current byte */
    entryByte = *ptr;

    /* Mask out the old value and replace with the new */
    *ptr = (entryByte & ~(valueMask << valueShift)) | value;

    /* printf("*ptr:0x%x (before: 0x%x), shift:%d, mask:0x%x, value:%d, ledIndex:%d, ptr:0x%x\n", 
                *ptr, entryByte, valueShift, valueMask, value, ledIndex, ptr); */

    return value;

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeLedValue_t aeAnimation_GetLedValue(aeAnimationData_t* animData, int frameIndex, int ledIndex)
{
	unsigned long	frame;
	unsigned long	mask;
	aeLedValue_t	value = 0;
    aeAnimSubtype_t subtype;


	AECHECKANIM(animData);

    /* Use the edit subtype for accessing the LED value */
    subtype = animData->editSubtype;

    if (subtype == aeAnimSubtype_eFrameByFrame)
	{
        if (animData->bytesPerEntry <= sizeof(int))
        {
            /* For Frame-by-frame animations, the frame contains one pixel per bit */
            frame = aeAnimation_p_GetEntryValue(animData, frameIndex);
            mask = (1 << ledIndex);

            value = (frame & mask) ? 1 : 0;
        }
        else
        {
            value = aeAnimation_p_GetLedValueField(animData, frameIndex, ledIndex);
        }
	}
	else if (subtype == aeAnimSubtype_ePixelList)
	{
		/* For Pixel list animations, the frame contains the LED index */
        frame = aeAnimation_p_GetEntryValue(animData, frameIndex);
		value = (frame == ledIndex) ? 1 : 0;
	}
	else if (subtype == aeAnimSubtype_eStrip)
	{
        AEASSERT(animData->layoutWidth > 0);
		/* For Strip animations, each 'frame' contains one row */
		/* We ignore the provided frameIndex and calculate this instead by the ledIndex */
		frameIndex = ledIndex / animData->layoutWidth;
        frame = aeAnimation_p_GetEntryValue(animData, frameIndex);

		mask = (1 << (ledIndex % animData->layoutWidth));

		value = (frame & mask) ? 1 : 0;
	}
	else
		AEASSERTED("Unknown subtype");

	return value;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetLedValue(aeAnimationData_t* animData, int frameIndex, int ledIndex, aeLedValue_t value)
{
	unsigned long	frame;
	unsigned long   mask;
    aeAnimSubtype_t subtype;

	AECHECKANIM(animData);

    subtype =animData->editSubtype;

	if (subtype == aeAnimSubtype_eFrameByFrame)
	{
        if (animData->bytesPerEntry <= sizeof(int))
        {
    		/* For Frame-by-frame animations, the frame contains one pixel per bit */
    		frame = aeAnimation_p_GetEntryValue(animData, frameIndex);
    	
    		mask = (1 << ledIndex);
    	
    		if (value)
    			frame = (frame | mask);
    		else
    			frame = (frame & ~mask);
    
    		aeAnimation_p_SetEntryValue(animData, frameIndex, frame);
        }
        else
        {
            aeAnimation_p_SetLedValueField(animData, frameIndex, ledIndex, value);
        }
	}
	else if (subtype == aeAnimSubtype_ePixelList)
	{
		/* For Pixel list animations, the frame contains the LED index */
		aeAnimation_p_SetEntryValue(animData, frameIndex, ledIndex);
	}
	else if (subtype == aeAnimSubtype_eStrip)
	{
        AEASSERT(animData->layoutWidth > 0);
		/* For Strip animations, each 'frame' contains one row */
		/* We ignore the provided frameIndex and calculate this instead by the ledIndex */
		frameIndex = ledIndex / animData->layoutWidth;
		mask = (1 << (ledIndex % animData->layoutWidth));

		frame = aeAnimation_p_GetEntryValue(animData, frameIndex);
	
		if (value)
			frame = (frame | mask);
		else
			frame = (frame & ~mask);

		aeAnimation_p_SetEntryValue(animData, frameIndex, frame);
	}
	else
		AEASSERTED("Unknown subtype");

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_Edit_UpdateLedValue(aeAnimationData_t* animData, int ledIndex, aeLedValue_t ledValue)
{
	aeAnimation_SetLedValue(animData, animData->editFrameIndex, ledIndex, ledValue);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_AppendFrame(aeAnimationData_t* animData)
{
    int newAllocatedCount;

	AECHECKANIM(animData);

    if (animData->frameCount == animData->allocatedCount)
    {
        newAllocatedCount = animData->allocatedCount + ANIM_ALLOCATE_INCREMENT;
        aeAnimation_p_IncreaseEntryAllocation(animData, newAllocatedCount);
    }

	if (animData->frameCount < animData->allocatedCount)
	{
		animData->frameCount++;
		animData->editFrameIndex = animData->frameCount - 1;
		aeAnimation_SetChangedFlag(animData);

		if (animData->needsResizeOnFrameCountChange)
			aeAnimation_SetFlag(animData, AEANIMFLAG_NeedsLayoutResize);
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetFrameData(aeAnimationData_t* animData, int frameIndex)
{
    int result;
    AECHECKANIM(animData);
    AEASSERT(frameIndex >= 0 && frameIndex < animData->frameCount);

    result = aeAnimation_p_GetEntryValue(animData, frameIndex);

    return result;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_GetFrameDataBuffer(aeAnimationData_t* animData, int frameIndex, unsigned char* pBuffer, int bufferLen)
{
    int copyLen;
    AECHECKANIM(animData);
    AEASSERT(frameIndex >= 0 && frameIndex < animData->frameCount);
    AEASSERT(bufferLen >= animData->bytesPerEntry);

    /* Don't over-run the buffer */
    copyLen = (bufferLen < animData->bytesPerEntry) ? bufferLen : animData->bytesPerEntry;

    /* Just to be safe */
    if (pBuffer != NULL)
        memcpy(pBuffer, aeAnimation_p_GetEntryPtr(animData,frameIndex), copyLen);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_DuplicateFrame(aeAnimationData_t* animData)
{
	int	 duplicatedIndex;
    int  i;
    int  newAllocatedCount;
    unsigned char*  ptr;
	
    AECHECKANIM(animData);
    AEASSERT(animData->editFrameIndex >= 0 && animData->editFrameIndex < animData->frameCount);

    /* Grow data, if necessary */
    if (animData->frameCount == animData->allocatedCount)
    {
        newAllocatedCount = animData->allocatedCount + ANIM_ALLOCATE_INCREMENT;
        aeAnimation_p_IncreaseEntryAllocation(animData, newAllocatedCount);
    }

    /* Only Allow duplicate if there is room */
	if (animData->frameCount < animData->allocatedCount)
	{
		duplicatedIndex = animData->editFrameIndex;
		
		/* Increase frame count */
		animData->frameCount++;

        ptr = aeAnimation_p_GetEntryPtr(animData, animData->frameCount - 2);

        
		for (i = animData->frameCount - 2; i >= duplicatedIndex; i--)
		{
            memcpy(ptr + animData->bytesPerEntry, ptr, animData->bytesPerEntry);

            ptr -= animData->bytesPerEntry;
		}
		
		
		/* Point to copied frame, not original */
		animData->editFrameIndex++;
		
		aeAnimation_SetChangedFlag(animData);

		/* Resize edit layout for strip animation */
		if (animData->needsResizeOnFrameCountChange)
			aeAnimation_SetFlag(animData, AEANIMFLAG_NeedsLayoutResize);
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ClearCurrFrame(aeAnimationData_t* animData)
{
    char            changed;

    AECHECKANIM(animData);
    AEASSERT(animData->editFrameIndex >= 0 && animData->editFrameIndex < animData->frameCount);

    changed = aeAnimation_p_ClearFrame(animData, animData->editFrameIndex);

    if (changed)
        aeAnimation_SetChangedFlag(animData);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_InvertCurrFrame(aeAnimationData_t* animData)
{
	int	index;
    int pixelIndex;
    unsigned int pixel;

    AECHECKANIM(animData);
    AEASSERT(animData->editFrameIndex >= 0 && animData->editFrameIndex < animData->frameCount);
	index = animData->editFrameIndex;

    /* For data edited as strip, then invert all */
    if (animData->editSubtype == aeAnimSubtype_eStrip)
    {
        for (index = 0; index < animData->frameCount; index++)
        {
            for (pixelIndex = 0; pixelIndex < animData->ledsPerEntry; pixelIndex++)
            {
                pixel = aeAnimation_p_GetLedValueField(animData, index, pixelIndex);

                pixel = ~pixel;

                aeAnimation_p_SetLedValueField(animData, index, pixelIndex, pixel);
            }

        }

        aeAnimation_SetChangedFlag(animData);
    }
    /* Otherwise, only invert frame-by-frame */
    else if (animData->subtype == aeAnimSubtype_eFrameByFrame)
    {
        for (pixelIndex = 0; pixelIndex < animData->ledsPerEntry; pixelIndex++)
        {
            pixel = aeAnimation_p_GetLedValueField(animData, index, pixelIndex);
    
            pixel = ~pixel;
    
            aeAnimation_p_SetLedValueField(animData, index, pixelIndex, pixel);
        }

        aeAnimation_SetChangedFlag(animData);
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_DeleteCurrFrame(aeAnimationData_t* animData)
{
	int	 deletedIndex;
    int  copyEntries;
    int  copySize;
    char changed;
    unsigned char*  ptr;

    AECHECKANIM(animData);

	/* Only Allow delete if more than one frame */
	if (animData->frameCount > 1)
	{
        changed = TRUE;

        deletedIndex = animData->editFrameIndex;
        ptr = aeAnimation_p_GetEntryPtr(animData, deletedIndex);

        /* Number of entries to copy */
        copyEntries = animData->frameCount - deletedIndex - 1;

        /* Number of bytes to copy */
        copySize = copyEntries * animData->bytesPerEntry;

        /* Copy the data 'down' */
        memcpy(ptr, ptr + animData->bytesPerEntry, copySize);

		/* Clear the old data so we don't see it again */
        aeAnimation_p_ClearFrame(animData, animData->frameCount - 1);

		/* Adjust frame count */
		animData->frameCount--;

		/* Keep editFrameIndex in range */
		if (animData->editFrameIndex >= animData->frameCount)
			animData->editFrameIndex = animData->frameCount-1;

		/* Keep animFrameIndex in range */
		if (animData->animFrameIndex >= animData->frameCount)
			aeAnimation_Reset(animData);

		/* Resize edit layout for strip animation */
		if (animData->needsResizeOnFrameCountChange)
			aeAnimation_SetFlag(animData, AEANIMFLAG_NeedsLayoutResize);
			
		/* Decrease runLength if necessary */
		if (animData->subtype == aeAnimSubtype_ePixelList && animData->settings[animData->currSetting].runLength > animData->frameCount)
			animData->settings[animData->currSetting].runLength = animData->frameCount;
		
	}
	/* If one frame, then clear it */
	else
	{
		changed = aeAnimation_p_ClearFrame(animData, 0);

	}

    if (changed)
        aeAnimation_SetChangedFlag(animData);

}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_UpdateCurrFrameIndex(aeAnimationData_t* animData, aePosition_t position)
{
	AECHECKANIM(animData);

	switch (position)
	{
		case aePosition_eFirst:
			animData->editFrameIndex = 0;
			break;

		case aePosition_ePrev:
			if (animData->editFrameIndex > 0)
				animData->editFrameIndex--;
			break;

		case aePosition_eCurrent:
			break;

		case aePosition_eNext:
			if (animData->editFrameIndex < animData->frameCount - 1)
				animData->editFrameIndex++;
			break;

		case aePosition_eLast:
			animData->editFrameIndex = animData->frameCount - 1;
			break;

		default:
			AEASSERTED("Unknown aePosition_t value");
			break;
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetFrameCount(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);
	return animData->frameCount;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetEditFrameIndex(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);
	return animData->editFrameIndex;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetAnimFrameIndex(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);
	return animData->animFrameIndex;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeAnimation_GetChangedFlag(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);

	return animData->changed;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetChangedFlag(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);

	animData->changed = TRUE;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ResetChangedFlag(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);

	animData->changed = FALSE;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ChangeSpeed(aeAnimationData_t* animData, aeChange_t change)
{
    int delta;
	int largedelta = 0x10;
    int smalldelta = 0x1;
	int speedmax   = 0xff;
	
    AECHECKANIM(animData);

	if (change == aeChange_eIncrease)
	{
        /* At the very low end, switch to a small delta to support values from 1..15 */
        delta = (animData->settings[animData->currSetting].speed - largedelta >= 0) ? largedelta : smalldelta;

		if (animData->settings[animData->currSetting].speed + delta < speedmax)
		{
			animData->settings[animData->currSetting].speed += delta;
		    aeAnimation_SetChangedFlag(animData);
		}
		
	}
	else if (change == aeChange_eDecrease)
	{
        /* At the very low end, switch to a small delta to support values from 1..15 */
        delta = (animData->settings[animData->currSetting].speed - largedelta > 0) ? largedelta : smalldelta;

        if (animData->settings[animData->currSetting].speed - delta > 0)
		{
			animData->settings[animData->currSetting].speed -= delta;
		    aeAnimation_SetChangedFlag(animData);
		}
	}
	
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ChangeRunLength(aeAnimationData_t* animData, aeChange_t change)
{
    AECHECKANIM(animData);

	if (change == aeChange_eIncrease)
	{
		if (animData->settings[animData->currSetting].runLength < animData->frameCount)
		{
			animData->settings[animData->currSetting].runLength++;
		    aeAnimation_SetChangedFlag(animData);
		}
		
	}
	else if (change == aeChange_eDecrease)
	{
		if (animData->settings[animData->currSetting].runLength > 1)
		{
			animData->settings[animData->currSetting].runLength--;
		    aeAnimation_SetChangedFlag(animData);
		}
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ChangeRepeat(aeAnimationData_t* animData, aeChange_t change)
{
	int delta = 1;
	int repeatmax = 0xff;

    AECHECKANIM(animData);

	if (change == aeChange_eIncrease)
	{
		if (animData->settings[animData->currSetting].repeat + 1 < repeatmax)
		{
			animData->settings[animData->currSetting].repeat += delta;
		    aeAnimation_SetChangedFlag(animData);
		}
		
	}
	else if (change == aeChange_eDecrease)
	{
		if (animData->settings[animData->currSetting].repeat - delta > 0)
		{
			animData->settings[animData->currSetting].repeat -= delta;
		    aeAnimation_SetChangedFlag(animData);
		}
	}
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimDirection_t aeAnimation_GetDirectionAt(aeAnimationData_t* animData, int index)
{
    AECHECKANIM(animData);

    return animData->settings[index].direction;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimDirection_t aeAnimation_GetDirection(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

    return aeAnimation_GetDirectionAt(animData, animData->currSetting);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ToggleDirection(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

  if (animData->settings[animData->currSetting].direction == aeAnimDirection_eHorizontal)
      animData->settings[animData->currSetting].direction = aeAnimDirection_eVertical;
  else
      animData->settings[animData->currSetting].direction = aeAnimDirection_eHorizontal;
      
}

void aeAnimation_SetDirection(aeAnimationData_t* animData, aeAnimDirection_t dir)
{
	animData->settings[animData->currSetting].direction = dir;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
const char* aeAnimation_GetName(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    return animData->name;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetName(aeAnimationData_t* animData, char* name)
{
    strncpy(animData->name, name, ANIM_NAME_LEN);
    animData->name[ANIM_NAME_LEN-1] = 0;

    aeAnimation_SetChangedFlag(animData);
}

/*----------------------------------------------------------------------
Summary
    Returns a modified version of the name

Desc
    
----------------------------------------------------------------------*/
char* aeAnimation_GetNameUnique(aeAnimationData_t* animData, char* nameStr, int nameStrLen)
{
    AECHECKANIM(animData);
    AEASSERT(nameStr && nameStrLen);

    strncpy(nameStr, animData->name, nameStrLen);

    /* Append the suffix if there is one */
    if (animData->uniqueSuffix[0] != 0)
        strncat(nameStr, animData->uniqueSuffix, nameStrLen);

    nameStr[nameStrLen-1] = 0;
    return nameStr;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static void aeAnimation_p_AppendVariant(aeAnimationData_t* animData, char* name, int namelen)
{
    char subname[5];

    subname[0] = 'A' + animData->currSetting;
    subname[1] = 0;

    if (animData->currSetting > 0 && (strlen(name) + strlen(subname)) < (namelen - 1))
        strcat(name, subname);

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char* aeAnimation_GetNameVariant(aeAnimationData_t* animData, char* name, int namelen)
{
    strcpy(name, animData->name);

    aeAnimation_p_AppendVariant(animData, name, namelen);

    return name;
}  

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char* aeAnimation_GetNameUniqueVariant(aeAnimationData_t* animData, char* name, int namelen)
{
    aeAnimation_GetNameUnique(animData, name, namelen);

    aeAnimation_p_AppendVariant(animData, name, namelen);

    return name;
}  

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetUniqueSuffix(aeAnimationData_t* animData, char* uniqueSuffixStr)
{
    AECHECKANIM(animData);
    AEASSERT(uniqueSuffixStr);
    AEASSERT(uniqueSuffixStr[0] != 0 && strlen(uniqueSuffixStr) < ANIM_NAME_UNIQUE_SUFFIX_LEN - 1);

    strncpy(animData->uniqueSuffix, uniqueSuffixStr, ANIM_NAME_UNIQUE_SUFFIX_LEN);
    animData->uniqueSuffix[ANIM_NAME_UNIQUE_SUFFIX_LEN - 1] = 0;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ClearUniqueSuffix(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

    animData->uniqueSuffix[0] = 0;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimSubtype_t aeAnimation_GetSubtype(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);

	return animData->subtype;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetSignature(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    return animData->signature;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetFlag(aeAnimationData_t* animData, int flag)
{
	AECHECKANIM(animData);

	animData->flags |= (flag);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ClearFlag(aeAnimationData_t* animData, int flag)
{
	AECHECKANIM(animData);

	animData->flags &= ~(flag);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetFlag(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);

	return animData->flags;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetWrapStyle(aeAnimationData_t* animData, aeAnimStyle_t style)
{
	AECHECKANIM(animData);
	animData->settings[animData->currSetting].style = style;
}

void aeAnimation_ToggleWrapStyle(aeAnimationData_t* animData)
{
	AECHECKANIM(animData);
	
	if (animData->settings[animData->currSetting].style == aeAnimStyle_eSinglePass)
		animData->settings[animData->currSetting].style = aeAnimStyle_eWrapped;
	else
		animData->settings[animData->currSetting].style = aeAnimStyle_eSinglePass;

    aeAnimation_SetChangedFlag(animData);

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimStyle_t aeAnimation_GetWrapStyle(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

    return aeAnimation_GetWrapStyleAt(animData, animData->currSetting);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimStyle_t aeAnimation_GetWrapStyleAt(aeAnimationData_t* animData, int index)
{
    AECHECKANIM(animData);
    AEASSERT(index >= 0 && index < animData->settingsCount);

    return animData->settings[index].style;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetRunLengthAt(aeAnimationData_t* animData, int index)
{
    return animData->settings[index].runLength;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetRunLength(aeAnimationData_t* animData)
{
    return aeAnimation_GetRunLengthAt(animData, animData->currSetting);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_UpdateVariantIndex(aeAnimationData_t* animData, aePosition_t position)
{
    AECHECKANIM(animData);

    if (position == aePosition_ePrev)
    {
        if (animData->currSetting > 0)
            animData->currSetting--;
    }
    else if (position == aePosition_eNext)
    {
        if (animData->currSetting < animData->settingsCount - 1)
            animData->currSetting++;

    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetVariantCount(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    return animData->settingsCount;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ChangeVariantCount(aeAnimationData_t* animData, aeChange_t change)
{
	int i;

    AECHECKANIM(animData);
    AEASSERT(animData->settingsCount > 0 && animData->settingsCount < ANIM_MAX_VARIANTS);
    AEASSERT(animData->currSetting >= 0 && animData->currSetting < animData->settingsCount);
    
    if (change == aeChange_eDecrease)
    {
    	/* Must leave at least one settings */
        if (animData->settingsCount > 1)
        {
        	/* Shift down all the settings above the current */
        	if (animData->currSetting < animData->settingsCount - 1)
        	{
        		for (i = animData->currSetting; i < animData->settingsCount - 1; i++)
        			animData->settings[i] = animData->settings[i + 1];
        	}
        	/* Otherwise current is the last, so decrement */
			else
                animData->currSetting--;
                
            /* Reduce the count */
            animData->settingsCount--;
        }
    }
    else if (change == aeChange_eIncrease)
    {
        if (animData->settingsCount < ANIM_MAX_VARIANTS)
        {
            /* Copy current settings to new settings */
            animData->settings[animData->settingsCount] = animData->settings[animData->currSetting];

            /* Increment */
            animData->settingsCount++;

            /* Set current to newly added */
            animData->currSetting = animData->settingsCount - 1;
        }

    }


    AEASSERT(animData->settingsCount > 0 && animData->settingsCount < ANIM_MAX_VARIANTS);
    AEASSERT(animData->currSetting >= 0 && animData->currSetting < animData->settingsCount);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_GetCurrentVariant(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

    return animData->currSetting;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetCurrentVariant(aeAnimationData_t* animData, int variant)
{
    AECHECKANIM(animData);
    AEASSERT(variant >= 0 && variant < animData->settingsCount);

    animData->currSetting = variant;
}

/*----------------------------------------------------------------------
Summary
    Set 'Needs Resize on Frame Count Change' flag

Desc
    Some animations want to change the edit layout as the frame
    counts changes. Setting this flag will notify the editor window
    when this is needed
----------------------------------------------------------------------*/
void aeAnimation_SetNeedsResizeOnFrameCountChange(aeAnimationData_t* animData, char flag)
{
    AECHECKANIM(animData);
    animData->needsResizeOnFrameCountChange = flag;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetFilePath(aeAnimationData_t* animData, char* path)
{
    AECHECKANIM(animData);
    AEASSERT(path);

    if (animData->pathptr == NULL)
        animData->pathptr = malloc(MAX_ANIM_PATH);
    
	if (animData->pathptr != NULL)
	{
		memset(animData->pathptr, 0, MAX_ANIM_PATH);
		strncpy(animData->pathptr, path, MAX_ANIM_PATH);
		animData->pathptr[MAX_ANIM_PATH - 1] = 0;
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeAnimation_HasFilePath(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    return animData->pathptr != NULL;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ClearFilePath(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    if (animData->pathptr != NULL)
    {
        free(animData->pathptr);
        animData->pathptr = NULL;
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_GetFilePath(aeAnimationData_t* animData, char* path, int pathlen)
{
	AEASSERT(path);
	AEASSERT(pathlen);
	
	if (animData->pathptr == NULL)
		*path = 0;
	else
	{
		strncpy(path, animData->pathptr, pathlen);
		path[pathlen - 1] = 0;
	}
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeAnimation_Read(aeAnimationData_t* animData, FILE* fileref, char* prefix, char* suffix)
{
    aErr_t  err = AERR_NO_ERROR;
    char    matched;
    char	line[500];
    char	fileline[500];

    animData->ioState = AnimIOState_eName;

    while (animData->ioState != AnimIOState_eDone)
    {
        err = aeIO_ReadLineFromFile(fileref, fileline, sizeof(fileline));
        matched = aeIO_RemovePrefixSuffix(fileline, prefix, suffix, line, sizeof(line));
        aeAnimation_IO_ReadLine(animData, line);
    }

    return err;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeAnimation_Write(aeAnimationData_t* animData, FILE* fileref, char* prefix, char* suffix)
{
    aErr_t  err = AERR_NO_ERROR;
    char    line[500];

    /* Create the first line: "ANIM:XXXX:YYYY" */
    aeArch_MakeSignatureStr(ASIG_ANIMATION, aeAnimation_GetSignature(animData), aeAnimation_GetSubtype(animData), line, sizeof(line));

    fprintf(fileref, "%s%s%s", prefix, line, suffix);

    /* Initial state */
    animData->ioState = AnimIOState_eName;

    do
    {
        aeAnimation_IO_WriteLine(animData, line, sizeof(line));
        fprintf(fileref, "%s%s%s", prefix, line, suffix);

    } while (animData->ioState != AnimIOState_eDone);

    return err;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeAnimation_Save(aeAnimationData_t* animData, char* filepath)
{
    aErr_t  err = AERR_NO_ERROR;
	FILE*	f;
    char    prefix[] = "; ";
    char    suffix[] = "\n";

	f = fopen(filepath, "w");

    if (f != NULL)
    {
        aeAnimation_Write(animData, f, prefix, suffix);

        fclose(f);
        f = NULL;

        aeAnimation_ResetChangedFlag(animData);
    }
    else
    {
        err = AERR_GENERAL_ERROR;
    }

    return err;
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeSequence_t* aeAnimation_GetSequence(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

    return animData->sequence;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_SetSequence(aeAnimationData_t* animData, aeSequence_t* sequence)
{
    AECHECKANIM(animData);
    AEASSERT(sequence);

    animData->sequence = sequence;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_ClearSequence(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);

    animData->sequence = NULL;
}




/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_IO_SetState(aeAnimationData_t* animData, AnimIOState_t iostate)
{
    AECHECKANIM(animData);
    AEASSERT(iostate >= 0 && iostate < AnimIOState_Count);
    animData->ioState = iostate;

    /* Always reset the ioCount */
    animData->ioCount = 0;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
AnimIOState_t aeAnimation_IO_GetState(aeAnimationData_t* animData)
{
    AECHECKANIM(animData);
    return animData->ioState;
}

void aeAnimation_IO_Count_Reset(aeAnimationData_t* animData)
{
    animData->ioCount = 0;
}

void aeAnimation_IO_Count_Incr(aeAnimationData_t* animData)
{
    animData->ioCount++;
}

int aeAnimation_IO_Count_Get(aeAnimationData_t* animData)
{
    return animData->ioCount;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_IO_ReadLine(aeAnimationData_t* animData, char* line)
{
    char	str[200];
    int     speed;
    int     count;
    int     repeat;
    int		style;
    int		direction;
    int		runlength;
    int		dirVector;
    int		transStyle;
    int		transDuration;
    int		settingsCount;
    int     reserved0;
    int     reserved1;
    int     scanCount = 0;
    int     i;
    aErr_t  err = AERR_NO_ERROR;
    unsigned int frame[6];

   
    /* Name */
    if (animData->ioState == AnimIOState_eName)
    {
        sscanf(line, "%s", str);
        strncpy(animData->name, str, sizeof(animData->name));
        animData->name[ANIM_NAME_LEN-1] = 0;

        animData->ioState = AnimIOState_eDetails;

        /* Probably unnecessary initialization */
        animData->animFrameIndex = 0;
        animData->editFrameIndex = 0;
    }

    /* speed, repeat, and count */
    else if (animData->ioState == AnimIOState_eDetails)
    {
        /* Determine if it is new-style */
    	style          = aeAnimStyle_eSinglePass;
    	direction      = aeAnimDirection_eHorizontal;
    	settingsCount  = 0;
    	runlength      = 1;
    	
        /* Frame-by-frame     */
        /*   Speed            */
        /*   Repeat           */
        /*   Frame count      */
 		if (animData->subtype == aeAnimSubtype_eFrameByFrame)
        	scanCount = sscanf(line, "%d %d %d", &speed, &repeat, &count);

        /* Strip              */
        /*   Speed            */
        /*   Repeat           */
        /*   Frame count      */
        /*   Style            */
        /*   Direction        */
        /*   Settings count   */
        else if (animData->subtype == aeAnimSubtype_eStrip)
	        scanCount = sscanf(line, "%d %d %d %d %d %d", &speed, &repeat, &count, &style, &direction, &settingsCount);

	    /* Pixel list         */
		/*   Speed            */
		/*   Repeat           */
		/*   Frame count      */
		/*   Style            */
		/*   Run length       */
		/*   Settings count   */
        else if (animData->subtype == aeAnimSubtype_ePixelList)
	        scanCount = sscanf(line, "%d %d %d %d %d %d", &speed, &repeat, &count, &style, &runlength, &settingsCount);
	    else
	    	AEASSERTED("aeAnimation_IO_ReadLine: Unknown subtype");


        /* Is it new style, where this line contains only settings count and frame count? */
        if (scanCount == 2)
        {
            sscanf(line, "%d %d", &count, &settingsCount);
            animData->frameCount = count;
            animData->settingsCount = settingsCount;

            animData->ioState = AnimIOState_eVariants;
        }
        /* Original version, including v0.1 frame-by-frame */
        else
        {
            animData->frameCount = count;
    
            animData->settings[0].speed          = speed;
            animData->settings[0].repeat         = repeat;
            animData->settings[0].style          = style;
            animData->settings[0].direction      = direction;
            animData->settings[0].runLength      = runlength;
            
            animData->settingsCount = 1;
    
   	        animData->ioState = AnimIOState_eFrames;
        }

        animData->ioCount = 0;

    }

    /* Variant */
    else if (animData->ioState == AnimIOState_eVariants)
    {
        /* Defaults */
        speed     = 0x80;
        repeat    = 1;
        runlength = 1;
        style     = aeAnimStyle_eSinglePass;
        direction = aeAnimDirection_eHorizontal;
        dirVector = 0x00;
        transStyle    = aeAnimTransition_eImmediate;
        transDuration = 0;
        reserved0     = 0;
        reserved0     = 1;


        /* Settings           */
        /*   Speed            */
        /*   Repeat           */
        /*   Run length       */
        /*   Style            */
        /*   Direction        */
        /*   Vector           */
        /*   Trans. style     */
        /*   Trans. duration  */
        /*   Reserved 0       */
        /*   Reserved 1       */
        sscanf(line, "%d %d %d %d %d %d %d %d %d %d", &speed, &repeat, &runlength, &style, &direction, &dirVector, &transStyle, 
               &transDuration, &reserved0, &reserved1);

        animData->settings[animData->ioCount].speed     = speed;
        animData->settings[animData->ioCount].repeat    = repeat; 
        animData->settings[animData->ioCount].runLength = runlength;
        animData->settings[animData->ioCount].style     = style;
        animData->settings[animData->ioCount].direction = direction;
        animData->settings[animData->ioCount].vector    = dirVector;
        animData->settings[animData->ioCount].transitionStyle    = transStyle;
        animData->settings[animData->ioCount].transitionDuration = transDuration;
        animData->settings[animData->ioCount].reserved0 = reserved0;
        animData->settings[animData->ioCount].reserved1 = reserved1;

	    if (animData->ioCount == animData->settingsCount - 1)
        {
     		animData->ioState = AnimIOState_eFrames;
            animData->ioCount = 0;
        }
        else
            animData->ioCount++;

    }

    /* Frames */
    else if (animData->ioState == AnimIOState_eFrames)
    {
        int nextSix = animData->ioCount + 6;
        /* Number of ints to read one entry */
        int valuesPerEntry = (animData->bytesPerEntry + sizeof(int) - 1)/sizeof(int);
        int entryIndex; /* Frame index */
        int valueIndex; /* Int (or smaller) within the frame */
        int valueWidth; /* Width of value within the frame */

        /* Insure we have enough space allocated */
        if (animData->frameCount > animData->allocatedCount)
        {
            err = aeAnimation_p_IncreaseEntryAllocation(animData, animData->frameCount);
            if (err != AERR_NO_ERROR)
                aeAnimation_p_SetError(animData, err);
        }

        sscanf(line, "%x %x %x %x %x %x", &frame[0], &frame[1], &frame[2], &frame[3], &frame[4], &frame[5]);

        /* Only save the frame data if we did not have an error */
        if (animData->err == AERR_NO_ERROR)
        {
            if (animData->bytesPerEntry <= sizeof(int))
            {
                AEASSERT(valuesPerEntry == 1);

                /* Use ioCount to track loaded frames */
                i = 0;
                while (animData->ioCount < animData->frameCount && animData->ioCount < nextSix)
                    aeAnimation_p_SetEntryValue(animData, animData->ioCount++, frame[i++]);
            }
            else
            {
                i = 0;
                while (animData->ioCount < animData->frameCount * valuesPerEntry && animData->ioCount < nextSix)
                {
                    unsigned char* ptr;

                    /* Divide ioCount by the valuesPerEntry to determine the frame index */
                    entryIndex = animData->ioCount / valuesPerEntry;
                    AEASSERT(entryIndex < animData->frameCount);

                    /* Use the modulo to find which value in the frame */
                    valueIndex = animData->ioCount % valuesPerEntry;

                    /* Determine the width of the value to write. Should be int size or smaller */
                    valueWidth = (animData->bytesPerEntry - valueIndex * sizeof(int)) > sizeof(int) ? 
                                    sizeof(int) : (animData->bytesPerEntry - valueIndex * sizeof(int));
                    AEASSERT(valueWidth > 0);
                    ptr = aeAnimation_p_GetEntryPtr(animData, entryIndex);

                    ptr += valueIndex * sizeof(int);

                    if (valueWidth == 1)
                        *(unsigned char*)ptr = frame[i++];
                    else if (animData->bytesPerEntry == 2)
                        *(unsigned short*)ptr = frame[i++];
                    else
                        *(unsigned int*)ptr = frame[i++];

                    animData->ioCount++;
                }
            }
        }
        else
        {
            /* Increment the ioCount pointer without saving frame data */
            /* Note that we only read/write 6 entries to a line at present */
            if (animData->ioCount + 6 < animData->frameCount * valuesPerEntry)
                animData->ioCount += 6;
            else
                animData->ioCount = animData->frameCount * valuesPerEntry;
        }

        if (animData->ioCount >= animData->frameCount * valuesPerEntry)
        {
            animData->ioState = AnimIOState_eDone;
        }
    }
    else
    {
        /* ERROR */
        AEASSERTED("Unknown iostate in arch_IO_ReadLine");
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_IO_WriteLine(aeAnimationData_t* animData, char* line, int lineLen)
{
    char    datastr[20];
    int     frame;

    /* Name */
    if (animData->ioState == AnimIOState_eName)
    {
        strcpy(line, animData->name);

        animData->ioState = AnimIOState_eDetails;
    }
    /* Speed, repeat, and count */
    else if (animData->ioState == AnimIOState_eDetails)
    {
        /*   Frame count      */
        /*   Variant count    */
        sprintf(line, "%d %d", 
                animData->frameCount, 
                animData->settingsCount);

        animData->ioState = AnimIOState_eVariants;
        animData->ioCount = 0;
    }
    /* Settings Variants */
    else if (animData->ioState == AnimIOState_eVariants)
    {
        /* Settings           */
        /*   Speed            */
        /*   Repeat           */
        /*   Run length       */
        /*   Style            */
        /*   Direction        */
        /*   Vector           */
        /*   Trans. style     */
        /*   Trans. duration  */
        /*   Reserved 0       */
        /*   Reserved 1       */

        /* Insure that the name is not empty */
        sprintf(line, "%d %d %d %d %d %d %d %d %d %d",  
                animData->settings[animData->ioCount].speed, 
                animData->settings[animData->ioCount].repeat, 
                animData->settings[animData->ioCount].runLength, 
                animData->settings[animData->ioCount].style,
                animData->settings[animData->ioCount].direction,
                animData->settings[animData->ioCount].vector,
                animData->settings[animData->ioCount].transitionStyle,
                animData->settings[animData->ioCount].transitionDuration,
                animData->settings[animData->ioCount].reserved0,
                animData->settings[animData->ioCount].reserved1);

	    if (animData->ioCount == animData->settingsCount - 1)
        {
     		animData->ioState = AnimIOState_eFrames;
            animData->ioCount = 0;
        }
        else
            animData->ioCount++;

    }
    /* Frames */
    else if (animData->ioState == AnimIOState_eFrames)
    {
        int nextSix = animData->ioCount + 6;
        /* Number of ints to write one entry */
        int valuesPerEntry = (animData->bytesPerEntry + sizeof(int) - 1)/sizeof(int);
        int entryIndex; /* Frame index */
        int valueIndex; /* Int (or smaller) within the frame */
        int valueWidth; /* Width of value within the frame */

        line[0] = 0;
        while (animData->ioCount < animData->frameCount * valuesPerEntry && animData->ioCount < nextSix)
        {
            /* Divide ioCount by the valuesPerEntry to determine the frame index */
            entryIndex = animData->ioCount / valuesPerEntry;
            AEASSERT(entryIndex < animData->frameCount);

            /* Use the modulo to find which value in the frame */
            valueIndex = animData->ioCount % valuesPerEntry;

            /* Determine the width of the value to write. Should be int size or smaller */
            valueWidth = (animData->bytesPerEntry - valueIndex * sizeof(int)) > sizeof(int) ? 
                            sizeof(int) : (animData->bytesPerEntry - valueIndex * sizeof(int));
            AEASSERT(valueWidth > 0);

            if (animData->bytesPerEntry <= sizeof(int))
            {
                frame = aeAnimation_p_GetEntryValue(animData, entryIndex);

                if (animData->bytesPerEntry == 1)
                    sprintf(datastr, " %2.2x", frame);
                else if (animData->bytesPerEntry == 2)
                    sprintf(datastr, " %4.4x", frame);
                else
                    sprintf(datastr, " %8.8x", frame);

                strcat(line, datastr);
            }
            else
            {
                unsigned char* ptr;
                ptr = aeAnimation_p_GetEntryPtr(animData, entryIndex);

                ptr += valueIndex * sizeof(int);

                if (valueWidth == 1)
                    sprintf(datastr, " %2.2x", *(unsigned char*)ptr);
                else if (animData->bytesPerEntry == 2)
                    sprintf(datastr, " %4.4x", *(unsigned short*)ptr);
                else
                    sprintf(datastr, " %8.8x", *(unsigned int*)ptr);

                strcat(line, datastr);
            }

            animData->ioCount++;
        }

        if (animData->ioCount >= animData->frameCount * valuesPerEntry)
        {
            animData->ioState = AnimIOState_eDone;
            AEASSERT(animData->ioCount == animData->frameCount * valuesPerEntry);
        }
    }
    else if (animData->ioState == AnimIOState_eCodeFileHeader      || 
             animData->ioState == AnimIOState_eCodeDispatchHeader  || 
             animData->ioState == AnimIOState_eCodeDispatchEntry   || 
             animData->ioState == AnimIOState_eCodeDispatchTrailer || 
             animData->ioState == AnimIOState_eCodeRoutineHeader   || 
             animData->ioState == AnimIOState_eCodeRoutineBody     ||
             animData->ioState == AnimIOState_eCodeRoutineTrailer  || 
             animData->ioState == AnimIOState_eCodeFileTrailer) 
    {
        {
            AEASSERTED("Architecture not provided for code write states");
        }
    }
    else
    {
        /* ERROR */
        AEASSERTED("Unknown iostate in arch_IO_WriteLine");
    }
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeAnimation_p_SetError(aeAnimationData_t* animData, aErr_t err)
{
    AECHECKANIM(animData);

    animData->err = err;
}


#if 0
void aeAnimation_Dbg_Log(aeAnimationData_t* animData)
{
    int value;
    int index;
    unsigned char* ptr;

    AEASSERT(animData);

    printf("Bytes per entry:%d\n", animData->bytesPerEntry);

    for (index = 0; index < animData->frameCount; index++)
    {
        value = aeAnimation_p_GetEntryValue(animData, index);

        printf("%12d %8x\n", value, value);
    }
    printf("....\n");

    index = animData->frameCount * animData->bytesPerEntry;
    ptr = animData->entries;

    while (index > 0)
    {
        printf("%2.2x ", *ptr);
        ptr++;
        index--;

    }

    printf("\n");

    printf("---\n");

}
#endif

