/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aearch.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aearch.h"
#include "aedebug.h"
#include "asig.h"
#include "aeanimation.h"
#include "aearchlist.h"
#include <stdio.h>

#define ANIMINFO_SAFETY 0x1d7733eb

#define AECHECKAI(_ai_) AEASSERT(_ai_ && _ai_->safety == ANIMINFO_SAFETY)

int aeArch_p_AnimSubtypeToIndex(aeAnimSubtype_t subtype)
{
    int result = -1;

    switch (subtype)
    {
        case aeAnimSubtype_eFrameByFrame:
            result = aeAnimSubtypeFlagBit_eFrameByFrame;
            break;
        case aeAnimSubtype_ePixelList:
            result = aeAnimSubtypeFlagBit_ePixelList;
            break;
        case aeAnimSubtype_eTile:
            result = aeAnimSubtypeFlagBit_eTile;
            break;
        case aeAnimSubtype_eStrip:
            result = aeAnimSubtypeFlagBit_eStrip;
            break;
        case aeAnimSubtype_eCommand:
            result = aeAnimSubtypeFlagBit_eCommand;
            break;
        case aeAnimSubtype_eFunction:
            result = aeAnimSubtypeFlagBit_eFunction;
            break;
        default:
            AEASSERTED("Unknown subtype");
            result = 0;
            break;
    }

    return result;
}

static aeAnimationInfo_t* aeArch_p_Create(aSignature_t layoutSignature, aeAnimSubtype_t subtype, aeAnimationData_t* animDataParam, char createAnimData)
{

    aeAnimationInfo_t*    animInfo = NULL;
    aeAnimationData_t*    animData = NULL;
    aeLayout_t*           editLayout = NULL;
    aeLayout_t*           animLayout = NULL;

    aeAnimationDataSettings_t   animSettings;


    int layoutWidth;
    int layoutHeight;
    int subtypeIndex;


    editLayout = aeLayout_Create(0);
    animLayout = aeLayout_Create(0);


    if (editLayout != NULL && animLayout != NULL)
    {
        animInfo = (aeAnimationInfo_t*)malloc(sizeof(aeAnimationInfo_t));
        if (animInfo != NULL)
        {
            memset(animInfo, 0, sizeof(aeAnimationInfo_t));
            animInfo->safety = ANIMINFO_SAFETY;
    
            /* Get Platform Info */
            aeArchList_Layout_GetBySignature(layoutSignature, &animInfo->platform);
        
            /* Get Architecture Info */
            aeArchList_Arch_GetBySignature(animInfo->platform.archSignature, &animInfo->archInfo);
        

            /* Set the layout size */
            aeLayout_SetLayoutDimensions(editLayout, animInfo->platform.layoutHeight, animInfo->platform.layoutWidth);
            aeLayout_SetLayoutDimensions(animLayout, animInfo->platform.layoutHeight, animInfo->platform.layoutWidth);

            /* Set layout element type */
            aeLayout_SetLayoutElementShape(editLayout, animInfo->platform.elementShape);
            aeLayout_SetLayoutElementShape(animLayout, animInfo->platform.elementShape);

            /* If the data was not provided, then allocate it */
            if (animDataParam == NULL && createAnimData)
            {
                subtypeIndex = aeArch_p_AnimSubtypeToIndex(subtype);

                /* Get configuration for the layout */
                layoutWidth  = animInfo->platform.layoutWidth;
                layoutHeight = animInfo->platform.layoutHeight;

                aeAnimation_InitDataSettings(&animSettings);
                animSettings.signature    = layoutSignature;
                animSettings.subtype      = subtype;
                animSettings.layoutWidth  = layoutWidth;
                animSettings.layoutHeight = layoutHeight;

                animSettings.bitsPerPixel  = animInfo->archInfo.bitsPerPixel;
                animSettings.bytesPerFrame = animInfo->archInfo.bytesPerFrame;
                animSettings.ledsPerEntry  = animInfo->archInfo.ledsPerEntry[subtypeIndex]; 
                animSettings.bytesPerEntry = animInfo->archInfo.bytesPerEntry[subtypeIndex];

                /* Create Animation Data */
                animData = aeAnimation_Create(&animSettings);
        
                /* Init the data */
                if (animData != NULL && animInfo->platform.initAnimData != NULL)
                    animInfo->platform.initAnimData(animData);
            }
            /* Else use the provided data */
            else
            {
                animData = animDataParam;
            }
    
            /* If we have data, then set it into the parent structure and create the layouts */
            /* We also allow no data for 'animation display only' usage */
            if (animData != NULL || !createAnimData)
            {
                animInfo->animData   = animData;
                animInfo->editLayout = editLayout;
                animInfo->animLayout = animLayout;

                /* Create Editing Layout, but only if we have animation data */
                /* We don't create the editing layout when we are using this */
                /* just to show an animation (e.g., in the sequence window) */
                if (animInfo->animData != NULL)
                    animInfo->platform.layoutCreateEdit(animInfo->editLayout, animInfo->animData);
    
                /* Create Animated Layout */
                animInfo->platform.layoutCreateAnim(animInfo->animLayout);
            }
            /* No animation data? Then clean-up */
            else
            {
                free(animInfo);
                animInfo = NULL;
    
            }
        }
    }

    if (animInfo == NULL)
    {
        if (editLayout != NULL)
        {
            aeLayout_Destroy(editLayout);
            editLayout = NULL;
        }

        if (animLayout != NULL)
        {
            aeLayout_Destroy(animLayout);
            animLayout = NULL;
        }

    }

    return animInfo;
}

aeAnimationInfo_t* aeArch_CreateWithSignatureSubtype(aSignature_t layoutSignature, aeAnimSubtype_t subtype)
{

	return aeArch_p_Create(layoutSignature, subtype, NULL, TRUE);
}

#if 0
aeAnimationInfo_t* aeArch_CreateWithNoData(aSignature_t layoutSignature)
{

	return aeArch_p_Create(layoutSignature, aeAnimSubtype_eFrameByFrame, NULL, FALSE);
}
#endif

aeAnimationInfo_t* aeArch_CreateWithData(aeAnimationData_t* animData)
{
    aSignature_t    layoutSignature;
    aeAnimSubtype_t subtype;

    layoutSignature = aeAnimation_GetSignature(animData);
    subtype         = aeAnimation_GetSubtype(animData);

	return aeArch_p_Create(layoutSignature, subtype, animData, FALSE);
}

void aeArch_Destroy(aeAnimationInfo_t* animInfo)
{
	AECHECKAI(animInfo);

    if (animInfo->animData != NULL)
    {
        aeAnimation_Delete(animInfo->animData);
        animInfo->animData = NULL;
    }

    if (animInfo->editLayout != NULL)
    {
        aeLayout_Destroy(animInfo->editLayout);
        animInfo->editLayout = NULL;
    }

    if (animInfo->animLayout != NULL)
    {
        aeLayout_Destroy(animInfo->animLayout);
        animInfo->animLayout = NULL;
    }

	free(animInfo);

}


aSignature_t aeArch_StrToSig(char* str, int len)
{
	aSignature_t sig = 0;
	int			 i = 0;
	
	while (i <= 4 && i < len)
	{
		sig = (sig << 8) | str[i];
		i++;
	}
	
	return sig;
}


void aeArch_SigToStr(aSignature_t sig, char* str, int len)
{
    if (len >= 4)
    {
        str[0] = (sig >> 24) & 0xff;
        str[1] = (sig >> 16) & 0xff;
        str[2] = (sig >>  8) & 0xff;
        str[3] = (sig      ) & 0xff;
    }
    if (len >= 5)
        str[4] = 0;
}



aeColor_t aeArch_EditData_GetElementColor(aeAnimationInfo_t* animInfo, int index)
{
	aeColor_t color = AECOLOR_UNKNOWN;

	AECHECKAI(animInfo);
    AEASSERT(animInfo->platform.dataGetElementColor);

	if (animInfo != NULL && animInfo->platform.dataGetElementColor != NULL)
		color = (animInfo->platform.dataGetElementColor)(animInfo->animData, index);

	return color;
}

void aeArch_EditData_AdvanceElementColor(aeAnimationInfo_t* animInfo, int index)
{
	AECHECKAI(animInfo);
    AEASSERT(animInfo->platform.dataAdvanceElementColor);

    if (animInfo != NULL && animInfo->platform.dataAdvanceElementColor != NULL)
        (animInfo->platform.dataAdvanceElementColor)(animInfo->animData, index);
}


aeColor_t aeArch_AnimData_GetElementColor(aeAnimationInfo_t* animInfo, int index)
{
	aeColor_t color = AECOLOR_UNKNOWN;

	AECHECKAI(animInfo);
	if (animInfo != NULL && animInfo->platform.animGetElementColor != NULL)
		color = (animInfo->platform.animGetElementColor)(animInfo->animData, index);

	return color;
}

int aeArch_AnimData_GetElementColorList(aeAnimationInfo_t* animInfo, aeColor_t* colorList, int colorListSize)
{
    int colorCount = 0;

	AECHECKAI(animInfo);
	if (animInfo != NULL && animInfo->platform.animGetElementColorList != NULL)
		colorCount = (animInfo->platform.animGetElementColorList)(animInfo->animData, colorList, colorListSize);

	return colorCount;
}


char* aeArch_AnimSubtype_GetStr(aeAnimSubtype_t subtype, char* str, int len)
{
	switch (subtype)
	{
		case aeAnimSubtype_eFrameByFrame:
			strncpy(str, "Frame by frame", len);
			break;
		case aeAnimSubtype_ePixelList:
			strncpy(str, "Pixel list", len);
			break;
		case aeAnimSubtype_eTile:
			strncpy(str, "Tile", len);
			break;
		case aeAnimSubtype_eStrip:
			strncpy(str, "Strip", len);
			break;
		case aeAnimSubtype_eCommand:
			strncpy(str, "Command", len);
			break;
		case aeAnimSubtype_eFunction:
			strncpy(str, "Function", len);
			break;

		default:
			AEASSERTED("Unknown subtype");
			strncpy(str, "<subtype?>", len);
		break;
	}

	str[len-1] = 0;

    return str;
}


char* aeArch_AnimStyle_GetStr(aeAnimStyle_t style, char* str, int len)
{
	switch (style)
	{
		case aeAnimStyle_eSinglePass:
			strncpy(str, "Single" /*"Single pass"*/, len);
			break;
		case aeAnimStyle_eWrapped:
			strncpy(str, "Wrapped", len);
			break;
		default:
			AEASSERTED("unknown style");
			strncpy(str, "<style?>", len);
		break;
	}

	str[len-1] = 0;
	
	return str;

}

aErr_t aeArch_ParseSignatureLine(char* line, aSignature_t* signature, aeAnimSubtype_t* subtype)
{
    aErr_t          err = AERR_NO_ERROR;
    int             count;
    aSignature_t    appSig;
    aSignature_t    animSig;
    aeAnimSubtype_t stSig;
    char            appSigStr[10];
    char            animSigStr[10];
    char            stSigStr[10];

    count = sscanf(line, "%4s:%4s:%4s", appSigStr, animSigStr, stSigStr);

    if (count >= 3)
    {
        appSig  = aeArch_StrToSig(appSigStr,  strlen(appSigStr));
        animSig = aeArch_StrToSig(animSigStr, strlen(animSigStr));
        stSig   = aeArch_StrToSig(stSigStr,   strlen(stSigStr));

        if (appSig != ASIG_ANIMATION)
            err = AERR_GENERAL_ERROR;
        else
        {
        }
        *signature = animSig;
        *subtype   = stSig;
    }
    else if (count == 1)
    {
        /* Is it a version 0.1 file? */
        animSig = aeArch_StrToSig(appSigStr, strlen(appSigStr));

        if (animSig == ASIG_FIVESQUARE)
        {
            *signature = animSig;
            *subtype   = aeAnimSubtype_eFrameByFrame;
        }
        else
        {
            err = AERR_GENERAL_ERROR;
        }
    }
    else
    {
        err = AERR_GENERAL_ERROR;
    }

    return err;
}



void aeArch_MakeSignatureStr(aSignature_t first, aSignature_t second, aSignature_t third, char* line, int linelen)
{
    char    signature[10];

    /* Create the signature line: "AAAA:XXXX:YYYY" */

    aeArch_SigToStr(first, signature, sizeof(signature));
    strcpy(line, signature);

    aeArch_SigToStr(second, signature, sizeof(signature));
    strcat(line,  ":");
    strcat(line,  signature);

    aeArch_SigToStr(third, signature, sizeof(signature));
    strcat(line,  ":");
    strcat(line,  signature);

}



