/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aearch.h
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#ifndef AEARCH_H_
#define AEARCH_H_
#include "astd.h"
#include "aetypes.h"
#include "aelayout.h"

#include <stdio.h>


/*-------------------
Function typedefs
    
-------------------*/

/* I/O */
typedef void (*aeArch_IO_ReadLine_t)(aeAnimationData_t* animData, char* line);
typedef void (*aeArch_IO_WriteLine_t)(aeAnimationData_t* animData, char* line, int lineLen);

/* Layout */
typedef void             (*aeArch_Layout_CreateAnim_t)(aeLayout_t* layout);
typedef void             (*aeArch_Layout_CreateEdit_t)(aeLayout_t* layout, aeAnimationData_t* animData);

typedef void             (*aeArch_Data_AdvanceElementColor_t)(aeAnimationData_t* animData, int elementIndex);
typedef aeColor_t        (*aeArch_Data_GetElementColor_t)(aeAnimationData_t* animData, int elementIndex);
typedef aeColor_t        (*aeArch_Animation_GetElementColor_t)(aeAnimationData_t* animData, int elementIndex);
typedef int              (*aeArch_Animation_GetElementColorList_t)(aeAnimationData_t* animData, aeColor_t* colorList, int colorListSize);

typedef void             (*aeArch_Data_SetElementValue_t)(aeAnimationData_t* animData, int elementIndex, aeElementValue_t);
typedef aeElementValue_t (*aeArch_Data_GetElementValue_t)(aeAnimationData_t* animData, int elementIndex);

/* Init Data */
typedef void         (*aeArch_InitData_t)(aeAnimationData_t* animData);

/*-------------------
Architecture Info
    Function pointers
-------------------*/
struct aeArchitectureInfo_tag
{
    aSignature_t        signature;                          /* Architecture signature, such as 'FiveSquare' or 'SLAK' or 'Crossbar' */
    aeAnimSubtypeFlag_t subtypeFlags;                       /* Subtypes supported by this architecture */
    int                 ledsPerEntry[AEANIMSUBTYPE_COUNT];
    int                 bytesPerEntry[AEANIMSUBTYPE_COUNT]; /* Bytes per entry needed for each subtype */
    int                 bytesPerFrame;                      /* Bytes for the entire frame, used by the animation to build the animated frame */
    int                 bitsPerPixel;                       /* Bits per LED value */
    char                name[ANIM_NAME_LEN];                /* Name */
    char                desc[ANIM_DESC_LEN];                /* Description */
};

/*-------------------
Architecture
    Function pointers
-------------------*/
struct aePlatform_tag
{
    aSignature_t                        layoutSignature;    /* Uniquely identifies layout; written to file */
    aSignature_t                        archSignature;      /* Identifies underlying architecture */

    aeAnimSubtypeFlag_t                 subtypeFlags;

    int                                 layoutWidth;
    int                                 layoutHeight;
    aeElementShape_t                    elementShape;

    int                                 codegenSpeedReduction;     /* Value to be passed from platform to codegen */

    char                                name[ANIM_NAME_LEN];
    char                                desc[ANIM_DESC_LEN];

    aeArch_InitData_t                   initAnimData;     /* Called when a new animation is created */
    aeArch_InitData_t                   initAnimDataRead; /* Called when an animation is read from a file */

    aeArch_Layout_CreateAnim_t          layoutCreateAnim;
    aeArch_Layout_CreateEdit_t          layoutCreateEdit;

    /* Routines to map led values to display colors */
    aeArch_Data_AdvanceElementColor_t   dataAdvanceElementColor;
    aeArch_Data_GetElementColor_t       dataGetElementColor;

    aeArch_Data_GetElementValue_t       dataGetElementValue;
    aeArch_Data_SetElementValue_t       dataSetElementValue;

    aeArch_Animation_GetElementColor_t  animGetElementColor;

    aeArch_Animation_GetElementColorList_t animGetElementColorList;
};

#if 0
struct aeArchitecture_tag
{
    aeArch_Layout_CreateAnim_t          layoutCreateAnim;
    aeArch_Layout_CreateEdit_t          layoutCreateEdit;

    /* Routine to write source code */
    aeArch_IO_WriteLine_t               ioWriteLine;

    /* Routines to map led values to display colors */
    aeArch_Data_UpdateElementColor_t    dataUpdateElementColor;
    aeArch_Data_AdvanceElementColor_t   dataAdvanceElementColor;
    aeArch_Data_GetElementColor_t       dataGetElementColor;
    aeArch_Animation_GetElementColor_t  animGetElementColor;

    #if 0
        aeArch_Layout_ResizeEdit_t          layoutResizeEdit;
        aeArch_Animation_GetLedValue_t      animGetLedValue;
        aeArch_Animation_Advance_t          animAdvance;
        aeArch_Animation_Reset_t            animReset;
    
        aeArch_IO_ReadLine_t                ioReadLine;
        aeArch_Data_AppendFrame_t           dataAppendFrame;
        aeArch_Data_DeleteCurrFrame_t       dataDeleteCurrFrame;
        aeArch_Data_DuplicateCurrFrame_t    dataDuplicateCurrFrame;
        aeArch_Data_ClearCurrFrame_t        dataClearCurrFrame;
        aeArch_Data_UpdateCurrFrameIndex_t  dataUpdateCurrFrameIndex;
        aeArch_Data_SetFrameSelection_t     dataSetFrameSelection;
        aeArch_Data_UpdateLedValue_t        dataUpdateLedValue;
        aeArch_Data_GetLedValue_t           dataGetLedValue;
    #endif
};
#endif


/*-------------------
Animation Info
    Complete information about an animation
    Includes architecture (list of function ptrs), data, and layouts
-------------------*/
typedef struct
{
    int                     safety;
    aePlatform_t            platform;
    aeArchitectureInfo_t    archInfo;
    aeAnimationData_t*      animData;
    aeLayout_t*             editLayout;
    aeLayout_t*             animLayout;
} aeAnimationInfo_t;


/*-------------------
Prototypes
    
-------------------*/

aeAnimationInfo_t*  aeArch_CreateWithSignatureSubtype(aSignature_t signature, aeAnimSubtype_t subtype);
aeAnimationInfo_t*  aeArch_CreateWithData(aeAnimationData_t* animData);

void                aeArch_Destroy(aeAnimationInfo_t* animInfo);


aeColor_t           aeArch_EditData_GetElementColor(aeAnimationInfo_t* animInfo, int index);
void                aeArch_EditData_AdvanceElementColor(aeAnimationInfo_t* animInfo, int index);
aeColor_t           aeArch_AnimData_GetElementColor(aeAnimationInfo_t* animInfo, int index);
int                 aeArch_AnimData_GetElementColorList(aeAnimationInfo_t* animInfo, aeColor_t* colorList, int colorListSize);


char*               aeArch_AnimSubtype_GetStr(aeAnimSubtype_t subtype, char* str, int len);
char*               aeArch_AnimStyle_GetStr(aeAnimStyle_t subtype, char* str, int len);

void                aeArch_SigToStr(aSignature_t sig, char* str, int strlen);
aSignature_t        aeArch_StrToSig(char* str, int strlen);

aErr_t              aeArch_ParseSignatureLine(char* line, aSignature_t* signature, aeAnimSubtype_t* subtype);


void                aeArch_MakeSignatureStr(aSignature_t first, aSignature_t second, aSignature_t third, char* line, int linelen);

int                 aeArch_p_AnimSubtypeToIndex(aeAnimSubtype_t subtype);

#endif





#if 0
    typedef void (*aeArch_Layout_ResizeEdit_t)(aeLayout_t* layout, aeChange_t change, aeOrientation_t orient); /* For strip only; increase or decrease; vert or horizontal */
    /* Data */
    typedef void (*aeArch_Data_AppendFrame_t)(aeAnimationData_t* animData);
    typedef void (*aeArch_Data_DeleteCurrFrame_t)(aeAnimationData_t* animData);
    typedef void (*aeArch_Data_DuplicateCurrFrame_t)(aeAnimationData_t* animData);
    typedef void (*aeArch_Data_ClearCurrFrame_t)(aeAnimationData_t* animData);
    
    typedef void (*aeArch_Data_UpdateCurrFrameIndex_t)(aeAnimationData_t* animData, aePosition_t position); /* First, Last, Next, Prev */
    typedef void (*aeArch_Data_SetFrameSelection_t)(aeAnimationData_t* animData); /* List? */
    typedef void         (*aeArch_Data_UpdateLedValue_t)(aeAnimationData_t* animData, int ledIndex, aeLedValue_t ledValue);
    typedef aeLedValue_t (*aeArch_Data_GetLedValue_t)(aeAnimationData_t* animData, int ledIndex);
    typedef aeLedValue_t (*aeArch_Animation_GetLedValue_t)(aeAnimationData_t* animData, int ledIndex);
    typedef void         (*aeArch_Animation_Advance_t)(aeAnimationData_t* animData);
    typedef void         (*aeArch_Animation_Reset_t)(aeAnimationData_t* animData);
#endif

