/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aearchlist.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aearchlist.h"
#include "asig.h"
#include "aedebug.h"


#define MAX_ITEMS 10

aeArchitectureInfo_t    sArchList[MAX_ITEMS];
int                     sArchListCount = 0;


aePlatform_t            sLayoutList[MAX_ITEMS];
int                     sLayoutListCount = 0;

aeCodeGen_t             sCodeGenList[MAX_ITEMS];
int                     sCodeGenListCount = 0;

static aePlatform_t* aeArchList_Layout_p_Find(aSignature_t layoutSignature);
static aeCodeGen_t*  aeArchList_CodeGen_p_Find(aSignature_t codegenSignature);



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeArchList_Arch_Add(aeArchitectureInfo_t* archInfo)
{
    int i;

    AEASSERT(sArchListCount < MAX_ITEMS - 1);

    /* Check the consistency of the archInfo */
    AEASSERT(archInfo->bitsPerPixel > 0);
    AEASSERT(archInfo->bytesPerFrame > 0);
    AEASSERT(archInfo->subtypeFlags != 0);

    /* Either we don't support a subtype or we specify the bytes per entry for the subtype */
    AEASSERT(((archInfo->subtypeFlags & aeAnimSubtypeFlag_eFrameByFrame) == 0) || archInfo->bytesPerEntry[aeAnimSubtypeFlagBit_eFrameByFrame] > 0 );
    AEASSERT(((archInfo->subtypeFlags & aeAnimSubtypeFlag_ePixelList   ) == 0) || archInfo->bytesPerEntry[aeAnimSubtypeFlagBit_ePixelList   ] > 0 );
    AEASSERT(((archInfo->subtypeFlags & aeAnimSubtypeFlag_eTile        ) == 0) || archInfo->bytesPerEntry[aeAnimSubtypeFlagBit_eTile        ] > 0 );
    AEASSERT(((archInfo->subtypeFlags & aeAnimSubtypeFlag_eStrip       ) == 0) || archInfo->bytesPerEntry[aeAnimSubtypeFlagBit_eStrip       ] > 0 );

    for (i = 0; i < sArchListCount; i++)
    {
        AEASSERT(archInfo->signature != sArchList[i].signature);
    }

    sArchList[sArchListCount] = *archInfo;
    sArchListCount++;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeArchList_Arch_GetCount(void)
{
    return sArchListCount;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aSignature_t aeArchList_Arch_GetSignature(int index)
{
    AEASSERT(index >= 0 && index < sArchListCount);
    return sArchList[index].signature;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char* aeArchList_Arch_GetName(int index, char* name, int namelen)
{
    AEASSERT(index >= 0 && index < sArchListCount);
    AEASSERT(name && namelen);

    strncpy(name, sArchList[index].name, namelen);
    name[namelen-1] = 0;

    return name;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeArchList_Arch_GetBySignature(aSignature_t archSignature, aeArchitectureInfo_t* archInfo)
{
    int index;
    int found = -1;

    for (index = 0; index < sArchListCount && found == -1; index++)
    {
        if (archSignature == sArchList[index].signature)
            found = index;
    }

    if (found != -1)
        *archInfo = sArchList[found];
    else
    {
        AEASSERT(0);
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static aePlatform_t* aeArchList_Layout_p_Find(aSignature_t layoutSignature)
{
    aePlatform_t* found = NULL;
    int index;

    for (index = 0; index < sLayoutListCount && found == NULL; index++)
    {
        if (layoutSignature == sLayoutList[index].layoutSignature)
            found = &sLayoutList[index];
    }

    return found;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeArchList_Layout_Add(aePlatform_t* platform)
{
    AEASSERT(sLayoutListCount < MAX_ITEMS - 1);
    AEASSERT(platform);
    AEASSERT(platform->archSignature && platform->layoutSignature && platform->name);
    AEASSERT(aeArchList_Layout_p_Find(platform->layoutSignature) == NULL);

    sLayoutList[sLayoutListCount] = *platform;
    sLayoutListCount++;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeArchList_Layout_GetCount(void)
{
    return sLayoutListCount;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char* aeArchList_Layout_GetName(int index, char* name, int namelen)
{
    AEASSERT(index >= 0 && index < sLayoutListCount);
    AEASSERT(name && namelen);

    strncpy(name, sLayoutList[index].name, namelen);
    name[namelen-1] = 0;

    return name;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aSignature_t aeArchList_Layout_GetSignature(int index)
{
    AEASSERT(index >= 0 && index < sLayoutListCount);

    return sLayoutList[index].layoutSignature;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aSignature_t aeArchList_Layout_GetArchSignatureFromLayoutSignature(aSignature_t layoutSignature)
{
    aePlatform_t* platform;

    platform = aeArchList_Layout_p_Find(layoutSignature);

    AEASSERT(platform);

    return platform->archSignature;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimSubtypeFlag_t aeArchList_Layout_GetSubtypeFlags(int index)
{
    AEASSERT(index >= 0 && index < sLayoutListCount);

    return sLayoutList[index].subtypeFlags;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/



void aeArchList_Layout_GetBySignature(aSignature_t layoutSignature, aePlatform_t* platform)
{
    aePlatform_t* found;

    found = aeArchList_Layout_p_Find(layoutSignature);

    if (found)
        *platform = *found;
    else
        AEASSERT(0);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeArchList_Layout_GetConfig(aSignature_t layoutSignature, int* layoutWidth, int* layoutHeight)
{
    aePlatform_t* platform;

    platform = aeArchList_Layout_p_Find(layoutSignature);

    if (platform != NULL)
    {
        *layoutWidth  = platform->layoutWidth;
        *layoutHeight = platform->layoutHeight;
    }
    else
        AEASSERT(0);
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeCodeGen_t* aeArchList_CodeGen_p_Find(aSignature_t codegenSignature)
{
    aeCodeGen_t* found = NULL;
    int index;

    for (index = 0; index < sCodeGenListCount && found == NULL; index++)
    {
        if (codegenSignature == sCodeGenList[index].codegenSignature)
            found = &sCodeGenList[index];
    }

    return found;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeArchList_CodeGen_Add(aeCodeGen_t* codegen)
{
    AEASSERT(sCodeGenListCount < MAX_ITEMS - 1);
    AEASSERT(aeArchList_CodeGen_p_Find(codegen->codegenSignature) == NULL);

    sCodeGenList[sCodeGenListCount] = *codegen;
    sCodeGenListCount++;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeArchList_CodeGen_GetBySignature(aSignature_t codegenSignature, aeCodeGen_t* codegen)
{
    aeCodeGen_t* found;

    found = aeArchList_CodeGen_p_Find(codegenSignature);

    if (found)
        *codegen = *found;
    else
        AEASSERT(0);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeArchList_CodeGen_GetCountByArch(aSignature_t archSignature)
{
    int count = 0;
    int index;

    for (index = 0; index < sCodeGenListCount; index++)
    {
        if (sCodeGenList[index].archSignature == archSignature)
            count++;
    }

    return count;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aSignature_t aeArchList_CodeGen_GetCodeGenSignatureByArchIndex(aSignature_t archSignature, int archIndex)
{
    int count = 0;
    int index;
    aSignature_t codegenSig = ASIG_UNKNOWN;

    for (index = 0; index < sCodeGenListCount && codegenSig == ASIG_UNKNOWN; index++)
    {
        if (sCodeGenList[index].archSignature == archSignature)
        {
            if (count == archIndex)
                codegenSig = sCodeGenList[index].codegenSignature;

            count++;
        }
    }

    return codegenSig;
}

