/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aegfxfile.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aegfxfile.h"
#include "aedebug.h"
#include "asig.h"
#include "aeanimation.h"
#include "aefileio.h"
#include "aearchlist.h"
#include "gd.h"

static aErr_t aeGfxFile_InitGifImages(int gifWidth, int gifHeight, 
                               gdImagePtr* imgRender, gdImagePtr* imgReference, 
                               aeColor_t* colorList, int* indexList, int colorCount);

static void   aeGfxFile_DrawFrame(gdImagePtr imgRender, aeAnimationInfo_t* animInfo, 
                                  aeColor_t* colorList, int* indexList, int colorCount,
                                  int gifWidth, int gifHeight);


#define ANIM_GFXFILE_MAX_GIF_COLORS 256
#define ANIM_GFXFILE_GIF_BORDER      5



aErr_t aeGfxFile_ExportAnimAsGif(aeAnimationData_t* animData, char* filepath)
{
    aErr_t  err = AERR_NO_ERROR;
    FILE*   giffile = 0;
    int     width;
    int     height;
    int     gifWidth;
    int     gifHeight;
    char    done = FALSE;
    int     speed;
    int     colorCount;
    char    first = TRUE;

    aeColor_t           colorList[ANIM_GFXFILE_MAX_GIF_COLORS];
    int                 indexList[ANIM_GFXFILE_MAX_GIF_COLORS];
    aeRect_t            bounds;
    gdImagePtr          imgRender;
    gdImagePtr          imgReference;
    aeAnimationData_t*  blankAnimData;
    aeAnimationInfo_t*  animInfo;

    animInfo = aeArch_CreateWithSignatureSubtype(aeAnimation_GetSignature(animData), aeAnimSubtype_eFrameByFrame);


    bounds = aeLayout_GetLayoutExtent(animInfo->animLayout);

    width  = bounds.right  - bounds.left;
    height = bounds.bottom - bounds.top;

    gifWidth  = width  + 2 * ANIM_GFXFILE_GIF_BORDER;
    gifHeight = height + 2 * ANIM_GFXFILE_GIF_BORDER;

    aeLayout_CalcBounds(animInfo->animLayout, &bounds);

    /* Keep the original around */
    blankAnimData = animInfo->animData;

    animInfo->animData = animData;

    speed = aeAnimation_GetSpeed(animData);
    speed = (speed+9)/10;
    if (speed == 0)
        speed = 1;

    /* Open a file for writing. "wb" means "write binary". */
    giffile = fopen(filepath, "wb");

    if (giffile != 0)
    {
        colorCount = aeArch_AnimData_GetElementColorList(animInfo, colorList, ANIM_GFXFILE_MAX_GIF_COLORS);

        /* add transparent color */
        colorList[colorCount] = 0x010203;
        colorCount++;

        err = aeGfxFile_InitGifImages(gifWidth, gifHeight, &imgRender, &imgReference, colorList, indexList, colorCount);

        if (err == AERR_NO_ERROR)
        {
            /* Set-up reference */
            gdImageFilledRectangle(imgRender,    0, 0, gifWidth - 1 , gifHeight - 1, indexList[0]);
            gdImageFilledRectangle(imgReference, 0, 0, gifWidth - 1 , gifHeight - 1, indexList[0]);

            /* This is not correct -- double-check the docs for how to create animated gifs */
            gdImageColorTransparent(imgRender, indexList[colorCount-1]);
            //gdImageColorTransparent(imgReference, indexList[colorCount-1]);

            gdImageGifAnimBegin(imgRender, giffile, 1 /* global color map */, 0 /* loop forever */);

            aeAnimation_Reset(animData);

            do
            {
                gdImageFilledRectangle(imgRender, 0, 0, gifWidth - 1, gifHeight - 1, indexList[0]);

                aeGfxFile_DrawFrame(imgRender, animInfo, colorList, indexList, colorCount, gifWidth, gifHeight);


                gdImageGifAnimAdd(imgRender, giffile, 0 /* local palette */, 0, 0 /* x, y Offset */,  speed, gdDisposalNone, first?NULL:imgReference);

                /* Copy current to ref */
                gdImageCopy(imgReference, imgRender, 0, 0, 0, 0, gifWidth - 1 , gifHeight - 1);

                aeAnimation_Advance(animData);

                done = aeAnimation_IsAtInitialAnimationState(animData);

                if (first)
                    gdImageColorTransparent(imgRender, indexList[colorCount-1]);
                first = FALSE;
            }
            while (!done);

            gdImageGifAnimEnd(giffile);

            /* Destroy the image in memory. */
            gdImageDestroy(imgRender);
            imgRender = 0;
            gdImageDestroy(imgReference);
            imgReference = 0;
        }
        else
        {
            err = AERR_GIF_IMAGE_CREATE_FAILED;
        }

        /* Close the file */
        fclose(giffile);
        giffile = 0;
    }
    else
    {
        err = AERR_FILE_CREATE_FAILED;
    }

    animInfo->animData = blankAnimData;



    return err;
}

aErr_t aeGfxFile_ExportSequenceAsGif(aeSequence_t* sequence, char* filepath)
{
    aErr_t  err = AERR_NO_ERROR;
    FILE*   giffile = 0;
    int     width;
    int     height;
    char    done = FALSE;
    int     speed;
    int     colorCount;
    int     gifWidth;
    int     gifHeight;

    aeColor_t           colorList[ANIM_GFXFILE_MAX_GIF_COLORS];
    int                 indexList[ANIM_GFXFILE_MAX_GIF_COLORS];
    aeRect_t            bounds;
    gdImagePtr          imgRender;
    gdImagePtr          imgReference;
    aeAnimationData_t*  blankAnimData;
    aeAnimationInfo_t*  animInfo;

    animInfo = aeArch_CreateWithSignatureSubtype(aeSequence_GetSignature(sequence), aeAnimSubtype_eFrameByFrame);


    bounds = aeLayout_GetLayoutExtent(animInfo->animLayout);

    width  = bounds.right  - bounds.left;
    height = bounds.bottom - bounds.top;

    gifWidth  = width  + 2 * ANIM_GFXFILE_GIF_BORDER;
    gifHeight = height + 2 * ANIM_GFXFILE_GIF_BORDER;

    aeLayout_CalcBounds(animInfo->animLayout, &bounds);

    /* Keep the original around */
    blankAnimData = animInfo->animData;


    colorCount = aeGfxFile_InitGifImages(width+1, height+1, &imgRender, &imgReference, colorList, indexList, ANIM_GFXFILE_MAX_GIF_COLORS);

    if (err == AERR_NO_ERROR)
    {

        colorCount = aeArch_AnimData_GetElementColorList(animInfo, colorList, ANIM_GFXFILE_MAX_GIF_COLORS);


        /* Open a file for writing. "wb" means "write binary". */
        giffile = fopen(filepath, "wb");

        if (giffile != 0)
        {
            /* Set-up reference */
            gdImageFilledRectangle(imgRender,    0, 0, width - 1 , height - 1, indexList[0]);
            gdImageFilledRectangle(imgReference, 0, 0, width - 1 , height - 1, indexList[0]);

            gdImageGifAnimBegin(imgRender, giffile, 1 /* global color map */, 0 /* loop forever */);

            aeSequence_Animated_Reset(sequence);

            do
            {
                gdImageFilledRectangle(imgRender, 0, 0, width - 1, height - 1, indexList[0]);

                animInfo->animData = aeSequence_Animated_GetData(sequence);
                aeGfxFile_DrawFrame(imgRender, animInfo, colorList, indexList, colorCount, gifWidth, gifHeight);

                speed = aeSequence_Animated_GetSpeed(sequence);
                speed = (speed+9)/10;
                if (speed == 0)
                    speed = 1;

                gdImageGifAnimAdd(imgRender, giffile, 0 /* local palette */, 0, 0 /* x, y Offset */,  speed, gdDisposalNone, imgReference);

                /* Copy current to ref */
                gdImageCopy(imgReference, imgRender, 0, 0, 0, 0, width - 1 , height - 1);

                aeSequence_Animated_Advance(sequence);
 
                done = aeSequence_Animated_SeqIsAtInitialAnimationState(sequence);
            }
            while (!done);

            gdImageGifAnimEnd(giffile);

            /* Close the file */
            fclose(giffile);
            giffile = 0;


        }
        else
        {
            err = AERR_FILE_CREATE_FAILED;
        }


        /* Destroy the image in memory. */
        gdImageDestroy(imgRender);
        imgRender = 0;
        gdImageDestroy(imgReference);
        imgReference = 0;
    }
    else
    {
        err = AERR_GIF_IMAGE_CREATE_FAILED;
    }


    return err;
}

aErr_t aeGfxFile_InitGifImages(int gifWidth, int gifHeight, 
                               gdImagePtr* imgRender, gdImagePtr* imgReference, 
                               aeColor_t* colorList, int* indexList, int colorCount)
{ 
    aErr_t      err = AERR_NO_ERROR;
    gdImagePtr  im = 0;
    gdImagePtr  imref = 0;
    int         i;
    int         red, green, blue;

    /* Allocate the image: 64 pixels across by 64 pixels tall */
    im = gdImageCreate(gifWidth, gifHeight);

    imref = gdImageCreate(gifWidth, gifHeight);


    if (im != NULL && imref != NULL)
    {
        /* Allocate the background, off, and on colors. */
        /* The first color in a new image will be the background color. */
        for (i = 0; i < colorCount; i++)
        {
            red   = (colorList[i] >> 16) & 0xff;
            green = (colorList[i] >>  8) & 0xff;
            blue  = (colorList[i]      ) & 0xff;

            indexList[i] = gdImageColorAllocate(im, red, green, blue);  
        }

        gdImagePaletteCopy(imref, im);

        *imgRender = im;
        *imgReference = imref;

    }
    else
    {
        AEDBG("error in aeGfxFile_InitGifImages");
        err = AERR_GENERAL_ERROR;
    }
    return err;
}



void aeGfxFile_DrawFrame(gdImagePtr imgRender, aeAnimationInfo_t* animInfo, 
                         aeColor_t* colorList, int* indexList, int colorCount, 
                         int gifWidth, int gifHeight)
{
    int         index;
    aeLayout_t*	layout = NULL;
    aeColor_t	elementColor;
    int         colorIndex;
    int         i;
    int         cX, cY;
    int         radiusX, radiusY;
    int         top, left, right, bottom;

    aeElementShape_t 	shape; 
    aeLayoutElement_t*	element;

    AEASSERT(animInfo);
    AEASSERT(animInfo->animData);
    AEASSERT(animInfo->animLayout);

    layout = animInfo->animLayout;


    element = aeLayout_GetElement(layout, 0);

    for (index = 0; index < aeLayout_GetElementCount(layout); index++)
    {
        element = aeLayout_GetElement(layout, index);

        /* Does the layout have a shape (which overrides the element shape?) */
        shape = aeLayout_GetLayoutElementShape(layout);

        /* If the layout shape is not specified, get the shape from the element */
        if (shape == aeElementShape_eUnknown)
            shape = aeLayout_GetElementShape(element->elementType);

        elementColor = aeArch_AnimData_GetElementColor(animInfo, index);

        colorIndex = indexList[0];

        for (i = 0; i < colorCount; i++)
            if (elementColor == colorList[i])
            {
                colorIndex = indexList[i];
                break;
            }

        if (shape == aeElementShape_eCircle)
        {

            cX = (element->bounds.left  + element->bounds.right )/2 + ANIM_GFXFILE_GIF_BORDER;
            cY = (element->bounds.top   + element->bounds.bottom)/2 + ANIM_GFXFILE_GIF_BORDER;

            radiusX = element->bounds.right  - element->bounds.left - 1;
            radiusY = element->bounds.bottom - element->bounds.top  - 1;

            if (cX - radiusX/2 >= 0 && cX + radiusX/2 < gifWidth && cY - radiusY/2 >= 0 && cY + radiusY/2 < gifHeight)
                gdImageFilledEllipse(imgRender, cX, cY, radiusX, radiusY, colorIndex);
        }
        else
        {
            left    = element->bounds.left       + ANIM_GFXFILE_GIF_BORDER;
            right   = element->bounds.right  - 1 + ANIM_GFXFILE_GIF_BORDER;
            top     = element->bounds.top        + ANIM_GFXFILE_GIF_BORDER;
            bottom  = element->bounds.bottom - 1 + ANIM_GFXFILE_GIF_BORDER;

            if (left >= 0 && left < gifWidth && right >= 0 && right <= gifWidth &&
                top >= 0 && top < gifHeight && bottom >= 0 && bottom < gifHeight &&
                left < right && top < bottom)
            {
                gdImageFilledRectangle(imgRender, left, top, right, bottom, colorIndex);
            }
        }
    }

}
