/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aelayout.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aelayout.h"
#include "aedebug.h"


#define AELAYOUT_SAFETY 0xeffe321e

#define AECHECKLAYOUT(_layout_) AEASSERT(_layout_ != NULL && _layout_->safety == AELAYOUT_SAFETY && _layout_->allocatedCount >= _layout_->elementCount)

static aErr_t aeLayout_p_IncreaseAllocatedElementCount(aeLayout_t* layout, int newElementCount);
 
/*-------------------
Layout Data
    Complete information about an animation
    Includes architecture (list of function ptrs), data, and layouts
-------------------*/
struct aeLayout_tag
{
    int                 safety;
    int                 elementCount;
    int                 allocatedCount;
	aeRect_t			layoutExtent;
    aeLayoutElement_t*  elements;

    /* Optional */
    int			            elementRowCount;        /* Count of items, vertical & horizontal */
    int			            elementColCount;
    aeElementType_t         elementShape;       /* Shape: circle, rectangle, etc */
    aeElementColor_t        elementOnColor;     /* Color info */
    aeElementColor_t        elementOffColor;
    aeElementColor_t        backgroundColor;
    int                     vertBorder;         /* Sizing and Spacing */
    int                     horzBorder;
    int                     vertSpacing;
    int                     horzSpacing;
    int                     vertSize;
    int                     horzSize;
    aeLayoutMatrixOrigin_t  matrixOrigin;       /* Starting corner and row-first or col-first layout */
    aeLayoutMatrixOrder_t   matrixOrder;

};

#if 0
void aeLayout_SetElementColors();
void aeLayout_SetBackgroundColor();
void aeLayout_SetElementBorder();
void aeLayout_SetElementSpacing();
void aeLayout_SetElementSize();
void aeLayout_SetMatrixOriginAndOrder();
#endif

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeLayout_t* aeLayout_Create(int elementCount)
{
    aErr_t              err;
    aeLayout_t*         layout;

    layout = (aeLayout_t*)malloc(sizeof(struct aeLayout_tag));

    if (layout != NULL)
    {
        memset(layout, 0, sizeof(struct aeLayout_tag));

        layout->safety = AELAYOUT_SAFETY;

        /* Initialize to empty */
        layout->elementCount   = 0;
        layout->allocatedCount = 0;
        layout->elements       = NULL;

        layout->elementShape   = aeElementShape_eUnknown;

        /* Allocate the elements */
        if (elementCount > 0)
            err = aeLayout_p_IncreaseAllocatedElementCount(layout, elementCount);
        else
            err = AERR_NO_ERROR;

        /* Did the element allocation succeed? */
        if (err != AERR_NO_ERROR)
        {
            free(layout);
            layout = NULL;
        }
        else
        {
            layout->elementCount = elementCount;
        }
    }


    return layout;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeLayout_Destroy(aeLayout_t* layout)
{
    AECHECKLAYOUT(layout);

    if (layout->elements != NULL)
    {
        free(layout->elements);
        layout->elements = NULL;
    }

    memset(layout, 0xee, sizeof(struct aeLayout_tag));
    free(layout);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeLayout_p_IncreaseAllocatedElementCount(aeLayout_t* layout, int newAllocatedCount)
{
    aErr_t              err = AERR_NO_ERROR;
    aeLayoutElement_t*  elements;
    int                 elementsSize;

    AECHECKLAYOUT(layout);
    AEASSERT(newAllocatedCount > layout->allocatedCount);
    
    elementsSize = newAllocatedCount * sizeof(aeLayoutElement_t);

    elements = (aeLayoutElement_t*)malloc(elementsSize);

    if (elements != NULL)
    {
        /* Clear the new block */
        memset(elements, 0, elementsSize);

        if (layout->elementCount > 0)
        {
            AEASSERT(layout->elements);
            /* Copy the old to the new */
            memcpy(elements, layout->elements, layout->elementCount * sizeof(aeLayoutElement_t));
        }

        if (layout->elements != NULL)
        {
            /* Free the old */
            free(layout->elements);
            layout->elements = NULL;
        }

        /* Update the allocated number */
        layout->allocatedCount = newAllocatedCount;

        /* Assign the new */
        layout->elements = elements;
    }
    else
    {
        err = AERR_OUT_OF_MEMORY;
    }

    return err;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeLayout_SetElementCount(aeLayout_t* layout, int newElementCount)
{
    char err = AERR_NO_ERROR;

    AECHECKLAYOUT(layout);
    AEASSERT(newElementCount >= 0);

    if (newElementCount > layout->allocatedCount)
    {
        err = aeLayout_p_IncreaseAllocatedElementCount(layout, newElementCount);
    }

    if (err == AERR_NO_ERROR)
    {
        layout->elementCount = newElementCount;
    }

    AECHECKLAYOUT(layout);
    return err;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeLayout_GetElementCount(aeLayout_t* layout)
{
    AECHECKLAYOUT(layout);
    return layout->elementCount;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeLayoutElement_t* aeLayout_GetElement(aeLayout_t* layout, int index)
{
    AECHECKLAYOUT(layout);
    AEASSERT(index >= 0 && index < layout->elementCount);

    return layout->elements + index;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeLayout_GetElementSize(aeElementType_t elementType, int* width, int* height)
{
	*width  = 5;
	*height = 5;

	switch (elementType)
	{
		case aeElementType_e3mmRound:
			*width  = 3;
			*height = 3;
			break;

		case aeElementType_e10mmRound:
			*width  = 10;
			*height = 10;
			break;

        default:
            *width  = 3;
            *height = 3;
            break;
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeLayout_ScaleElement(aeLayoutElement_t* layout, aeRect_t* displayExtent)
{
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static void aeRect_ScaleRect(aeRect_t* srcRect, aeRect_t* dstRect, aeRect_t* scaleRect, char constrainSquare)
{
	int		numWidth;
	int		denWidth;
	int		numHeight;
	int		denHeight;
	float	ratioWidth;
	float	ratioHeight;

	numWidth  = dstRect->right  - dstRect->left;
	denWidth  = srcRect->right  - srcRect->left;

	numHeight = dstRect->bottom - dstRect->top;
	denHeight = srcRect->bottom - srcRect->top;

	ratioWidth  = numWidth /(float)denWidth;
	ratioHeight = numHeight/(float)denHeight;

	if (constrainSquare)
	{
		if (ratioWidth < ratioHeight)
		{
			numHeight = numWidth;
			denHeight = denWidth;
		}
		else
		{
			numWidth = numHeight;
			denWidth = denHeight;
		}
	}

	scaleRect->left   = dstRect->left + (scaleRect->left   * numWidth )/denWidth;
	scaleRect->right  = dstRect->left + (scaleRect->right  * numWidth )/denWidth;
	scaleRect->top    = dstRect->top  + (scaleRect->top    * numHeight)/denHeight;
	scaleRect->bottom = dstRect->top  + (scaleRect->bottom * numHeight)/denHeight;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeLayout_CalcBounds(aeLayout_t* layout, aeRect_t* displayExtent)
{
	int			index;
	aeRect_t	lr;
	aeRect_t	dr;
	aeRect_t	er;
	int			locX, locY;
	int			width, height;

    aeLayoutElement_t* element;

    AECHECKLAYOUT(layout);

	lr = layout->layoutExtent;
	dr = *displayExtent;

	for (index = 0; index < layout->elementCount; index++)
	{
        element = aeLayout_GetElement(layout, index);

		locX = element->locX;
		locY = element->locY;

		aeLayout_GetElementSize(element->elementType, &width, &height);

		er.left   = locX - width/2;
		er.right  = locX + width/2;
		er.top    = locY - height/2;
		er.bottom = locY + height/2;

		aeRect_ScaleRect(&lr, &dr, &er, TRUE);

		element->bounds = er;
	}
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeElementShape_t aeLayout_GetElementShape(aeElementType_t elementType)
{
	aeElementShape_t	shape;

	switch (elementType)
	{
		case aeElementType_eCustom:
			shape = aeElementType_eCustom;
			break;

		case aeElementType_e3mmRound:
		case aeElementType_e5mmRound:
		case aeElementType_e8mmRound:
		case aeElementType_e10mmRound:
		case aeElementType_e20mmRound:
			shape = aeElementShape_eCircle;
			break;

		default:
			AEASSERTED("unknown elementType");
			shape = aeElementShape_eCircle;
			break;
	}

	return shape;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeLayout_SetLayoutExtent(aeLayout_t* layout, aeRect_t extent)
{
    AECHECKLAYOUT(layout);
    AEASSERT(extent.left < extent.right && extent.top < extent.bottom);

    layout->layoutExtent = extent;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeRect_t aeLayout_GetLayoutExtent(aeLayout_t* layout)
{
    AECHECKLAYOUT(layout);
    return layout->layoutExtent;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    Used to pass layout size to aematrixlayoututils
----------------------------------------------------------------------*/
void aeLayout_SetLayoutDimensions(aeLayout_t* layout, int rowCount, int colCount)
{
    AECHECKLAYOUT(layout);

    layout->elementRowCount = rowCount;
    layout->elementColCount = colCount;
}

int aeLayout_GetLayoutDimensions_Rows(aeLayout_t* layout)
{
    return layout->elementRowCount;
}

int aeLayout_GetLayoutDimensions_Cols(aeLayout_t* layout)
{
    return layout->elementColCount;
}

void aeLayout_SetLayoutElementShape(aeLayout_t* layout, aeElementShape_t elementShape)
{
    AECHECKLAYOUT(layout);

    layout->elementShape = elementShape;
}

aeElementShape_t aeLayout_GetLayoutElementShape(aeLayout_t* layout)
{
    AECHECKLAYOUT(layout);

    return layout->elementShape;
}

