/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aeSequence.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aesequence.h"
#include "aedebug.h"
#include "asig.h"
#include "aeanimation.h"
#include "aefileio.h"
#include "aearchlist.h"

#define MAX_SEQ_ITEMS 150
#define AESEQ_MAX_ACTIVE_SETS 6




/*-------------------
Sequence Info
    Information about one entry in the sequence list
    For pixel list, could refer to another entry's data
-------------------*/
typedef struct
{
    aeAnimationData_t*  animData;       /* Pointer to animation data. Null indicates that varData refers to anuim data elsewhere in list */
    int                 variant;        /* Index into pixel list variants */
    aeAnimationData_t*  varData;        /* Pointer to identify original animation data elsewhere is list. */
    char                active;         /* Bitmask: [6:0] Animation is included in assembly list */
    aeAnimationData_t*  openedData;     /* Non-null when animation is opened in a window; used to find window */

} aeSequenceInfo_t;

/*-------------------
Sequence
    List of animations
-------------------*/
struct aeSequence_tag
{
    int                 safety;
    aSignature_t        signature;
    aSignature_t        codegenSignature;
    aeCodeGen_t         codegen;
    char                name[ANIM_NAME_LEN];
    char                path[MAX_ANIM_PATH];
    int                 count;
    char                changed;        /* Set if animations added, removed, or rearranged; animations also carry changed flag */
    int                 selected;       /* Index of selected item */
    int                 animated;       /* Index of currently animated item */
    int                 currActiveSet;
    int                 activeSetCount;
    aeSequenceInfo_t    seq[MAX_SEQ_ITEMS];
    

};



/*-------------------
Local prototypes
-------------------*/
static aeAnimationData_t*   aeSequence_p_GetAnimData(aeSequence_t* sequence, int index);

static void                 aeSequence_p_SetChangedFlag(aeSequence_t* sequence);
static void                 aeSequence_p_ActivateAll(aeSequence_t* sequence, int animIndex);
static char                 aeSequence_p_GetActive(aeSequence_t* sequence, int animIndex, int activeSet);
static void                 aeSequence_p_SetActive(aeSequence_t* sequence, int animIndex, int activeSet, char active);
static void                 aeSequence_p_Validate(aeSequence_t* sequence);

/*----------------------------------------------------------------------
    Constants
----------------------------------------------------------------------*/
#define SEQUENCE_SAFETY 0x1d7735ee

#define AECHECKSEQ(_seq_) AEASSERT(_seq_ && _seq_->safety == SEQUENCE_SAFETY)
#define AECHECKIDX(_seq_, _idx_) AEASSERT(_idx_ >= 0 && _idx_ < _seq_->count)


/*----------------------------------------------------------------------
    Location Variables
----------------------------------------------------------------------*/
static int sSequenceUntitledCount = 1;


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeSequence_t* aeSequence_Create(aSignature_t layoutSignature)
{
	aeSequence_t*   seq;
    aSignature_t    archSignature;
    int             codegenCount;
	
	seq = malloc(sizeof(aeSequence_t));
	
	if (seq != NULL)
	{
		memset(seq, 0, sizeof(aeSequence_t));
        seq->safety    = SEQUENCE_SAFETY;
        seq->signature = layoutSignature;

        /* Start with at least some name */
        strcpy(seq->name, SEQUENCE_UNTITLED_NAME);

        /* Always at least one active set */
        seq->activeSetCount = 1;

        /* Use the layout signature to get to get the architecture signature */
        archSignature = aeArchList_Layout_GetArchSignatureFromLayoutSignature(layoutSignature);

        /* Determine if there are any codegen */
        codegenCount = aeArchList_CodeGen_GetCountByArch(archSignature);

        /* If any, then take the first for now */
        seq->codegenSignature = aeArchList_CodeGen_GetCodeGenSignatureByArchIndex(archSignature, 0);
	}

	return seq;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Destroy(aeSequence_t* sequence)
{
    int             index;
	
    AECHECKSEQ(sequence);

    for (index = sequence->count - 1; index >= 0; index--)
    {
        AEASSERT(sequence->seq[index].openedData == NULL);

        if (sequence->seq[index].animData != NULL)
        {
            aeAnimation_Delete(sequence->seq[index].animData);
            sequence->seq[index].animData = NULL;
        }
    }

    memset(sequence, 0, sizeof(aeSequence_t));
    free(sequence);
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_p_Validate(aeSequence_t* sequence)
{
    int index;
    int j;
    aeAnimationData_t* adI;
    aeAnimationData_t* adJ;

    AECHECKSEQ(sequence);
    AEASSERT(sequence->count >= 0);

    if (sequence->count > 0)
        AECHECKIDX(sequence, sequence->selected);
    else
        AEASSERT(sequence->selected == 0);

    if (sequence->count > 0)
        AECHECKIDX(sequence, sequence->animated);
    else
        AEASSERT(sequence->animated == 0);

    /* Only one of animData or varData can have a value; use xor in the test */
    for (index = 0; index < sequence->count; index++)
    {
        AEASSERT(sequence->seq[index].animData != NULL || sequence->seq[index].varData != NULL);
        AEASSERT(sequence->seq[index].animData == NULL || sequence->seq[index].varData == NULL);
    }

    /* If varData is set, it must refer to valid animData in the list */
    /* DDK 2009.05.12: The animData must be located earlier than the varData entry */
    for (index = 0; index < sequence->count; index++)
        if (sequence->seq[index].varData != NULL)
        {
            for (j = 0; j < index ; j++) /* DDK 2009.05.12: Was 'j < sequence->count' */
                if (sequence->seq[j].animData == sequence->seq[index].varData)
                    break;
            AEASSERT(j < index); /* DDK 2009.05.12: Was 'j < sequence->count' */
        }

    /* Validated the openedData */
    for (index = 0; index < sequence->count; index++)
    {
        adI = sequence->seq[index].animData != NULL ? sequence->seq[index].animData : sequence->seq[index].varData;
        for (j = 0; j < sequence->count; j++)
        {
            adJ = sequence->seq[j].animData != NULL ? sequence->seq[j].animData : sequence->seq[j].varData;

            if (adI == adJ)
                AEASSERT(sequence->seq[j].openedData == sequence->seq[index].openedData);

        }
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_AssignUntitledName(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);
    /* Default name */
    sprintf(sequence->name, "%s%d", SEQUENCE_UNTITLED_NAME, sSequenceUntitledCount);
    sSequenceUntitledCount++;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static char aeSequence_FindAnimationUsingOrigRef(aeSequence_t* sequence, aeAnimationData_t* origData, int* pOpenedDataIndex, int* pVarDataIndex)
{
    int     index;
    int     foundOpenedDataIndex = -1;
    int     foundVarDataIndex = -1;

    /* Find the opened window in the list */
    for (index = 0; index < MAX_SEQ_ITEMS && foundOpenedDataIndex == -1; index++)
    {
        if (sequence->seq[index].openedData == origData)
            foundOpenedDataIndex = index;
    }

    /* Does the opened item have animData? If so, then that is where the data is stored */
    if (sequence->seq[foundOpenedDataIndex].animData != NULL)
    {
        foundVarDataIndex = foundOpenedDataIndex;
	}
    /* Else find the index where the data is stored */
    else
    {
        AEASSERT(sequence->seq[foundOpenedDataIndex].varData);

        for (index = 0; index < MAX_SEQ_ITEMS && foundVarDataIndex == -1; index++)
        {
            if (sequence->seq[index].animData == sequence->seq[foundOpenedDataIndex].varData)
                foundVarDataIndex = index;
        }
    }

    if (pOpenedDataIndex != NULL && foundOpenedDataIndex != -1)
        *pOpenedDataIndex = foundOpenedDataIndex;

    if (pVarDataIndex != NULL && foundVarDataIndex != -1)
        *pVarDataIndex = foundVarDataIndex;


    return (foundOpenedDataIndex != -1 && foundVarDataIndex != -1);
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static char aeSequence_p_EntryNotInUse(aeSequence_t* sequence, int index)
{
    return sequence->seq[index].animData == NULL && sequence->seq[index].varData == NULL;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeSequence_AttachAnimation(aeSequence_t* sequence, aeAnimationData_t* openedAnimData)
{
    aErr_t  err = AERR_NO_ERROR;
    int     index;
    int     newEntryIndex = -1;
    int     openedEntryIndex = -1;
    char    needsUpdate = FALSE;
    aeSequenceInfo_t*  si;
    aeAnimationData_t* copyData;

    AECHECKSEQ(sequence);

    aeSequence_p_Validate(sequence);

    /* Find an empty spot */
    for (index = 0; index < MAX_SEQ_ITEMS && newEntryIndex == -1; index++)
    {
        if (aeSequence_p_EntryNotInUse(sequence, index))
            newEntryIndex = index;
    }

    /* Determine if the data is already in the list in another entry */
    for (index = 0; index < MAX_SEQ_ITEMS && openedEntryIndex == -1; index++)
    {
        if (sequence->seq[index].openedData == openedAnimData)
            openedEntryIndex = index;
    }

    /* Add animation */
    if (newEntryIndex != -1)
    {
        si = &sequence->seq[newEntryIndex];

        si->varData  = NULL;
        si->variant  = aeAnimation_GetCurrentVariant(openedAnimData);

        if (openedEntryIndex == -1)
        {
            /* Copy the data to give to the sequence */
            copyData = aeAnimation_Copy(openedAnimData);

            if (copyData != NULL)
            {
                si->animData = copyData;
                si->openedData = openedAnimData; /* Reference to find the opened window */
            }
            else
                err = AERR_OUT_OF_MEMORY;
        }
        else
        {
            /* Don't store the original data */
            si->animData = NULL;

            /* Refer to another entry instead */
            if (sequence->seq[openedEntryIndex].animData != NULL)
                si->varData  = sequence->seq[openedEntryIndex].animData;
            else
                si->varData  = sequence->seq[openedEntryIndex].varData;

            /* Reference to find the opened window */
            si->openedData = openedAnimData; 

            /* If the data is the window is changed, then we need to save it, */
            /* but we can't do that until a little later, below */
            if (aeAnimation_GetChangedFlag(openedAnimData))
                needsUpdate = TRUE;
        }

        if (err == AERR_NO_ERROR)
        {
            aeAnimation_SetSequence(openedAnimData, sequence);
            sequence->count++;
            aeSequence_p_ActivateAll(sequence, newEntryIndex);

            /* We can't call this until we have updated the sequence count */
            if (needsUpdate)
                err = aeSequence_UpdateAnimation(sequence, openedAnimData);
        }
    }
    else
    {
        err = AERR_OUT_OF_MEMORY;
    }

    if (err == AERR_NO_ERROR)
        aeSequence_p_SetChangedFlag(sequence);

    aeSequence_p_Validate(sequence);

    return err;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_DeleteAnimation(aeSequence_t* sequence, int animIndex)
{
    int                 index;
    aeAnimationData_t*  animData;

    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);

    aeSequence_p_Validate(sequence);

    /* Not legal to call this if the window is open */
    /* This is because a sequence can not notify 'up' to a window */
    /* or an animation's window */
    AEASSERT(sequence->seq[animIndex].openedData == NULL);

    /* Bail if necessary -- just to be safe for now. */
    if (sequence->seq[animIndex].openedData != NULL)
        return; 

    /* Does this entry have animation data? */
    if (sequence->seq[animIndex].animData != NULL)
        /* Determine if animData is used by another animation in the list */
        for (index = 0; index < sequence->count; index++)
            if (sequence->seq[index].varData == sequence->seq[animIndex].animData)
            {
                AEASSERT(index != animIndex);
                /* Move the animation data to the other entry that was referring to it */
                sequence->seq[index].animData = sequence->seq[animIndex].animData;
                sequence->seq[index].varData  = NULL;

                /* For integrity, also set the 'reference' field to point to the animation data */
                /* of the entry to be deleted */
                sequence->seq[animIndex].varData  = sequence->seq[animIndex].animData;
                sequence->seq[animIndex].animData = NULL;

                /* Validate the integrity */
                aeSequence_p_Validate(sequence);

                /* We can only do this for the first one found */
                break;
            }

    /* If selected is at the end, back it down one, but only if we are not deleting the last item */
    if (sequence->selected == sequence->count - 1 && sequence->count > 1)
        sequence->selected = sequence->count - 2;

    /* Same for animated */
    if (sequence->animated == sequence->count - 1 && sequence->count > 1)
        sequence->animated = sequence->count - 2;

    /* Get the data out of the entry */
    /* Note that it may NULL if we have another entry referring to the same data. */
    animData = sequence->seq[animIndex].animData;
    sequence->seq[animIndex].animData = NULL;

    // Free data, if last entry for this animation reference
    if (animData != NULL)
    {
        aeAnimation_Delete(animData);
        animData = NULL;
    }

    for (index = animIndex; index < sequence->count - 1; index++)
    {
        sequence->seq[index] = sequence->seq[index+1];
    }

    /* Reduce the count */
    sequence->count--;

    /* Clean-out the newly available entry */
    memset(&sequence->seq[sequence->count], 0, sizeof(sequence->seq[sequence->count]));

    aeSequence_p_SetChangedFlag(sequence);

    aeSequence_p_Validate(sequence);

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimationData_t* aeSequence_p_GetAnimData(aeSequence_t* sequence, int index)
{
    aeAnimationData_t* ad;

    AECHECKSEQ(sequence);

    ad = (sequence->seq[index].animData != NULL) ? sequence->seq[index].animData : sequence->seq[index].varData;

    AEASSERT(ad);

    return ad;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    Also called from aeSequence_AttachAnimation if the data being
    attached has been attached before and it has changed
----------------------------------------------------------------------*/
aErr_t aeSequence_UpdateAnimation(aeSequence_t* sequence, aeAnimationData_t* origData)
{
    aErr_t  err = AERR_NO_ERROR;
    int     foundOpenedDataIndex;
    int     foundVarDataIndex;
    int     index;
    int     variantCount;
    aeAnimationData_t* oldData;
    aeAnimationData_t* copyData;

    /* Insure that animation is not already in the list */
    AECHECKSEQ(sequence);
    AEASSERT(origData);

    aeSequence_p_Validate(sequence);

    copyData = aeAnimation_Copy(origData);
    if (copyData == NULL)
        err = AERR_OUT_OF_MEMORY;

    else
    {
        if (aeSequence_FindAnimationUsingOrigRef(sequence, origData, &foundOpenedDataIndex, &foundVarDataIndex))
        {
            /* Get the old one */
            oldData = sequence->seq[foundVarDataIndex].animData;

            /* Place the new one */
            sequence->seq[foundVarDataIndex].animData = copyData;

            /* Replace all references to the old one */
            for (index = 0; index < sequence->count; index++)
                if (sequence->seq[index].varData == oldData)
                    sequence->seq[index].varData = copyData;

            /* Sanity check */
            for (index = 0; index < sequence->count; index++)
                AEASSERT(sequence->seq[index].animData != oldData && sequence->seq[index].varData != oldData);

            /* Keep the variants in range */
            variantCount = aeAnimation_GetVariantCount(copyData);
            for (index = 0; index < sequence->count; index++)
                if ((sequence->seq[index].varData == copyData || sequence->seq[index].animData == copyData) && 
                    sequence->seq[index].variant >= variantCount)
                    sequence->seq[index].variant = variantCount - 1;

            /* Insure that the animation is not hidden */
            aeAnimation_HideAnimatedFrame(copyData, FALSE);

            /* Delete the old one */
            aeAnimation_Delete(oldData);
            oldData = NULL;
        }
        else
        {
            AEASSERTED("could not find openedData in list");
            err = AERR_GENERAL_ERROR;
        }
    }

    aeSequence_p_Validate(sequence);

    return err;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_ClearOpenedDataRef(aeSequence_t* sequence, aeAnimationData_t* origRef)
{
    int     index;
	char	foundOne = FALSE;

    AECHECKSEQ(sequence);
	
    aeSequence_p_Validate(sequence);

    for (index = 0; index < sequence->count; index++)
	{
		if (sequence->seq[index].openedData == origRef)
        {
            sequence->seq[index].openedData = NULL;
            foundOne = TRUE;
        }
	}

    aeSequence_p_Validate(sequence);

    AEASSERT(foundOne);  
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_SetAllMatchingOpenedDataRef(aeSequence_t* sequence, int setIndex)
{
    int     index;

    aeAnimationData_t* animDataRef;

    AECHECKSEQ(sequence);

    /* Get either the animData from the index or the varData, whichever is not NULL */
    animDataRef = (sequence->seq[setIndex].animData != NULL) ? sequence->seq[setIndex].animData : sequence->seq[setIndex].varData;

    /* Validate some stuff */
    AEASSERT(sequence->seq[setIndex].animData == NULL ||  sequence->seq[setIndex].varData == NULL);
    AEASSERT(sequence->seq[setIndex].openedData != NULL);
    AEASSERT(animDataRef);
	
    /* Look through the entire list for any matching references to the animation specified and set this all to open */
    for (index = 0; index < sequence->count; index++)
	{
		if ((index != setIndex) && (sequence->seq[index].animData == animDataRef || sequence->seq[index].varData == animDataRef))
            sequence->seq[index].openedData = sequence->seq[setIndex].openedData;
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
const char* aeSequence_GetName(aeSequence_t* sequence)
{
    AEASSERT(strlen(sequence->name) > 0);
    return sequence->name;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_SetName(aeSequence_t* sequence, char* name)
{
    strncpy(sequence->name, name, ANIM_NAME_LEN);
    sequence->name[ANIM_NAME_LEN-1] = 0;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeSequence_GetCount(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);

    return sequence->count;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeSequence_Selected_GetIndex(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);
    AEASSERT((sequence->selected == 0) || (sequence->selected >= 0 && sequence->selected < sequence->count));
    return sequence->selected;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Selected_SetIndex(aeSequence_t* sequence, int animIndex)
{
    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);

    sequence->selected = animIndex;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Selected_UpdateIndex(aeSequence_t* sequence, aePosition_t position)
{
    AECHECKSEQ(sequence);

    if (sequence->count > 0)
    {
        AECHECKIDX(sequence, sequence->selected);
    
        switch (position)
        {
            case aePosition_eFirst:
                sequence->selected = 0;
                break;
    
            case aePosition_ePrev:
                if (sequence->selected > 0)
                    sequence->selected--;
                break;
    
            case aePosition_eCurrent:
                break;
    
            case aePosition_eNext:
                if (sequence->selected < sequence->count - 1)
                    sequence->selected++;
                break;
    
            case aePosition_eLast:
                sequence->selected = sequence->count - 1;
                break;
    
            default:
                AEASSERTED("Unknown aePosition_t value");
                break;
        }
    
    
        AECHECKIDX(sequence, sequence->selected);
    }

}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Selected_Move(aeSequence_t* sequence, aePosition_t position)
{
    int                 fromIndex = 0;
    int                 toIndex = 0;
    aeSequenceInfo_t    si;

    AECHECKSEQ(sequence);

    if (sequence->count > 0)
    {
        AEASSERT(sequence->selected >= 0 && sequence->selected < sequence->count);
    
        /* Initial 'from' and 'to indices */
        fromIndex = sequence->selected;
        toIndex   = fromIndex;
    
        /* Determine new 'to' index', if any */
        switch (position)
        {
            case aePosition_eFirst:
                toIndex = 0;
                break;
    
            case aePosition_ePrev:
                if (toIndex > 0)
                    toIndex--;
                break;
    
            case aePosition_eCurrent:
                break;
    
            case aePosition_eNext:
                if (toIndex < sequence->count - 1)
                    toIndex++;
                break;
    
            case aePosition_eLast:
                toIndex = sequence->count - 1;
                break;
    
            default:
                AEASSERTED("Unknown aePosition_t value");
                break;
        }
    
        /* If indicies are different, then swap entries */
        if (fromIndex != toIndex)
        {
            si = sequence->seq[toIndex];
    
            sequence->seq[toIndex] = sequence->seq[fromIndex];
            sequence->seq[fromIndex] = si;
    
            /* Remember to keep pointing to the same selected item */
            sequence->selected = toIndex;

            /* If one animation refered to another, then keep the */
            /* source animation data at the lower index */
            /* 2009.05.28: Also must have animData; that is, don't compare NULLs */
            if (sequence->seq[fromIndex].animData != NULL && 
                sequence->seq[fromIndex].animData == sequence->seq[toIndex].varData && 
                fromIndex > toIndex)
            {
                /* Set the animData in the lower index */
                sequence->seq[toIndex].animData = sequence->seq[toIndex].varData;
                sequence->seq[toIndex].varData = NULL;

                /* And the varData in the higher index */
                sequence->seq[fromIndex].varData = sequence->seq[fromIndex].animData;
                sequence->seq[fromIndex].animData = NULL;

            }
            else if (sequence->seq[toIndex].animData != NULL && 
                     sequence->seq[toIndex].animData == sequence->seq[fromIndex].varData && 
                     toIndex > fromIndex)
            {
                /* Set the animData in the lower index */
                sequence->seq[fromIndex].animData = sequence->seq[fromIndex].varData;
                sequence->seq[fromIndex].varData = NULL;

                /* And the varData in the higher index */
                sequence->seq[toIndex].varData = sequence->seq[toIndex].animData;
                sequence->seq[toIndex].animData = NULL;
            }
            
            aeSequence_p_SetChangedFlag(sequence);

            aeSequence_p_Validate(sequence);
        }
    
    
        AEASSERT(sequence->selected >= 0 && sequence->selected < sequence->count);
    }
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_p_ActivateAll(aeSequence_t* sequence, int animIndex)
{
    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);

    sequence->seq[animIndex].active  = 0xff;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeSequence_p_GetActive(aeSequence_t* sequence, int animIndex, int activeSet)
{
    char    isAct;
    char    actFlags;
    int     mask;

    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);
    AEASSERT(activeSet >= 0 && activeSet < sequence->activeSetCount);

    mask     = (1 << activeSet);
    actFlags = sequence->seq[animIndex].active;
    isAct    = ((actFlags & mask) != 0);

    return isAct;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_p_SetActive(aeSequence_t* sequence, int animIndex, int activeSet, char active)
{
    char    isAct;
    char    actFlags;
    int     mask;

    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);
    AEASSERT(activeSet >= 0 && activeSet < sequence->activeSetCount);

    /* Get the current active flag */
    mask     = (1 << activeSet);
    actFlags = sequence->seq[animIndex].active;
    isAct    = ((actFlags & mask) != 0);

    if ((isAct != 0) != (active != 0))
    {
        if (active)
            actFlags = (actFlags | mask);
        else
            actFlags = (actFlags & ~mask);

        sequence->seq[animIndex].active = actFlags;
        aeSequence_p_SetChangedFlag(sequence);
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_ActiveSet_Append(aeSequence_t* sequence)
{
    int     index;
    int     curSet;
    int     newSet;
    char    isAct;

    if (sequence->activeSetCount < AESEQ_MAX_ACTIVE_SETS)
    {
        curSet = sequence->currActiveSet;
        newSet = sequence->activeSetCount;

        /* Increase the count here to avoid the asserts in the calls */
        /* in the loop, below */
        sequence->activeSetCount++;

        /* Copy the current active flags to the new set */
        for (index = 0; index < sequence->count; index++)
        {
            isAct = aeSequence_p_GetActive(sequence, index, curSet);
            aeSequence_p_SetActive(sequence, index, newSet, isAct);
        }

        aeSequence_p_SetChangedFlag(sequence);

        /* Select the newly appended set at the current set */
        sequence->currActiveSet = sequence->activeSetCount - 1;
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_ActiveSet_DeleteCurr(aeSequence_t* sequence)
{
    int     index;
    int     setIndex;
    char    isAct;

    /* Don't delete the last one! */
    if (sequence->activeSetCount > 1)
    {
        /* Move all the sets 'down' to replace the deleted set */
        for (setIndex = sequence->currActiveSet; setIndex < sequence->activeSetCount - 1; setIndex++)
        {
            for (index = 0; index < sequence->count; index++)
            {
                isAct = aeSequence_p_GetActive(sequence, index, setIndex + 1);
                aeSequence_p_SetActive(sequence, index, setIndex, isAct);
            }
        }

        sequence->activeSetCount--;

        if (sequence->currActiveSet > sequence->activeSetCount - 1)
            sequence->currActiveSet = sequence->activeSetCount - 1;

        aeSequence_p_SetChangedFlag(sequence);
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_ActiveSet_Select(aeSequence_t* sequence, aePosition_t position)
{
    AECHECKSEQ(sequence);

    switch (position)
    {
        case aePosition_eFirst:
            sequence->currActiveSet = 0;
            break;

        case aePosition_ePrev:
            if (sequence->currActiveSet > 0)
                sequence->currActiveSet--;
            break;

        case aePosition_eCurrent:
            break;

        case aePosition_eNext:
            if (sequence->currActiveSet < sequence->activeSetCount - 1)
                sequence->currActiveSet++;
            break;

        case aePosition_eLast:
            sequence->currActiveSet = sequence->activeSetCount - 1;
            break;

        default:
            AEASSERTED("Unknown aePosition_t value");
            break;
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeSequence_ActiveSet_GetCount(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);
    AEASSERT(sequence->activeSetCount > 0 && sequence->activeSetCount <= AESEQ_MAX_ACTIVE_SETS);

    return sequence->activeSetCount;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeSequence_ActiveSet_GetIndex(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);
    AEASSERT(sequence->activeSetCount > 0 && sequence->activeSetCount <= AESEQ_MAX_ACTIVE_SETS);
    AEASSERT(sequence->currActiveSet >= 0 && sequence->currActiveSet < sequence->activeSetCount);
    return sequence->currActiveSet;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeSequence_ActiveSet_GetCountItems(aeSequence_t* sequence)
{
    int activeItems = 0;
    int index;

    AECHECKSEQ(sequence);
    AEASSERT(sequence->activeSetCount > 0 && sequence->activeSetCount <= AESEQ_MAX_ACTIVE_SETS);

    for (index = 0; index < sequence->count; index++)
        if (aeSequence_Anim_IsActive(sequence, index))
            activeItems++;

    return activeItems;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeAnimationData_t* aeSequence_Anim_Get(aeSequence_t* sequence, int animIndex)
{
    aeAnimationData_t* ad;
    ad = aeSequence_p_GetAnimData(sequence, animIndex);

    return ad;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
const char* aeSequence_Anim_GetName(aeSequence_t* sequence, int animIndex)
{
    aeAnimationData_t* ad;
    AECHECKSEQ(sequence);

    ad = aeSequence_p_GetAnimData(sequence, animIndex);

    return aeAnimation_GetName(ad);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeSequence_Anim_IsOpened(aeSequence_t* sequence, int animIndex)
{
    AECHECKSEQ(sequence);
    AEASSERT(animIndex >= 0 && animIndex < sequence->count);

    return (sequence->seq[animIndex].openedData != NULL);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeSequence_Anim_IsChanged(aeSequence_t* sequence, int animIndex)
{
    aeAnimationData_t* ad;

    AECHECKSEQ(sequence);
    AEASSERT(animIndex >= 0 && animIndex < sequence->count);

    ad = aeSequence_p_GetAnimData(sequence, animIndex);
    return aeAnimation_GetChangedFlag(ad);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeSequence_Anim_IsActive(aeSequence_t* sequence, int animIndex)
{
    char    isAct;
    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);

    isAct = aeSequence_p_GetActive(sequence, animIndex, sequence->currActiveSet);

    return isAct;
}

/*----------------------------------------------------------------------
Summary
    Is Animation Data Needed by an Active Animation?

Desc
    The routine returns TRUE is the animation data in this entry
    is needed by some other animation. 

    If this entry does not have animdata then this will return false.

    Note that this also references the "active set"
----------------------------------------------------------------------*/
char aeSequence_Anim_HasNeededAnimData(aeSequence_t* sequence, int animIndex)
{
    char    isDataNeeded;
    int     i;

    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);

    isDataNeeded = FALSE;

    /* First, the anim must have data */
    if (aeSequence_Anim_HasOrigData(sequence, animIndex))
    {
        /* Is this animation active? If so, then we need the data */
        isDataNeeded = aeSequence_p_GetActive(sequence, animIndex, sequence->currActiveSet);
    
        /* Check all other animations to find any that use this data */
        /* and are active */
        for (i = 0; i < sequence->count && !isDataNeeded; i++)
        {
            if (aeSequence_Anim_IsActive(sequence, i) &&
                aeSequence_p_GetAnimData(sequence, animIndex) == aeSequence_p_GetAnimData(sequence, i))
                isDataNeeded = TRUE;
        }
    }

    return isDataNeeded;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Anim_SetActive(aeSequence_t* sequence, int animIndex, char active)
{
    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);

    aeSequence_p_SetActive(sequence, animIndex, sequence->currActiveSet, active);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeSequence_Anim_HasOrigData(aeSequence_t* sequence, int animIndex)
{
    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, animIndex);

    return (sequence->seq[animIndex].animData != NULL);

}

/*----------------------------------------------------------------------
Summary
    Is Animation Active And Unique So Far

Desc
    Odd name...this routine will determine if this the first time that
    this animation/variant combination has appeared in the list as an
    active animation
----------------------------------------------------------------------*/
char aeSequence_Anim_IsFirstActiveOccurance(aeSequence_t* sequence, int animIndex)
{
    int     i;
    int     variant;
    char    firstActive;
    aeAnimationData_t*  animData;

    AECHECKSEQ(sequence);
    AEASSERT(animIndex >= 0 && animIndex < sequence->count);

    animData = aeSequence_p_GetAnimData(sequence, animIndex);
    variant  = sequence->seq[animIndex].variant;

    /* Is it active? */
    firstActive = aeSequence_Anim_IsActive(sequence, animIndex);

    /* Loop until will find another or until we are about to reach the */
    /* target index; note that we don't enter the loop if it was not */
    /* active to start with */
    for (i = 0; i < animIndex && firstActive; i++)
    {
        if (aeSequence_Anim_IsActive(sequence, i) &&
            animData == aeSequence_p_GetAnimData(sequence, i) && 
            variant  == sequence->seq[i].variant)
            firstActive = FALSE;
    }

    return firstActive;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeSequence_Anim_GetVariant(aeSequence_t* sequence, int animIndex)
{
    AECHECKSEQ(sequence);
    AEASSERT(animIndex >= 0 && animIndex < sequence->count);

    return sequence->seq[animIndex].variant;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeSequence_Anim_GetVariantCount(aeSequence_t* sequence, int animIndex)
{
    aeAnimationData_t*  animData;
    AECHECKSEQ(sequence);
    AEASSERT(animIndex >= 0 && animIndex < sequence->count);

    animData = aeSequence_p_GetAnimData(sequence, animIndex);

    return aeAnimation_GetVariantCount(animData);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Anim_VariantSelect(aeSequence_t* sequence, int animIndex, aePosition_t position)
{
    int variantCount;
    int oldVariant;

    AECHECKSEQ(sequence);

    variantCount = aeSequence_Anim_GetVariantCount(sequence, animIndex);
    oldVariant = sequence->seq[animIndex].variant;

    switch (position)
    {
        case aePosition_eFirst:
            sequence->seq[animIndex].variant = 0;
            break;

        case aePosition_ePrev:
            if (sequence->seq[animIndex].variant > 0)
                sequence->seq[animIndex].variant--;
            break;

        case aePosition_eCurrent:
            break;

        case aePosition_eNext:
            if (sequence->seq[animIndex].variant < variantCount - 1)
                sequence->seq[animIndex].variant++;
            break;

        case aePosition_eLast:
            sequence->seq[animIndex].variant = variantCount - 1;
            break;

        default:
            AEASSERTED("Unknown aePosition_t value");
            break;
    }

    if (oldVariant != sequence->seq[animIndex].variant)
        aeSequence_p_SetChangedFlag(sequence);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
char aeSequence_GetChangedFlag(aeSequence_t* sequence)
{
    char    changed = FALSE;
    int     index;

    AECHECKSEQ(sequence);

    changed = sequence->changed;

    for (index = 0; index < sequence->count && !changed; index++)
        if (sequence->seq[index].animData != NULL)
            changed = aeAnimation_GetChangedFlag(sequence->seq[index].animData);

    return changed;
}


/*----------------------------------------------------------------------
Summary
    Sets the sequence changed flag

Desc
    This is private since only the sequence knows when it gets changed
----------------------------------------------------------------------*/
void aeSequence_p_SetChangedFlag(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);
    sequence->changed = TRUE;
}

/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/

aErr_t aeSequence_Read(aeSequence_t* sequence, FILE* fileref, char* prefix, char* suffix)
{
    aErr_t  err = AERR_NO_ERROR;
    char    fileline[500];
    char    line[500];
    int     seqcount;
    int		index;
    int     active;
    int     variantIndex;
    int     animDataFollows;
    int     animDataIndex;
    int     layoutWidth;
    int     layoutHeight;
    int     activeSetCount;
    int     subtypeIndex;
    aeAnimationData_t*      animData;
    aSignature_t		    signature;
    aeAnimSubtype_t		    subtype;
    aePlatform_t            platform;
    aeArchitectureInfo_t    archInfo;
    aeAnimationDataSettings_t animSettings;

    /* Get Architecture Info */
    aeArchList_Layout_GetBySignature(sequence->signature, &platform);

    aeArchList_Arch_GetBySignature(platform.archSignature, &archInfo);

    err = aeIO_ReadLineFromFile(fileref, fileline, sizeof(fileline));
    aeIO_RemovePrefixSuffix(fileline, prefix, suffix, line, sizeof(line));
    sscanf(line, "%d %d", &seqcount, &activeSetCount);

    if (err == AERR_NO_ERROR)
    {
        if (seqcount <= 0)
            err = AERR_FILE_DATA_FORMAT;
        else if (seqcount > MAX_SEQ_ITEMS)
            seqcount = MAX_SEQ_ITEMS;

        if (activeSetCount < 1)
            activeSetCount = 1;
        else if (activeSetCount >= AESEQ_MAX_ACTIVE_SETS)
            activeSetCount = AESEQ_MAX_ACTIVE_SETS;
    }

    sequence->count = seqcount;
    sequence->activeSetCount = activeSetCount;

    for (index = 0; index < seqcount && err == AERR_NO_ERROR; index++)
    {
        err = aeIO_ReadLineFromFile(fileref, fileline, sizeof(fileline));
        if (err == AERR_NO_ERROR)
        {
            aeIO_RemovePrefixSuffix(fileline, prefix, suffix, line, sizeof(line));
            sscanf(line, "%d %d %d %d", &active, &variantIndex, &animDataFollows, &animDataIndex);
        }

        /* If variant is false, then what follows in the animation data */
        if (animDataFollows)
        {
            err = aeIO_ReadLineFromFile(fileref, fileline, sizeof(fileline));
            if (err == AERR_NO_ERROR)
            {
                aeIO_RemovePrefixSuffix(fileline, prefix, suffix, line, sizeof(line));
                err = aeArch_ParseSignatureLine(line, &signature, &subtype);
            }
    
            if (err == AERR_NO_ERROR)
            {
                subtypeIndex = aeArch_p_AnimSubtypeToIndex(subtype);

                aeArchList_Layout_GetConfig(signature, &layoutWidth, &layoutHeight);

                aeAnimation_InitDataSettings(&animSettings);
                animSettings.signature    = signature;
                animSettings.subtype      = subtype;
                animSettings.layoutWidth  = layoutWidth;
                animSettings.layoutHeight = layoutHeight;

                animSettings.bitsPerPixel  = archInfo.bitsPerPixel;
                animSettings.bytesPerFrame = archInfo.bytesPerFrame;
                animSettings.bytesPerEntry = archInfo.bytesPerEntry[subtypeIndex];
                animSettings.ledsPerEntry  = archInfo.ledsPerEntry[subtypeIndex]; 

                animData = aeAnimation_Create(&animSettings);
                if (animData == NULL)
                    err = AERR_OUT_OF_MEMORY;
            }
    
            if (err == AERR_NO_ERROR)
            {
                if (platform.initAnimDataRead)
                    (platform.initAnimDataRead)(animData);

                aeAnimation_Read(animData, fileref, prefix, suffix);
            }
    
            if (err == AERR_NO_ERROR)
            {
                sequence->seq[index].animData = animData;
                sequence->seq[index].active   = active;
                sequence->seq[index].varData  = NULL;
                sequence->seq[index].variant  = variantIndex;
                sequence->seq[index].openedData = NULL;
            }
        }
        /* If variant is true, then nothing follows and variantIndex points to a position */
        /* higher in the list to find the animation data */
        else
        {
            /* For a variant, the original data must have already been loaded */
            /* and the original data must exist at the specified index */
            if (animDataIndex >= index || sequence->seq[animDataIndex].animData == NULL)
                err = AERR_FILE_DATA_FORMAT;
            else
            {
                sequence->seq[index].animData = NULL;
                sequence->seq[index].active   = active;
                sequence->seq[index].varData  = sequence->seq[animDataIndex].animData;
                sequence->seq[index].variant  = variantIndex;
                sequence->seq[index].openedData = NULL;
            }
        }
    }

    /* We only need to validate the sequence if we successfully read it */
    if (err == AERR_NO_ERROR)
        aeSequence_p_Validate(sequence);

	return err;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static int aeSequence_FindIndexOfOriginalAnimData(aeSequence_t* sequence, int animIndex)
{
    char    found;
    int		j;
    int     varAnimIndex;

    AECHECKSEQ(sequence);

    varAnimIndex = 0;

    /* If this entry does not have animation data, then find the index of the original animation data */
    if (sequence->seq[animIndex].animData == NULL)
    {
        found = FALSE;

        /* The original animation index must exist prior to this index */
        for (j = 0; j < animIndex && !found; j++)
        {
            if (sequence->seq[j].animData == sequence->seq[animIndex].varData)
            {
                varAnimIndex = j;
                found = TRUE;
            }
        }
        AEASSERT(found);
    }

    return varAnimIndex;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeSequence_Write(aeSequence_t* sequence, FILE* fileref, char* prefix, char* suffix)
{
    aErr_t  err = AERR_NO_ERROR;
    char    line[500];
    int     index;
    int     varAnimIndex;
    int     variantIndex;
    unsigned char	active;
    char	animDataFollows;

    AECHECKSEQ(sequence);

    /* Create the first line: "ANIM:XXXX:SEQU" */
    aeArch_MakeSignatureStr(ASIG_ANIMATION, aeSequence_GetSignature(sequence), ASIG_SEQUENCE, line, sizeof(line));

    fprintf(fileref, "%s%s%s", prefix, line, suffix);

    /* Write the count and name */
    fprintf(fileref, "%s%d %d%s", prefix, sequence->count, sequence->activeSetCount, suffix);

    /* Loop over animation */
    for (index = 0; index < sequence->count; index++)
    {
        /* If this animation is a variant of another, this routine will find the index of the original data */
        varAnimIndex = aeSequence_FindIndexOfOriginalAnimData(sequence, index);

        /* Write the sequence-related info for the animation */
        /* active, variant, data flag if follows, original animation data index */

        /* Whether or not animation is active in exported code */
        active  = sequence->seq[index].active;
        /* Which of the settings variants to use */
        variantIndex = sequence->seq[index].variant;
        /* Whether or not the animation data will follow */
        animDataFollows = (sequence->seq[index].animData != NULL);
        sprintf(line, "%u %d %d %d", active, variantIndex, animDataFollows?1:0, varAnimIndex);
        fprintf(fileref, "%s%s%s", prefix, line, suffix);

        if (sequence->seq[index].animData != NULL)
        {
            aeAnimation_Write(sequence->seq[index].animData, fileref, prefix, suffix);
    
            aeAnimation_ResetChangedFlag(sequence->seq[index].animData);
        }
    }
    
    return err;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static void aeSequence_MakeAnimNamesUnique(aeSequence_t* sequence)
{
    int     indexI;
    int     indexJ;
    char    uniqueStr[ANIM_NAME_UNIQUE_SUFFIX_LEN];

    aeAnimationData_t*  animDataI;
    aeAnimationData_t*  animDataJ;


    /* Iterate over all of the animations, starting with the second */
    for (indexI = 1; indexI < sequence->count; indexI++)
    {
        animDataI = sequence->seq[indexI].animData;

        /* Insure we have one to examine */
        if (animDataI != NULL)
        {
            /* Clear any previous unique suffix */
            aeAnimation_ClearUniqueSuffix(animDataI);

            /* Compare 'I' to all that came before */
            for (indexJ = 0; indexJ < indexI; indexJ++)
            {
                animDataJ = sequence->seq[indexJ].animData;

                if (animDataJ != NULL &&
                    strcmp(aeAnimation_GetName(animDataI), aeAnimation_GetName(animDataJ)) == 0 )
                {
                    /* Make 'I' unique by adding a suffix */
                    sprintf(uniqueStr, "%3.3d", indexI);
                    aeAnimation_SetUniqueSuffix(animDataI, uniqueStr);

                    break;
                }
            }
        }
    }
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeSequence_WriteCode(aeSequence_t* sequence, FILE* fileref)
{
    aErr_t          err = AERR_NO_ERROR;
    aeCodeGen_t     codegen;
    aePlatform_t    platform;

    /* Get Architecture Info */
    aeArchList_Layout_GetBySignature(sequence->signature, &platform);

    aeArchList_CodeGen_GetBySignature(sequence->codegenSignature, &codegen);

    AEASSERT(codegen.codegenfunc);

    /* Probably would be better to simply pass the entire platform structure to the codegen */
    codegen.speedReduction = platform.codegenSpeedReduction;

    if (codegen.requiresUniqueNames)
        aeSequence_MakeAnimNamesUnique(sequence);

    if (codegen.codegenfunc != NULL)
        err = (codegen.codegenfunc)(sequence, &codegen, fileref);


    return err;
}

#if 0
aErr_t aeSequence_WriteCode(aeSequence_t* sequence, FILE* fileref, aeAnimSourceCodeType_t sourceCodeType)
{
    aErr_t  err = AERR_NO_ERROR;
    char    line[500];
    int     index;
    int     animIndex;

    /*------------------------*/
    /* Code File Header       */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeFileHeader);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, line, sizeof(line));
    fprintf(fileref, line);

    /*------------------------*/
    /* Dispatch Header        */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeDispatchHeader);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, line, sizeof(line));
    fprintf(fileref, line);

    /*------------------------*/
    /* Dispatch Entries       */
    /*------------------------*/
    /* Loop over animation */
    for (index = 0; index < sequence->count; index++)
    {
        /* Only write dispatch entries for active animations */
        if (sequence->seq[index].active)
        {
            /* Use either the current index or the variant index, as necessary */
            if (sequence->seq[index].animData != NULL)
                animIndex = index;
            else
                animIndex = aeSequence_FindIndexOfOriginalAnimData(sequence, index);
    
            /* Tell the animation we want to write the dispatch line */
            aeAnimation_IO_SetState(sequence->seq[animIndex].animData, AnimIOState_eCodeDispatchEntry);
            
            /* Specify which variant we want */
            aeAnimation_SetCurrentVariant(sequence->seq[animIndex].animData, sequence->seq[animIndex].variant);
    
            /* Get the line and write it to the file */
            aeAnimation_IO_WriteLine(sequence->seq[animIndex].animData, line, sizeof(line));
            fprintf(fileref, line);
        }

    }

    /*------------------------*/
    /* Dispatch Trailer       */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeDispatchTrailer);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, line, sizeof(line));
    fprintf(fileref, line);



    /*------------------------*/
    /* Animation Entry Points */
    /*------------------------*/
    /* Loop over animation */
    for (index = 0; index < sequence->count; index++)
    {
        /* Only write active routines */
        if (sequence->seq[index].active)
        {
            /* Use either the current index or the variant index, as necessary */
            if (sequence->seq[index].animData != NULL)
                animIndex = index;
            else
                animIndex = aeSequence_FindIndexOfOriginalAnimData(sequence, index);

            /* Specify which variant we want */
            aeAnimation_SetCurrentVariant(sequence->seq[animIndex].animData, sequence->seq[animIndex].variant);

            /* Tell the animation we want to write the routine header */
            aeAnimation_IO_SetState(sequence->seq[animIndex].animData, AnimIOState_eCodeRoutineHeader);
            aeAnimation_IO_WriteLine(sequence->seq[animIndex].animData, line, sizeof(line));
            fprintf(fileref, line);
        }
    }

    /*------------------------*/
    /* Animation Routines (tables) */
    /*------------------------*/
    /* Loop over animations */
    for (index = 0; index < sequence->count; index++)
    {
        /* Only write active routines with animation data */
        if (sequence->seq[index].active && sequence->seq[index].animData != NULL)
        {
            animIndex = index;

            /* Tell the animation we want to write the routine body and trailer */
            aeAnimation_IO_SetState(sequence->seq[animIndex].animData, AnimIOState_eCodeRoutineBody);
            do
            {
                aeAnimation_IO_WriteLine(sequence->seq[animIndex].animData, line, sizeof(line));
                fprintf(fileref, line);
            } while (aeAnimation_IO_GetState(sequence->seq[animIndex].animData) != AnimIOState_eDone);
        }
    }

    /*------------------------*/
    /* Code File Trailer      */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeFileTrailer);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, line, sizeof(line));
    fprintf(fileref, line);


    return err;
}
#endif

#if 0
aErr_t aeSequence_WriteCode(aeSequence_t* sequence, aeArchitecture_t* animArch, FILE* fileref, aeAnimSourceCodeType_t sourceCodeType)
{
    aErr_t  err = AERR_NO_ERROR;
    char    line[500];
    char    index;
    int     animIndex;

    /*------------------------*/
    /* Code File Header       */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeFileHeader);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, animArch, line, sizeof(line));
    fprintf(fileref, line);

    /*------------------------*/
    /* Dispatch Header        */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeDispatchHeader);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, animArch, line, sizeof(line));
    fprintf(fileref, line);

    /*------------------------*/
    /* Dispatch Entries       */
    /*------------------------*/
    /* Loop over animation */
    for (index = 0; index < sequence->count; index++)
    {
        /* Only write dispatch entries for active animations */
        if (sequence->seq[index].active)
        {
            /* Use either the current index or the variant index, as necessary */
            if (sequence->seq[index].animData != NULL)
                animIndex = index;
            else
                animIndex = aeSequence_FindIndexOfOriginalAnimData(sequence, index);
    
            /* Tell the animation we want to write the dispatch line */
            aeAnimation_IO_SetState(sequence->seq[animIndex].animData, AnimIOState_eCodeDispatchEntry);
            
            /* Specify which variant we want */
            aeAnimation_SetCurrentVariant(sequence->seq[animIndex].animData, sequence->seq[animIndex].variant);
    
            /* Get the line and write it to the file */
            aeAnimation_IO_WriteLine(sequence->seq[animIndex].animData, animArch, line, sizeof(line));
            fprintf(fileref, line);
        }

    }

    /*------------------------*/
    /* Dispatch Trailer       */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeDispatchTrailer);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, animArch, line, sizeof(line));
    fprintf(fileref, line);



    /* Loop over animation */
    for (index = 0; index < sequence->count; index++)
    {
        /* Only write active routines */
        if (sequence->seq[index].active)
        {
            /* Use either the current index or the variant index, as necessary */
            if (sequence->seq[index].animData != NULL)
                animIndex = index;
            else
                animIndex = aeSequence_FindIndexOfOriginalAnimData(sequence, index);

            /* Specify which variant we want */
            aeAnimation_SetCurrentVariant(sequence->seq[animIndex].animData, sequence->seq[animIndex].variant);

            /* Tell the animation we want to write the routine header */
            aeAnimation_IO_SetState(sequence->seq[animIndex].animData, AnimIOState_eCodeRoutineHeader);
            aeAnimation_IO_WriteLine(sequence->seq[animIndex].animData, animArch, line, sizeof(line));
            fprintf(fileref, line);

            if (sequence->seq[index].animData != NULL)
            {
                /* Tell the animation we want to write the routine body and trailer */
                aeAnimation_IO_SetState(sequence->seq[animIndex].animData, AnimIOState_eCodeRoutineBody);
                do
                {
                    aeAnimation_IO_WriteLine(sequence->seq[animIndex].animData, animArch, line, sizeof(line));
                    fprintf(fileref, line);
                } while (aeAnimation_IO_GetState(sequence->seq[animIndex].animData) != AnimIOState_eDone);
            }

        }
    }

    /*------------------------*/
    /* Code File Trailer      */
    /*------------------------*/
    aeAnimation_IO_SetState(sequence->seq[0].animData, AnimIOState_eCodeFileTrailer);
    aeAnimation_IO_WriteLine(sequence->seq[0].animData, animArch, line, sizeof(line));
    fprintf(fileref, line);


    return err;
}
#endif
/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aErr_t aeSequence_Save(aeSequence_t* sequence, char* filepath)
{
    aErr_t  err = AERR_NO_ERROR;
	FILE*	f;
    char    prefix[] = "; ";
    char    suffix[] = "\n";

    AECHECKSEQ(sequence);

    aeSequence_p_Validate(sequence);

    f = fopen(filepath, "w");

    if (f != NULL)
    {
        err = aeSequence_Write(sequence, f, prefix, suffix);

        if (err == AERR_NO_ERROR && sequence->codegenSignature != ASIG_UNKNOWN && sequence->count > 0)
        {
            err = aeSequence_WriteCode(sequence, f);
        }

        if (err == AERR_NO_ERROR)
            sequence->changed = FALSE;

        fclose(f);
        f = NULL;
    }
    else
    {
        err = AERR_GENERAL_ERROR;
    }

    return err;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aSignature_t aeSequence_GetSignature(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);
	return sequence->signature;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_GetFilePath(aeSequence_t* sequence, char* path, int pathlen)
{
    AECHECKSEQ(sequence);
	AEASSERT(path && pathlen);
	
    strncpy(path, sequence->path, pathlen);
    path[pathlen - 1] = 0;
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeSequence_SetFilePath(aeSequence_t* sequence, char* path)
{
    AECHECKSEQ(sequence);
    strncpy(sequence->path, path, MAX_ANIM_PATH);
    sequence->path[MAX_ANIM_PATH - 1] = 0;
}



/*----------------------------------------------------------------------
Summary
    Sequence Opened For Edit

Desc
    Called by sequence window when animation is ready to be
    opened for editing
----------------------------------------------------------------------*/
void aeSequence_Anim_OpenedForEdit(aeSequence_t* sequence, int animIndex, aeAnimationData_t* refData)
{
    AECHECKSEQ(sequence);
	AEASSERT(animIndex >= 0 && animIndex < sequence->count);
	
    /* Tell the sequence the animation is opened */
    sequence->seq[animIndex].openedData = refData;

}


/*----------------------------------------------------------------------
Summary
    Returns the reference to find the open window

Desc
    
----------------------------------------------------------------------*/
aeAnimationData_t*  aeSequence_Anim_GetOpenedDataRef(aeSequence_t* sequence, int animIndex)
{
    AECHECKSEQ(sequence);
    AEASSERT(animIndex >= 0 && animIndex < sequence->count);

    return sequence->seq[animIndex].openedData;
}

/*----------------------------------------------------------------------
Summary
    Returns a copy of the animation data for editing

Desc
    
----------------------------------------------------------------------*/
aeAnimationData_t* aeSequence_Anim_GetDataCopyForEdit(aeSequence_t* sequence, int animIndex)
{
    aeAnimationData_t* animData;
    aeAnimationData_t* copyData;
    int                variant;

    AECHECKSEQ(sequence);
    animData = aeSequence_p_GetAnimData(sequence, animIndex);

    copyData = aeAnimation_Copy(animData);

    variant = sequence->seq[animIndex].variant;
    aeAnimation_SetCurrentVariant(copyData, variant);

    return copyData;
}

/*----------------------------------------------------------------------
Summary
    Resets the animation of the sequence

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Animated_Reset(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);

    sequence->animated = 0;

    if (sequence->count > 0)
        aeAnimation_Reset(aeSequence_Animated_GetData(sequence));
}

/*----------------------------------------------------------------------
Summary
    Returns the index of the currently animated animation 

Desc
    Used by the seqwindow to identify the animation
----------------------------------------------------------------------*/
int aeSequence_Animated_GetIndex(aeSequence_t* sequence)
{
    AECHECKSEQ(sequence);

    return sequence->animated;
}

/*----------------------------------------------------------------------
Summary
    Advances the animation

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Animated_Advance(aeSequence_t* sequence)
{
    aeAnimationData_t* animData;

    AECHECKSEQ(sequence);

    /* We can only animate the data if we have some */
    if (sequence->count > 0)
    {
        /* Get the currently animated data */
        animData = aeSequence_Animated_GetData(sequence);

        /* Advance it */
        aeAnimation_Advance(animData);

        /* If the current animation has finish, then move on to the next */
        if (aeAnimation_IsAtInitialAnimationState(animData))
            aeSequence_Animated_Select(sequence, aePosition_eNext, TRUE);
    }
}

/*----------------------------------------------------------------------
Summary
    Select which animation to run

Desc
    
----------------------------------------------------------------------*/
void aeSequence_Animated_Select(aeSequence_t* sequence, aePosition_t position, char allowLoop)
{
    int prevAnimated;
    aeAnimationData_t* animData;

    AECHECKSEQ(sequence);
    AECHECKIDX(sequence, sequence->animated);

    prevAnimated = sequence->animated;


    switch (position)
    {
        case aePosition_eFirst:
            sequence->animated = 0;
            break;

        case aePosition_ePrev:
            if (sequence->animated > 0)
                sequence->animated--;
            else if (allowLoop)
                sequence->animated = sequence->count - 1;
            break;

        case aePosition_eCurrent:
            break;

        case aePosition_eNext:
            if (sequence->animated < sequence->count - 1)
                sequence->animated++;
            else if (allowLoop)
                sequence->animated = 0;
            break;

        case aePosition_eLast:
            sequence->animated = sequence->count - 1;
            break;

        default:
            AEASSERTED("Unknown aePosition_t value");
            break;
    }

    if (prevAnimated != sequence->animated)
    {
        /* Reset the new animation */
        animData = aeSequence_Animated_GetData(sequence);
        aeAnimation_Reset(animData);
        aeAnimation_BuildFrame(animData);
    }

    AECHECKIDX(sequence, sequence->animated);
}

/*----------------------------------------------------------------------
Summary
    Get animated speed

Desc
    
----------------------------------------------------------------------*/
int aeSequence_Animated_GetSpeed(aeSequence_t* sequence)
{
    int speed = 0;
    aeAnimationData_t* animData;

    AECHECKSEQ(sequence);

    /* We can only animate the data if we have some */
    if (sequence->count > 0)
    {
        animData = aeSequence_Animated_GetData(sequence);
        speed = aeAnimation_GetSpeed(animData);
    }

    return speed;
}

/*----------------------------------------------------------------------
Summary
    Get the current animated data

Desc
    Used by the sequence window to show the animation
----------------------------------------------------------------------*/
aeAnimationData_t* aeSequence_Animated_GetData(aeSequence_t* sequence)
{
    aeAnimationData_t*  animData = NULL;
    int                 variant;

    AECHECKSEQ(sequence);

    if (sequence->count > 0)
    {
        AECHECKIDX(sequence, sequence->animated);

        animData = aeSequence_Anim_Get(sequence, sequence->animated);

        /* Set the variant */
        variant = aeSequence_Anim_GetVariant(sequence, sequence->animated);
        aeAnimation_SetCurrentVariant(animData, variant);
    }

    return animData;
}

/*----------------------------------------------------------------------
Summary
    Returns true when the current animated animation is just beginning

Desc
    Used by the sequence window to update the timer duration
----------------------------------------------------------------------*/
char aeSequence_Animated_AnimIsAtInitialAnimationState(aeSequence_t* sequence)
{
    char isInitial = TRUE;
    aeAnimationData_t* animData;

    AECHECKSEQ(sequence);

    /* We can only get the state of the data if we have some */
    if (sequence->count > 0)
    {
        animData = aeSequence_Animated_GetData(sequence);
        isInitial = aeAnimation_IsAtInitialAnimationState(animData);
    }

    return isInitial;
}

/*----------------------------------------------------------------------
Summary
    Returns true when the animated sequense is just beginning

Desc
    Used by the gfxfile to determine when to end write the 
    animated graphics (GIF)
----------------------------------------------------------------------*/
char aeSequence_Animated_SeqIsAtInitialAnimationState(aeSequence_t* sequence)
{
    char isInitial = TRUE;
    aeAnimationData_t* animData;

    AECHECKSEQ(sequence);

    /* If we have no data, then return TRUE */
    if (sequence->count == 0)
        isInitial = TRUE;
    /* If we are not at the first animation, then we are not at the intial state */
    else if (sequence->animated > 0)
        isInitial = FALSE;
    /* If we are the first animation, then we return query the animation to see if */
    /* it is at the initial state */
    else
    {
        animData = aeSequence_Animated_GetData(sequence);
        isInitial = aeAnimation_IsAtInitialAnimationState(animData);
    }

    return isInitial;
}

#if 0
    static char aeSequence_FindAnimationUsingData(aeSequence_t* sequence, aeAnimationData_t* animData, int* foundIndex)
    {
        int index;
        int found = -1;
    
        AEASSERT(sequence && animData);
    
        for (index = 0; index < MAX_SEQ_ITEMS && found == -1; index++)
            if (sequence->seq[index].animData == animData)
                found = index;
    
        if (found != -1 && foundIndex != NULL)
            *foundIndex = found;
    
        return (found != -1);
    }
#endif


    #if 0
static void aeSequence_UpdateEntry(aeSequence_t* sequence, int location, aeAnimationData_t* animData, aeAnimationData_t* openedRef)
{
    aeSequenceInfo_t* si;

    AECHECKSEQ(sequence);
    AEASSERT(animData);
    AEASSERT(animData->currSetting >= 0 && animData->currSetting < animData->settingsCoun);

    si = &sequence->seq[location];

    si->animData = animData;
    si->active   = TRUE;
    si->varData  = NULL;
    si->variant  = animData->currSetting;
    si->openedData = openedRef; /* Reference to find the opened window */

    animData->sequence = sequence;

}
#endif

