/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aedialogs.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aedialogs.h"
#include <windows.h>
#include "aedebug.h"
#include <stdio.h>

static LRESULT aeWnd_Dialog_p_GetName(HINSTANCE hinst, HWND hwndOwner, LPSTR titleStr, LPSTR messageStr, LPSTR nameStr, int nameStrLen);

#define ID_TEXT    100
#define ID_EDIT    101

static char szDialogItemName[256];

    
/*----------------------------------------------------------------------
Summary
    Display an error

Desc
    
----------------------------------------------------------------------*/
void aeuiWnd_Error(char* desc, aErr_t err)
{
    MessageBox(NULL, desc, NULL, MB_ICONEXCLAMATION + MB_OK);
}

/*----------------------------------------------------------------------
Summary
    Display a message

Desc
    
----------------------------------------------------------------------*/
void aeuiWnd_Message(char* desc)
{
    MessageBox(NULL, desc, NULL, MB_ICONINFORMATION + MB_OK);
}

/*----------------------------------------------------------------------
Summary
    Ask user to confirm closing a window with changed data

Desc
    
----------------------------------------------------------------------*/
aeuiWndResponse_t aeuiWnd_DataChangedConfirmClose(const char* name)
{
    char    desc[200];
    int     value; 
    aeuiWndResponse_t response = aeuiWndResponse_eOK;

    sprintf(desc, "'%s' has changed. Save data before closing?", name ? name : "<unknown>");

    value = MessageBox(NULL, desc, NULL, MB_ICONEXCLAMATION + MB_YESNOCANCEL);

    if (value == IDYES)
        response = aeuiWndResponse_eYes;
    else if (value == IDNO)
        response = aeuiWndResponse_eNo;
    else if (value == IDCANCEL)
        response = aeuiWndResponse_eCancel;
    else
        AEASSERTED("Unknown 'response' from Win32 MessageBox");

    return response;
}


/*----------------------------------------------------------------------
Summary
    Ask user to accept or cancel

Desc
    
----------------------------------------------------------------------*/
aeuiWndResponse_t aeuiWnd_AskOKCancel(const char* desc)
{
    int     value; 
    aeuiWndResponse_t response = aeuiWndResponse_eOK;

    value = MessageBox(NULL, desc, NULL, MB_ICONEXCLAMATION + MB_OKCANCEL);

    if (value == IDOK)
        response = aeuiWndResponse_eOK;
    else if (value == IDCANCEL)
        response = aeuiWndResponse_eCancel;
    else
        AEASSERTED("Unknown 'response' from Win32 MessageBox");

    return response;
}

/*----------------------------------------------------------------------
Summary
    Bring a window to the top

Desc
    
----------------------------------------------------------------------*/
void aeuiWnd_BringForward(void* wndcontext)
{
    HWND hWnd = (HWND)wndcontext;

    AEASSERT(wndcontext);

    BringWindowToTop(hWnd);
}


/*----------------------------------------------------------------------
Summary
    Change the window's title

Desc
    
----------------------------------------------------------------------*/
void aeuiWnd_ChangeTitle(void* wndcontext, const char* title)
{
    HWND hWnd = (HWND)wndcontext;

    AEASSERT(wndcontext);

    (void)SetWindowText(hWnd, title);
}


/*----------------------------------------------------------------------
Summary
    Get the maximum window size

Desc

BOOL EnumDisplayDevices(
  __in   LPCTSTR lpDevice,
  __in   DWORD iDevNum,
  __out  PDISPLAY_DEVICE lpDisplayDevice,
  __in   DWORD dwFlags
);


    BOOL EnumDisplaySettings(
  __in   LPCTSTR lpszDeviceName,
  __in   DWORD iModeNum,
  __out  DEVMODE *lpDevMode
);

----------------------------------------------------------------------*/
void aeuiWnd_Window_GetMaxSize(int* maxWidth, int* maxHeight)
{
    char    success;
    int     devIndex;
    int     flags;
    int     modeNum;
    int     width;
    int     height;

    DISPLAY_DEVICE  displayDevice;
    DEVMODE         devMode;

    AEASSERT(maxWidth && maxHeight);

    /* Start with very conservative defaults */
    width  = 640;
    height = 400;

    memset(&devMode, 0, sizeof(devMode));
    devMode.dmSize = sizeof(devMode);

    memset(&displayDevice, 0, sizeof(displayDevice));
    displayDevice.cb = sizeof(displayDevice);

    devIndex = 0;
    flags = 0;
    success = EnumDisplayDevices(NULL, devIndex, &displayDevice, flags);

    if (1 /*success*/)
    {

        /* Get the current settings for the primary monitor */
        modeNum = ENUM_CURRENT_SETTINGS;
        success = EnumDisplaySettings(NULL /*displayDevice.DeviceName*/, modeNum, &devMode);

        if (success)
        {
            width  = devMode.dmPelsWidth;
            height = devMode.dmPelsHeight;
        }
    }

    /* Leave room for the window frame */
    width  -= 2 * 10;
    height -= 25;

    /* Return the values */
    *maxWidth  = width;
    *maxHeight = height;

    #if 0
    /* Test tiny EEE ASUS screen */
    *maxWidth = 640;
    *maxHeight = 400;
    #endif
}



/*----------------------------------------------------------------------
Summary
    

Desc
    GetSystemMetrics Function
    SM_CXBORDER The width of a window border, in pixels. 
    SM_CYBORDER The height of a window border, in pixels

BOOL AdjustWindowRect(      
    LPRECT lpRect,
    DWORD dwStyle,
    BOOL bMenu
);    
----------------------------------------------------------------------*/
/*
*/

/*----------------------------------------------------------------------
Summary
    Present a dialog asking for a string

Desc
    
----------------------------------------------------------------------*/
aeuiWndResponse_t aeuiWnd_RequestString(void* instancecontext, void* wndcontext, char* titleStr, char* messageStr, char* nameStr, int nameStrLen)
{
    aeuiWndResponse_t response = aeuiWndResponse_eCancel;

    HINSTANCE   hinst = (HINSTANCE)instancecontext;
    HWND        hwnd  = (HWND)wndcontext;
    LRESULT     result;

    result = aeWnd_Dialog_p_GetName(hinst, hwnd, titleStr, messageStr, nameStr, nameStrLen);

    if (result == IDOK)
    {
        strncpy(nameStr, szDialogItemName, nameStrLen);
        nameStr[nameStrLen - 1] = 0;

        response = aeuiWndResponse_eOK;
    }

    return response;
}

/*----------------------------------------------------------------------
Summary
    Align a LPWORD

Desc
    
----------------------------------------------------------------------*/
static LPWORD lpwAlign(LPWORD lpIn)
{
    ULONG ul;

    if ( ((int)lpIn % 2) == 1)
        AEDBGN("lpIn was odd: ", (int)lpIn);

    ul = (ULONG)lpIn;
    ul ++;
    ul >>=1;
    ul <<=1;
    return (LPWORD)ul;
}

 
/*----------------------------------------------------------------------
Summary
    DialogProc routine

Desc
    
----------------------------------------------------------------------*/
static BOOL CALLBACK aeWnd_Dialog_p_RequestStringProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{ 

    /* AEDBGN("msg: ", message); */
    switch (message) 
    { 
        case WM_COMMAND: 
            /* AEDBGN("wParam: ", wParam); */
            switch (LOWORD(wParam)) 
            { 
                case IDOK: 
                    if (!GetDlgItemText(hwndDlg, ID_EDIT, szDialogItemName, sizeof(szDialogItemName))) 
                         szDialogItemName[0] = 0; 
                    /* Fall through. */
 
                case IDCANCEL: 
                    EndDialog(hwndDlg, wParam); 
                    return TRUE; 
            } 
    } 
    return FALSE; 
} 


/*----------------------------------------------------------------------
Summary
    Utility routine to add controls to in-memory dialog template

Desc
    
----------------------------------------------------------------------*/
#if 0
static void aeWnd_Dialog_p_AddControl(LPWORD pDialogData, LPWORD pDialogEndAddr, int xLoc, int yLoc, int xSize, int ySize, char controlType, char* label)
{
}
#endif

/*----------------------------------------------------------------------
Summary
    Build a dialog in memory and display it; return when user presses
    OK or Cancel

Desc
    
----------------------------------------------------------------------*/
static LRESULT aeWnd_Dialog_p_GetName(HINSTANCE hinst, HWND hwndOwner, LPSTR titleStr, LPSTR messageStr, LPSTR nameStr, int nameStrLen)
{
    HGLOBAL             hgbl;
    LPDLGTEMPLATE       lpdt;
    LPDLGITEMTEMPLATE   lpdit;
    LPWORD  lpw;
    LPWSTR  lpwsz;
    LRESULT ret;
    int     nchar;
    int     i;

    AEASSERT(messageStr);
    AEASSERT(titleStr);
    AEASSERT(nameStr && nameStrLen);

    hgbl = GlobalAlloc(GMEM_ZEROINIT, 2048);
    if (!hgbl)
        return -1;
 
    lpdt = (LPDLGTEMPLATE)GlobalLock(hgbl);
 
    /* Define a dialog box. */
    lpdt->style = WS_POPUP | WS_BORDER | WS_SYSMENU | DS_MODALFRAME | WS_CAPTION;
    /* Number of controls */
    lpdt->cdit = 4;

    /* Location, relative to parent window */
    lpdt->x  = 10;  
    lpdt->y  = 10;
    lpdt->cx = 120; 
    lpdt->cy = 70;

    /* Point past the structure */
    lpw = (LPWORD)(lpdt + 1);

    /* No menu */
    *lpw++ = 0;
    /* Predefined dialog box class (by default) */
    *lpw++ = 0;

    lpwsz = (LPWSTR)lpw;
    nchar = MultiByteToWideChar(CP_ACP, 0, titleStr, -1, lpwsz, 50);

    /* Round-up to next ODD number */
    nchar = (nchar & ~1) + 1;

    lpw += nchar;

    //-----------------------
    // Define an OK button.
    //-----------------------
    lpw = lpwAlign(lpw);    // Align DLGITEMTEMPLATE on DWORD boundary
    lpdit = (LPDLGITEMTEMPLATE)lpw;
    lpdit->x  = 10; 
    lpdit->y  = 50;
    lpdit->cx = 40; 
    lpdit->cy = 12;
    lpdit->id = IDOK;
    lpdit->style = WS_CHILD | WS_VISIBLE | BS_DEFPUSHBUTTON;

    /* Point past structure and set Button class */
    lpw = (LPWORD)(lpdit + 1);
    *lpw++ = 0xFFFF;
    *lpw++ = 0x0080;

    /* lpwsz = (LPWSTR)lpw; */
    nchar = MultiByteToWideChar(CP_ACP, 0, "OK", -1, (LPWSTR)lpw, 50);
    /* Round-up to next even number */
    nchar = (nchar + 1) & ~1;
    lpw += nchar;

    /* Align to DWORD boundary and indicate no creation data for control */
    lpw = lpwAlign(lpw);
    *lpw++ = 0;

    //-----------------------
    // Define a Cancel button.
    //-----------------------
    /* Align DLGITEMTEMPLATE on DWORD boundary */
    lpw = lpwAlign(lpw);
    lpdit = (LPDLGITEMTEMPLATE)lpw;
    lpdit->x  = 60; 
    lpdit->y  = 50;
    lpdit->cx = 40; 
    lpdit->cy = 12;
    lpdit->id = IDCANCEL;
    lpdit->style = WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON;

    lpw = (LPWORD)(lpdit + 1);
    *lpw++ = 0xFFFF;
    *lpw++ = 0x0080; /* Button class atom */

    /* lpwsz = (LPWSTR)lpw; */
    nchar = MultiByteToWideChar(CP_ACP, 0, "Cancel", -1, (LPWSTR)lpw, 50);

    /* Round-up string length to next even number */
    nchar = (nchar + 1) & ~1;
    lpw += nchar;

    /* Align to DWORD boundary and indicate no creation data for control */
    lpw = lpwAlign(lpw);
    *lpw++ = 0;

    //-----------------------
    // Define an edit text control.
    //-----------------------
    /* Align DLGITEMTEMPLATE on DWORD boundary */
    lpw = lpwAlign(lpw);
    lpdit = (LPDLGITEMTEMPLATE)lpw;
    lpdit->x  = 10; 
    lpdit->y  = 25;
    lpdit->cx = 100; 
    lpdit->cy = 14;
    lpdit->id = ID_EDIT;
    lpdit->style = WS_CHILD | WS_VISIBLE | SS_LEFT | WS_BORDER;

    lpw = (LPWORD)(lpdit + 1);
    *lpw++ = 0xFFFF;
    *lpw++ = 0x0081; /* Edit class */

    lpwsz = (LPWSTR)lpw;
    for (i = 0; i < strlen(nameStr)+1; i++)
        *lpwsz++ = (WCHAR)nameStr[i];
    nchar = strlen(nameStr)+1;

    /* Round-up string length to next even number */
    nchar = (nchar + 1) & ~1;
    lpw += nchar;

    /* Align to DWORD boundary and indicate no creation data for control */
    lpw = lpwAlign(lpw);
    *lpw++ = 0;
    
    //-----------------------
    // Define a static text control.
    //-----------------------
    /* Align DLGITEMTEMPLATE on DWORD boundary */
    lpw = lpwAlign(lpw);
    lpdit = (LPDLGITEMTEMPLATE)lpw;
    lpdit->x  = 10; 
    lpdit->y  = 5;
    lpdit->cx = 100; 
    lpdit->cy = 14;
    lpdit->id = ID_TEXT;
    lpdit->style = WS_CHILD | WS_VISIBLE | SS_LEFT;

    /* Point past the structure */
    lpw = (LPWORD)(lpdit + 1);
    *lpw++ = 0xFFFF;
    *lpw++ = 0x0082; /* Static class */

    lpwsz = (LPWSTR)lpw;
    for (i = 0; i < strlen(messageStr)+1; i++)
        *lpwsz++ = (WCHAR)messageStr[i];

    nchar = strlen(messageStr)+1;

    /* Round-up to next even number */
    nchar = (nchar + 1) & ~1;
    lpw += nchar;
    
    /* Align to DWORD boundary and indicate no creation data for control */
    lpw = lpwAlign(lpw);
    *lpw++ = 0;
    



    GlobalUnlock(hgbl); 
    ret = DialogBoxIndirect(hinst, (LPDLGTEMPLATE)hgbl, hwndOwner, (DLGPROC)aeWnd_Dialog_p_RequestStringProc); 
    GlobalFree(hgbl); 

    return ret;
}


/*----------------------------------------------------------------------
Summary
    Dialog test routine
    
Desc
    
----------------------------------------------------------------------*/
void aeuiWnd_TestDialog(void* instancecontext, void* wndcontext)
{
    HINSTANCE   hinst = (HINSTANCE)instancecontext;
    HWND        hwnd  = (HWND)wndcontext;
    char        name[50];

    strcpy(name, "David Kaufman");

    aeWnd_Dialog_p_GetName(hinst, hwnd, "Example dialog2", "Test message", name, sizeof(name));

}
