/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aefivesquare.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aefivesquare.h"
#include "aedebug.h"
#include "asig.h"
#include "aeanimation.h"
#include "aearchlist.h"

#include <stdio.h>

#define AEFIVESQUARE_COLOR_BG       0x000000
#define AEFIVESQUARE_COLOR_ON       0xff2020
#define AEFIVESQUARE_COLOR_OFF_EDIT 0xe0d0d0
#define AEFIVESQUARE_COLOR_OFF_ANIM 0xf0f0f0
#define AEFIVESQUARE_COLOR_OFF_GIF  0x202020

#define AEMATRIX_WIDTH  5
#define AEMATRIX_HEIGHT 5


/* Layout */
static void         aeFiveSquare_Layout_CreateAnim(aeLayout_t* layout);
static void         aeFiveSquare_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData);

static aeColor_t    aeFiveSquare_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex);
static aeColor_t    aeFiveSquare_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex);
static void         aeFiveSquare_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex);
static int          aeFiveSquare_Animation_GetElementColorList(aeAnimationData_t* animData, aeColor_t* colorList, int colorListSize);

static void         aeFiveSquare_InitData(aeAnimationData_t* data);
static void         aeFiveSquare_InitDataRead(aeAnimationData_t* data);


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeFiveSquare_Arch_Register(void)
{
    aeArchitectureInfo_t    archInfo;

    archInfo.signature = ASIG_FIVESQUARE;

    archInfo.subtypeFlags = 
        aeAnimSubtypeFlag_eFrameByFrame |
        aeAnimSubtypeFlag_ePixelList |
        aeAnimSubtypeFlag_eStrip;

    /* LEDs either on or off */
    archInfo.bitsPerPixel = 1;

    /* Needs four bytes to hold 25 bits */
    archInfo.bytesPerFrame = 4; 

    archInfo.bytesPerEntry[aeAnimSubtypeFlagBit_eFrameByFrame] = 4;
    archInfo.ledsPerEntry[aeAnimSubtypeFlagBit_eFrameByFrame]  = 25;

    archInfo.bytesPerEntry[aeAnimSubtypeFlagBit_ePixelList]    = 1;
    archInfo.ledsPerEntry[aeAnimSubtypeFlagBit_ePixelList]     = 1;

    archInfo.bytesPerEntry[aeAnimSubtypeFlagBit_eStrip]        = 1;
    archInfo.ledsPerEntry[aeAnimSubtypeFlagBit_eStrip]         = 5;


    aeArchList_Arch_Add(&archInfo);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeFiveSquare_Layout_Register(void)
{
    aePlatform_t    platform;

    memset(&platform, 0, sizeof(platform));

    strcpy(platform.name, "FiveSquare");

    platform.layoutSignature         = ASIG_FIVESQUARE;
    platform.archSignature           = ASIG_FIVESQUARE;

    platform.layoutWidth             = 5;
    platform.layoutHeight            = 5;

    platform.elementShape            = aeElementShape_eUnknown;

    platform.animGetElementColor     = aeFiveSquare_Animation_GetElementColor;
    platform.animGetElementColorList = aeFiveSquare_Animation_GetElementColorList;
    platform.dataAdvanceElementColor = aeFiveSquare_Data_AdvanceElementColor;
    platform.dataGetElementColor     = aeFiveSquare_Data_GetElementColor;
    platform.layoutCreateAnim        = aeFiveSquare_Layout_CreateAnim;
    platform.layoutCreateEdit        = aeFiveSquare_Layout_CreateEdit;

    platform.initAnimData            = aeFiveSquare_InitData;
    platform.initAnimDataRead        = aeFiveSquare_InitDataRead;

    /* Recommended or meaningful subtype flags */
    platform.subtypeFlags = 
        aeAnimSubtypeFlag_eFrameByFrame |
        aeAnimSubtypeFlag_ePixelList |
        aeAnimSubtypeFlag_eStrip;


    aeArchList_Layout_Add(&platform);
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeFiveSquare_InitData(aeAnimationData_t* animData)
{
    AEASSERT(animData);

	/* Defaults are good for FrameByFrame and PixelList. */

    /* For Strip, we set-up eight columns */
	if (aeAnimation_GetSubtype(animData) == aeAnimSubtype_eStrip)
	{
		int addFrame, i;
		
		aeAnimation_SetWrapStyle(animData, aeAnimStyle_eWrapped);
		aeAnimation_SetDirection(animData, aeAnimDirection_eHorizontal);

		if (aeAnimation_GetFrameCount(animData) < 8)
		{
			addFrame = 8 - aeAnimation_GetFrameCount(animData);
			for (i = 0; i < addFrame; i++)
				aeAnimation_AppendFrame(animData);
		}

        aeAnimation_SetNeedsResizeOnFrameCountChange(animData, TRUE);

        /* Clear the changed flag. It was set when the frames were added */
        aeAnimation_ResetChangedFlag(animData);
	}
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeFiveSquare_InitDataRead(aeAnimationData_t* data)
{
    AEASSERT(data);

	if (aeAnimation_GetSubtype(data) == aeAnimSubtype_eStrip)
	{
        aeAnimation_SetNeedsResizeOnFrameCountChange(data, TRUE);
    }
}


/* Layout */

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
static void aeFiveSquare_ComputeLayout(aeLayout_t* layout, int rowsOrColumns, aeAnimDirection_t direction)
{
	int			row;
	int			col;
	aeRect_t	bounds;
	int			ledSpace;
	int			ledSize;
	int			elementIndex;
	int			rowCount;
	int			colCount;
	int			elementCount;
	int			fixedDim;
	int			growDim;

	aeLayoutElement_t* le;

	/* Determine rows and columns */
	rowCount = AEMATRIX_HEIGHT;
	colCount = AEMATRIX_WIDTH;

	rowCount = rowsOrColumns;

	ledSize = 10; /* millimeters */
	ledSpace = 2;

	fixedDim = ledSpace + rowCount * (ledSpace + ledSize);
	growDim  = ledSpace + colCount * (ledSpace + ledSize);

	bounds.left   = 0;
	bounds.top    = 0;

	/* For a vertical arrangment, the bottom grows; for a horizontal arrangement, the right grows */
	bounds.right  = (direction == aeAnimDirection_eVertical) ? growDim  : fixedDim;
	bounds.bottom = (direction == aeAnimDirection_eVertical) ? fixedDim : growDim;

	/* Set bounds for layout */
	aeLayout_SetLayoutExtent(layout, bounds);

	/* Total elements */
	elementCount = rowCount * colCount;

	/* Calculate spacing around each LED */
	/*                1                   2                   3                   4                   5                */
	/* |<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->| */
	/*      1                   2                   3                   4                   5                   6      */
	/* ledSpace = ((bounds.right - bounds.left) - (rowCount * ledSize)) / 6; */


	aeLayout_SetElementCount(layout, elementCount);
	
	for (elementIndex = 0; elementIndex < elementCount; elementIndex++)
	{
        row = elementIndex / colCount;
        col = elementIndex % colCount;

        le = aeLayout_GetElement(layout,elementIndex);

        /* locX and locY contain the center of the LED */
        if (direction == aeAnimDirection_eVertical)
        {
            le->locX = ledSpace + ledSize/2 + (4 - col) * (ledSize + ledSpace);
            le->locY = ledSpace + ledSize/2 + row * (ledSize + ledSpace);
        }
        else /* Horizontal */
        {
            le->locX = ledSpace + ledSize/2 + row * (ledSize + ledSpace);
            le->locY = ledSpace + ledSize/2 + col * (ledSize + ledSpace);
        }

        le->elementType = aeElementType_e10mmRound;

        le->ledCount = 1;
        le->colors[0] = AEFIVESQUARE_COLOR_ON;
	}


}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeFiveSquare_Layout_CreateAnim(aeLayout_t* layout)
{
	aeFiveSquare_ComputeLayout(layout, AEMATRIX_HEIGHT, aeAnimDirection_eVertical);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeFiveSquare_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData)
{
	aeAnimDirection_t	dir = aeAnimDirection_eVertical;
	int					size = AEMATRIX_HEIGHT;


	if (aeAnimation_GetSubtype(animData) == aeAnimSubtype_eStrip)
	{
		dir  = aeAnimation_GetDirection(animData);
		size = aeAnimation_GetFrameCount(animData);
	}
	aeFiveSquare_ComputeLayout(layout, size, dir);
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeFiveSquare_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex)
{
    int          ledIndex;
	aeLedValue_t value;

    ledIndex = elementIndex;

    value = aeAnimation_Edit_GetLedValue(animData, ledIndex);

    if (value == 0)
        value = 1;
    else
        value = 0;

    aeAnimation_Edit_UpdateLedValue(animData, ledIndex, value);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeColor_t aeFiveSquare_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t value;
 
	value = aeAnimation_Edit_GetLedValue(animData, elementIndex);

	return (value) ? AEFIVESQUARE_COLOR_ON : AEFIVESQUARE_COLOR_OFF_EDIT;
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeColor_t aeFiveSquare_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t	value;

	value = aeAnimation_Anim_GetLedValue(animData, elementIndex);

	return (value) ? AEFIVESQUARE_COLOR_ON : AEFIVESQUARE_COLOR_OFF_ANIM;
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeFiveSquare_Animation_GetElementColorList(aeAnimationData_t* animData, aeColor_t* colorList, int colorListSize)
{
    int count = 0;

    colorList[count++] = AEFIVESQUARE_COLOR_BG;
    colorList[count++] = AEFIVESQUARE_COLOR_OFF_GIF;
    colorList[count++] = AEFIVESQUARE_COLOR_ON;

    return count;
}

