/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aehexagon.c
 *
	
	         
		      _2_  _1_  _0_
		      		
		    _6_  _5_  _4_  _3_  
		    
		  _11  _10  _9_  _8_ _7_  
		  
		    _15  _14  _13  _12
		    
		      _18  _17  _16


 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aehexagon.h"
#include "aedebug.h"
#include "asig.h"
#include "aeanimation.h"
#include "aearchlist.h"

//#include <stdio.h>
#include <math.h>

#define AEHEXAGON_COLOR_BG       0xffffff
#define AEHEXAGON_COLOR_ON       0x20ff20
#define AEHEXAGON_COLOR_OFF_EDIT 0xe0d0d0
#define AEHEXAGON_COLOR_OFF_ANIM 0xf0f0f0
#define AEHEXAGON_COLOR_OFF_GIF  0x404040

/* Layout */
static void         aeHexagon_Layout_CreateAnim(aeLayout_t* layout);
static void         aeHexagon_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData);

static aeColor_t    aeHexagon_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex);
static aeColor_t    aeHexagon_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex);
static void         aeHexagon_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex);
static int          aeHexagon_Animation_GetElementColorList(aeAnimationData_t* animData, aeColor_t* colorList, int colorListSize);

static void         aeHexagon_InitData(aeAnimationData_t* data);


void aeHexagon_Arch_Register(void)
{
    aeArchitectureInfo_t    archInfo;

    archInfo.signature = ASIG_HEXAGON;

    archInfo.subtypeFlags = 
        aeAnimSubtypeFlag_eFrameByFrame |
        aeAnimSubtypeFlag_ePixelList;

    /* LEDs either on or off */
    archInfo.bitsPerPixel = 1;

    /* Needs three bytes to hold 19 bits */
    archInfo.bytesPerFrame = 3; 

    archInfo.bytesPerEntry[aeAnimSubtypeFlagBit_eFrameByFrame] = 3;
    archInfo.ledsPerEntry[aeAnimSubtypeFlagBit_eFrameByFrame]  = 19;

    archInfo.bytesPerEntry[aeAnimSubtypeFlagBit_ePixelList]    = 1;
    archInfo.ledsPerEntry[aeAnimSubtypeFlagBit_ePixelList]     = 1;

    aeArchList_Arch_Add(&archInfo);
}

void aeHexagon_Layout_Register(void)
{
    aePlatform_t    platform;

    memset(&platform, 0, sizeof(platform));
    strcpy(platform.name, "Hexagon");

    platform.layoutSignature         = ASIG_HEXAGON;
    platform.archSignature           = ASIG_HEXAGON;

    platform.layoutWidth             = 0;
    platform.layoutHeight            = 0;

    platform.animGetElementColor     = aeHexagon_Animation_GetElementColor;
    platform.animGetElementColorList = aeHexagon_Animation_GetElementColorList;
    platform.dataAdvanceElementColor = aeHexagon_Data_AdvanceElementColor;
    platform.dataGetElementColor     = aeHexagon_Data_GetElementColor;
    platform.layoutCreateAnim        = aeHexagon_Layout_CreateAnim;
    platform.layoutCreateEdit        = aeHexagon_Layout_CreateEdit;

    platform.initAnimData            = aeHexagon_InitData;

    /* Recommended or meaningful subtype flags */
    platform.subtypeFlags = 
        aeAnimSubtypeFlag_eFrameByFrame |
        aeAnimSubtypeFlag_ePixelList;


    aeArchList_Layout_Add(&platform);
}


void aeHexagon_InitData(aeAnimationData_t* data)
{
    AEASSERT(data);

	/* defaults are good. Don't need to set these */
}

/* Layout */

static void aeHexagon_ComputeLayout(aeLayout_t* layout)
{
	int			row = 0;
	int			col = 0;
	aeRect_t	bounds;
	int			ledSpace;
	int			ledSize;
	int			elementIndex;
    int         offset = 0;
    int         vertDelta;
    int         horzDelta;

    aeLayoutElement_t* le;

    /*
            _2_  _1_  _0_
        
          _6_  _5_  _4_  _3_  
        
        _11  _10  _9_  _8_ _7_  
        
          _15  _14  _13  _12
        
            _18  _17  _16
    */

    #define LED_COUNT 19

	ledSize = 10; /* millimeters */
	ledSpace = 2;


    horzDelta = (ledSize + ledSpace);
    vertDelta = sin((60 * 3.14159)/180) * (ledSize + ledSpace);

    bounds.left   = 0;
	bounds.top    = 0;

	bounds.right  = horzDelta * 5 + ledSpace;
	bounds.bottom = vertDelta * 5 + ledSpace * 2;

	/* Set bounds for layout */
	aeLayout_SetLayoutExtent(layout, bounds);

	/* Calculate spacing around each LED */
	/*                1                   2                   3                   4                   5                */
	/* |<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->| */
	/*      1                   2                   3                   4                   5                   6      */
	/* ledSpace = ((bounds.right - bounds.left) - (rowCount * ledSize)) / 6; */


	


	aeLayout_SetElementCount(layout, LED_COUNT);
	
	for (elementIndex = 0; elementIndex < LED_COUNT; elementIndex++)
	{

        /* Convert index to rows and "columns", where column is offset by "offset" */
        if (elementIndex <= 2)
        {
            row    = 0;
            col    = 4 - elementIndex;
            offset = 2;
        }
        else if (elementIndex <= 6)
        {
            row    = 1;
            col    = 7 - elementIndex;
            offset = 1;
        }
        else if (elementIndex <= 11)
        {
            row    = 2;
            col    = 11 - elementIndex;
            offset = 0;
        }
        else if (elementIndex <= 15)
        {
            row    = 3;
            col    = 16 - elementIndex;
            offset = 1;
        }
        else if (elementIndex <= 18)
        {
            row    = 4;
            col    = 20 - elementIndex;
            offset = 2;
        }


        le = aeLayout_GetElement(layout, elementIndex);

        /* locX and locY contain the center of the LED */
        le->locX = ledSpace + ledSize/2 + col * horzDelta - (offset * horzDelta)/2;
        le->locY = ledSpace + ledSize/2 + row * vertDelta;

        le->elementType = aeElementType_e10mmRound;

        le->ledCount = 1;
        le->colors[0] = AEHEXAGON_COLOR_ON;
	}


}

void aeHexagon_Layout_CreateAnim(aeLayout_t* layout)
{
	aeHexagon_ComputeLayout(layout);
}

void aeHexagon_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData)
{
	aeHexagon_ComputeLayout(layout);
}



void aeHexagon_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex)
{
    int          ledIndex;
	aeLedValue_t value;

    ledIndex = elementIndex;

    value = aeAnimation_Edit_GetLedValue(animData, ledIndex);

    if (value == 0)
        value = 1;
    else
        value = 0;

    aeAnimation_Edit_UpdateLedValue(animData, ledIndex, value);
}

aeColor_t aeHexagon_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t value;
 
	value = aeAnimation_Edit_GetLedValue(animData, elementIndex);

	return (value) ? AEHEXAGON_COLOR_ON : AEHEXAGON_COLOR_OFF_EDIT;
}



aeColor_t aeHexagon_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t	value;

	value = aeAnimation_Anim_GetLedValue(animData, elementIndex);

	return (value) ? AEHEXAGON_COLOR_ON : AEHEXAGON_COLOR_OFF_ANIM;
}


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeHexagon_Animation_GetElementColorList(aeAnimationData_t* animData, aeColor_t* colorList, int colorListSize)
{
    int count = 0;

    colorList[count] = AEHEXAGON_COLOR_BG;
    count++;

    colorList[count] = AEHEXAGON_COLOR_OFF_ANIM;
    count++;

    colorList[count] = AEHEXAGON_COLOR_ON;
    count++;

    return count;
}

