/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aematrixlayoututils.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aematrixlayoututils.h"
#include "aedebug.h"
#include "asig.h"
#include "aeanimation.h"
#include "aearchlist.h"

#include <stdio.h>

#define AEFIVESQUARE_COLOR_BG       0xffffff
#define AEFIVESQUARE_COLOR_ON       0xff2020
#define AEFIVESQUARE_COLOR_OFF_EDIT 0xe0d0d0
#define AEFIVESQUARE_COLOR_OFF_ANIM 0xf0f0f0
#define AEFIVESQUARE_COLOR_OFF_GIF  0x202020


/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeMatrix_Layout_GetPlatformDefaults(aePlatform_t* platform)
{
    memset(platform, 0, sizeof(aePlatform_t));

    platform->layoutSignature         = ASIG_UNKNOWN;
    platform->archSignature           = ASIG_UNKNOWN;

    platform->layoutWidth             = 0; /* These need to be set by routine that calls this routine */
    platform->layoutHeight            = 0;

    platform->elementShape            = aeElementShape_eUnknown;

    platform->animGetElementColor     = aeMatrix_Layout_Animation_GetElementColor;
    platform->animGetElementColorList = aeMatrix_Layout_Animation_GetElementColorList;
    platform->dataAdvanceElementColor = aeMatrix_Layout_Data_AdvanceElementColor;
    platform->dataGetElementColor     = aeMatrix_Layout_Data_GetElementColor;
    platform->layoutCreateAnim        = aeMatrix_Layout_CreateAnim;
    platform->layoutCreateEdit        = aeMatrix_Layout_CreateEdit;

    platform->initAnimDataRead        = aeMatrix_Layout_InitDataRead;

    /* Recommended or meaningful subtype flags */
    platform->subtypeFlags = 
        aeAnimSubtypeFlag_eFrameByFrame |
        aeAnimSubtypeFlag_ePixelList |
        aeAnimSubtypeFlag_eStrip;
}




/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeMatrix_Layout_InitDataRead(aeAnimationData_t* data)
{
    AEASSERT(data);

	if (aeAnimation_GetSubtype(data) == aeAnimSubtype_eStrip)
	{
        aeAnimation_SetNeedsResizeOnFrameCountChange(data, TRUE);
    }
}


/* Layout */

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeMatrix_Layout_ComputeLayout(aeLayout_t* layout, int adjustableCount, int cropSize, aeAnimDirection_t direction)
{
	int			row;
	int			col;
	aeRect_t	bounds;
	int			ledSpace;
	int			ledSize;
	int			actualElementIndex;
    int         displayedElementIndex;
	int			rowCount;
	int			displayedColCount;
    int         actualColCount;
	int			displayedElementCount;
	int			fixedDim;
	int			growDim;

	aeLayoutElement_t* le;

	/* Determine rows and columns */
	rowCount = aeLayout_GetLayoutDimensions_Rows(layout);
    actualColCount = aeLayout_GetLayoutDimensions_Cols(layout);

    #if 1 /* Don't support cropping yet */
    cropSize = 0;
    #endif
    
    displayedColCount = (cropSize == 0) ? actualColCount : cropSize;

    AEASSERT(displayedColCount <= actualColCount);
    AEASSERT(rowCount > 0);
    AEASSERT(actualColCount > 0);
    AEASSERT(displayedColCount > 0);

    if (adjustableCount != 0)
        rowCount = adjustableCount;

	ledSize = 10; /* millimeters */
	ledSpace = 2;

	fixedDim = ledSpace +          rowCount * (ledSpace + ledSize);
	growDim  = ledSpace + displayedColCount * (ledSpace + ledSize);

	bounds.left   = 0;
	bounds.top    = 0;

	/* For a vertical arrangment, the bottom grows; for a horizontal arrangement, the right grows */
	bounds.right  = (direction == aeAnimDirection_eVertical) ? growDim  : fixedDim;
	bounds.bottom = (direction == aeAnimDirection_eVertical) ? fixedDim : growDim;

	/* Set bounds for layout */
	aeLayout_SetLayoutExtent(layout, bounds);

	/* Total elements */
	displayedElementCount = rowCount * displayedColCount;

	/* Calculate spacing around each LED */
	/*                1                   2                   3                   4                   5                */
	/* |<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->| */
	/*      1                   2                   3                   4                   5                   6      */
	/* ledSpace = ((bounds.right - bounds.left) - (rowCount * ledSize)) / 6; */

    /* Set the element count according to how many items will be displayed */
	aeLayout_SetElementCount(layout, displayedElementCount);

    /* Iterate over the elements according to the actual number of elements */
	for (actualElementIndex = 0; actualElementIndex < rowCount * actualColCount; actualElementIndex++)
	{
        /* Compute the row and col based on the actual column count */
        row = actualElementIndex / actualColCount;
        col = actualElementIndex % actualColCount;

        /* Only add the element if the computed column is within the displayed region */
        if (col < displayedColCount)
        {
            displayedElementIndex = row * displayedColCount + col;
            AEASSERT(displayedElementIndex < displayedElementCount);

            le = aeLayout_GetElement(layout, displayedElementIndex);
    
            /* locX and locY contain the center of the LED */
            if (direction == aeAnimDirection_eVertical)
            {
                le->locX = ledSpace + ledSize/2 + (aeLayout_GetLayoutDimensions_Cols(layout) - 1 - col) * (ledSize + ledSpace);
                le->locY = ledSpace + ledSize/2 + row * (ledSize + ledSpace);
            }
            else /* Horizontal */
            {
                le->locX = ledSpace + ledSize/2 + row * (ledSize + ledSpace);
                le->locY = ledSpace + ledSize/2 + col * (ledSize + ledSpace);
            }
    
            le->elementType = aeElementType_e10mmRound;
    
            le->ledCount = 1;
            le->colors[0] = AEFIVESQUARE_COLOR_ON;
        }
	}


}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeMatrix_Layout_CreateAnim(aeLayout_t* layout)
{
	aeMatrix_Layout_ComputeLayout(layout, 0, 0, aeAnimDirection_eVertical);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeMatrix_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData)
{
	aeAnimDirection_t	dir = aeAnimDirection_eVertical;
	int					adjustable;
    int                 crop;

    /* Normally, we use the dimensions set by the layout */
    adjustable = 0;
    crop = 0;

    /* For a strip layout, we set the adjustable parameter */
    /* and the direction of the strip */
	if (aeAnimation_GetSubtype(animData) == aeAnimSubtype_eStrip)
	{
		dir  = aeAnimation_GetDirection(animData);
		adjustable = aeAnimation_GetFrameCount(animData);

        if (dir == aeAnimDirection_eHorizontal)
            crop = aeLayout_GetLayoutDimensions_Rows(layout);

	}
	aeMatrix_Layout_ComputeLayout(layout, adjustable, crop, dir);
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
void aeMatrix_Layout_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex)
{
    int          ledIndex;
	aeLedValue_t value;

    ledIndex = elementIndex;

    value = aeAnimation_Edit_GetLedValue(animData, ledIndex);

    if (value == 0)
        value = 1;
    else
        value = 0;

    aeAnimation_Edit_UpdateLedValue(animData, ledIndex, value);
}

/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeColor_t aeMatrix_Layout_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t value;
 
	value = aeAnimation_Edit_GetLedValue(animData, elementIndex);

	return (value) ? AEFIVESQUARE_COLOR_ON : AEFIVESQUARE_COLOR_OFF_EDIT;
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
aeColor_t aeMatrix_Layout_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t	value;

	value = aeAnimation_Anim_GetLedValue(animData, elementIndex);

	return (value) ? AEFIVESQUARE_COLOR_ON : AEFIVESQUARE_COLOR_OFF_ANIM;
}



/*----------------------------------------------------------------------
Summary
    

Desc
    
----------------------------------------------------------------------*/
int aeMatrix_Layout_Animation_GetElementColorList(aeAnimationData_t* animData, aeColor_t* colorList, int colorListSize)
{
    int count = 0;

    colorList[count++] = AEFIVESQUARE_COLOR_BG;
    colorList[count++] = AEFIVESQUARE_COLOR_OFF_ANIM;
    colorList[count++] = AEFIVESQUARE_COLOR_ON;

    return count;
}


