/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aeslakcircletwocolor.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aepov8.h"
#include "aedebug.h"
#include "aeanimation.h"
#include "asig.h"
#include "aearchlist.h"


#define POV8_HEIGHT 8

#define AEPOV8_COLOR_ON  0xff2020
#define AEPOV8_COLOR_OFF 0xf0e0e0
#define AEPOV8_COLOR_OFF_ANIM 0xf8ffff

static void aePOV8_InitData(aeAnimationData_t* animData);
static void aePOV8_InitDataRead(aeAnimationData_t* animData);

/* Layout */
static void         aePOV8_Layout_CreateAnim(aeLayout_t* layout);
static void         aePOV8_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData);

static void         aePOV8_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex);
static aeColor_t    aePOV8_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex);
static aeColor_t    aePOV8_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex);


aeLedValue_t aePOV8_GetLedValue(aeAnimationData_t* animData, int frameIndex, int ledIndex);
void         aePOV8_SetLedValue(aeAnimationData_t* animData, int frameIndex, int ledIndex, aeLedValue_t value);

void aePOV8_Layout_Register(void)
{
    aePlatform_t platform;

    memset(&platform, 0, sizeof(platform));

    platform.layoutSignature            = ASIG_POV8;
    platform.archSignature              = ASIG_SLAK;

    strcpy(platform.name, "POV8");

    platform.layoutWidth  = POV8_HEIGHT; /* Number of LEDs in a column */
    platform.layoutHeight = POV8_HEIGHT; /* Hmm... */


    /* */
    platform.codegenSpeedReduction      = 10;

    platform.initAnimData               = aePOV8_InitData;
    platform.initAnimDataRead           = aePOV8_InitDataRead;

	platform.layoutCreateAnim			= aePOV8_Layout_CreateAnim;
	platform.layoutCreateEdit			= aePOV8_Layout_CreateEdit;

	platform.dataAdvanceElementColor    = aePOV8_Data_AdvanceElementColor;
	platform.dataGetElementColor		= aePOV8_Data_GetElementColor;
	platform.animGetElementColor		= aePOV8_Animation_GetElementColor;

    /* Recommended or meaningful subtype flags */
    platform.subtypeFlags = 
        aeAnimSubtypeFlag_eFrameByFrame;

    aeArchList_Layout_Add(&platform);

}

static void aePOV8_p_InitData_Common(aeAnimationData_t* animData)
{
    aeAnimation_SetEditSubtype(animData, aeAnimSubtype_eStrip);
    aeAnimation_SetNeedsResizeOnFrameCountChange(animData, TRUE);

}

void aePOV8_InitData(aeAnimationData_t* animData)
{
    int addFrame, i;

    AEASSERT(animData);

    if (aeAnimation_GetFrameCount(animData) < 8)
    {
        addFrame = 8 - aeAnimation_GetFrameCount(animData);
        for (i = 0; i < addFrame; i++)
            aeAnimation_AppendFrame(animData);
    }

    /* Clear the changed flag. It was set when the frames were added */
    aeAnimation_ResetChangedFlag(animData);

    aePOV8_p_InitData_Common(animData);
}

void aePOV8_InitDataRead(aeAnimationData_t* animData)
{
    aePOV8_p_InitData_Common(animData);
}

/* Layout */

static void aePOV8_ComputeLayout(aeLayout_t* layout, int columns)
{
    int			row;
    int			col;
    aeRect_t	bounds;
    int			ledSpace;
    int			ledSize;
    int			elementIndex;
    int			elementCount;
    int			vertDim;
    int			horzDim;

    aeLayoutElement_t* le;

    /* LED Info */
    ledSize = 10; /* millimeters */
    ledSpace = 2;

    /* Dimensions */
    vertDim = ledSpace + POV8_HEIGHT * (ledSpace + ledSize);
    horzDim = ledSpace + columns * (ledSpace + ledSize);

    bounds.left   = 0;
    bounds.top    = 0;
    bounds.right  = horzDim;
    bounds.bottom = vertDim;

    /* Set bounds for layout */
    aeLayout_SetLayoutExtent(layout, bounds);

    /* Total elements */
    elementCount = POV8_HEIGHT * columns;

    /* Calculate spacing around each LED */
    /*                1                   2                   3                   4                   5                */
    /* |<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->[---LED---]<-space->| */
    /*      1                   2                   3                   4                   5                   6      */
    /* ledSpace = ((bounds.right - bounds.left) - (rowCount * ledSize)) / 6; */

    aeLayout_SetElementCount(layout, elementCount);

    /* Increasing ElementIndex identifies LEDs vertically */
    for (elementIndex = 0; elementIndex < elementCount; elementIndex++)
    {
        row = elementIndex % POV8_HEIGHT;
        col = elementIndex / POV8_HEIGHT;

        le = aeLayout_GetElement(layout, elementIndex);

        /* locX and locY contain the center of the LED */
        le->locX = ledSpace + ledSize/2 + col * (ledSize + ledSpace);
        le->locY = ledSpace + ledSize/2 + row * (ledSize + ledSpace);

        le->elementType = aeElementType_e10mmRound;

        le->ledCount = 1;
        le->colors[0] = AEPOV8_COLOR_ON;
    }
}


void aePOV8_Layout_CreateAnim(aeLayout_t* layout)
{
    /* No animation layout */
    aeLayout_SetElementCount(layout, 8);
    aePOV8_ComputeLayout(layout, 1);
}

void aePOV8_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData)
{
    int columns;

    columns= aeAnimation_GetFrameCount(animData);
    aePOV8_ComputeLayout(layout, columns);
}



void aePOV8_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex)
{
    int value;

    value = aeAnimation_Edit_GetLedValue(animData, elementIndex);

    value = !value;

    aeAnimation_Edit_UpdateLedValue(animData, elementIndex, value);
}

aeColor_t aePOV8_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t value;
	aeColor_t	colorOn;
    int         ledIndex;

	colorOn = AEPOV8_COLOR_ON;

    ledIndex = elementIndex;
	value = aeAnimation_Edit_GetLedValue(animData, ledIndex);

	return (value) ? colorOn : AEPOV8_COLOR_OFF;
}


aeColor_t aePOV8_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t	value;

    value = aeAnimation_Anim_GetLedValue(animData, elementIndex);

	return (value) ? AEPOV8_COLOR_ON : AEPOV8_COLOR_OFF_ANIM;
}




