/*-----------------------------------------------------------------------------
 *
 *  ANIMATOR
 *  LED Animation Sequence Editor
 *
 *  Copyright David D. Kaufman, 2009
 *  
 *  aeslakcircletwocolor.c
 *
 *
 *
 *
 *
 *------------------------------------------------------------------------------*/

#include "aeslakcircletwocolor.h"
#include "aedebug.h"
#include "aeanimation.h"
#include "asig.h"
#include "aearchlist.h"
#include <math.h>


#define AECIRCLETWO_COLOR_ON_EVEN  0xff2020
#define AECIRCLETWO_COLOR_ON_ODD   0x20ff20
#define AECIRCLETWO_COLOR_OFF      0xe0d0d0
#define AECIRCLETWO_COLOR_OFF_ANIM 0xf8ffff

#define AECIRCLETWO_UNIQUECOUNT		12


/* Layout */
static void         arch_Layout_CreateAnim(aeLayout_t* layout);
static void         arch_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData);

static void         arch_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex);
static aeColor_t    arch_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex);
static aeColor_t    arch_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex);


aeLedValue_t aeSlakCircleTwoColor_GetLedValue(aeAnimationData_t* animData, int frameIndex, int ledIndex);
void         aeSlakCircleTwoColor_SetLedValue(aeAnimationData_t* animData, int frameIndex, int ledIndex, aeLedValue_t value);

void aeSLAK_CircleTwoColor_Layout_Register(void)
{
    aePlatform_t platform;

    memset(&platform, 0, sizeof(platform));

    platform.layoutSignature            = ASIG_CIRCLETWO;
    platform.archSignature              = ASIG_SLAK;

    strcpy(platform.name, "CircleTwoColor");

	platform.layoutCreateAnim			= arch_Layout_CreateAnim;
	platform.layoutCreateEdit			= arch_Layout_CreateEdit;

	platform.dataAdvanceElementColor    = arch_Data_AdvanceElementColor;
	platform.dataGetElementColor		= arch_Data_GetElementColor;
	platform.animGetElementColor		= arch_Animation_GetElementColor;

    /* Recommended or meaningful subtype flags */
    platform.subtypeFlags = 
        aeAnimSubtypeFlag_eFrameByFrame;

    aeArchList_Layout_Add(&platform);

}

/* Layout */

static void aeSlakCircleTwoColor_ComputeLayout(aeLayout_t* layout)
{
	int	i;
	int	centerX, centerY;
	int radius;
	int	angleDelta;
	int	offsetX,  offsetY;
	aeRect_t	lr;
	aeLayoutElement_t* le;

    aeLayout_SetElementCount(layout, AECIRCLETWO_UNIQUECOUNT * 2);
	//layout->elementCount = AECIRCLETWO_UNIQUECOUNT * 2;

	lr.top    = 0;
	lr.left   = 0;
	lr.bottom = 140;
	lr.right  = 140;

    aeLayout_SetLayoutExtent(layout, lr);
	//layout->layoutExtent = lr;


	centerX = lr.right/2;
	centerY = lr.bottom/2;
	radius  = 60;
	angleDelta = 360 / aeLayout_GetElementCount(layout);

	for (i = 0; i < AECIRCLETWO_UNIQUECOUNT; i++)
	{
		offsetX = radius * cos((angleDelta * i * 3.14159)/(180.0));
		offsetY = radius * sin((angleDelta * i * 3.14159)/(180.0));

		le = aeLayout_GetElement(layout, i);

		le->locX = centerX + offsetX;
		le->locY = centerY + offsetY;
		le->elementType = aeElementType_e10mmRound;
		le->ledCount = 1;


		le = aeLayout_GetElement(layout, i + AECIRCLETWO_UNIQUECOUNT);

		le->locX = centerX - offsetX;
		le->locY = centerY - offsetY;
		le->elementType = aeElementType_e10mmRound;
		le->ledCount = 1;
	}
}


static int aeSlakCircleTwoColor_MapElementIndexToLEDIndex(int elementIndex)
{
    int ledIndex;

    ledIndex = elementIndex % AECIRCLETWO_UNIQUECOUNT;

    return ledIndex;
}

void arch_Layout_CreateAnim(aeLayout_t* layout)
{
	aeSlakCircleTwoColor_ComputeLayout(layout);
}

void arch_Layout_CreateEdit(aeLayout_t* layout, aeAnimationData_t* animData)
{
	aeSlakCircleTwoColor_ComputeLayout(layout);
}



void arch_Data_AdvanceElementColor(aeAnimationData_t* animData, int elementIndex)
{
    int ledIndex;
    int value;


    ledIndex = aeSlakCircleTwoColor_MapElementIndexToLEDIndex(elementIndex);

    value = aeAnimation_Edit_GetLedValue(animData, ledIndex);

    value = !value;

    aeAnimation_Edit_UpdateLedValue(animData, ledIndex, value);
}

aeColor_t arch_Data_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t value;
	aeColor_t	colorOn;
    int         ledIndex;

	colorOn = (elementIndex & 1) ? AECIRCLETWO_COLOR_ON_ODD : AECIRCLETWO_COLOR_ON_EVEN;

    ledIndex = aeSlakCircleTwoColor_MapElementIndexToLEDIndex(elementIndex);
	value = aeAnimation_Edit_GetLedValue(animData, ledIndex);

	return (value) ? colorOn : AECIRCLETWO_COLOR_OFF;
}


aeColor_t arch_Animation_GetElementColor(aeAnimationData_t* animData, int elementIndex)
{
	aeLedValue_t	value;
	aeColor_t	colorOn;
    int         ledIndex;

	colorOn = (elementIndex & 1) ? AECIRCLETWO_COLOR_ON_ODD : AECIRCLETWO_COLOR_ON_EVEN;

    ledIndex = aeSlakCircleTwoColor_MapElementIndexToLEDIndex(elementIndex);

    value = aeAnimation_Anim_GetLedValue(animData, ledIndex);

	return (value) ? colorOn : AECIRCLETWO_COLOR_OFF_ANIM;
}




